@@ -40,21 +40,23 @@ let print_usage_common co command =
 \n  -l f                   (idem)\
 \n  -load-vernac-source-verbose f  load Coq file f.v (Load Verbose \"f\".)\
 \n  -lv f	           (idem)\
-\n  -load-vernac-object lib\
-\n                         load Coq library lib (Require lib)\
-\n  -rfrom root lib        load Coq library lib (From root Require lib.)\
+\n  -require lib           load Coq library lib (Require lib)\
 \n  -require-import lib, -ri lib\
 \n                         load and import Coq library lib\
 \n                         (equivalent to Require Import lib.)\
 \n  -require-export lib, -re lib\
 \n                         load and transitively import Coq library lib\
 \n                         (equivalent to Require Export lib.)\
+\n  -require-from root lib, -rfrom root lib
+\n                         load Coq library lib (From root Require lib.)\
 \n  -require-import-from root lib, -rifrom root lib\
 \n                         load and import Coq library lib\
 \n                         (equivalent to From root Require Import lib.)\
 \n  -require-export-from root lib, -refrom root lib\
 \n                         load and transitively import Coq library lib\
 \n                         (equivalent to From root Require Export lib.)\
+\n  -load-vernac-object lib\
+\n                         (obsolete synonym of -require lib)\
 \n\
 \n  -where                 print Coq's standard library location and exit\
 \n  -config, --config      print Coq's configuration information and exit\
@@ -86,7 +88,8 @@ let print_usage_common co command =
 \n  -set \"Foo Bar\"         enable Foo Bar (as Set Foo Bar. in a file)\
 \n  -set \"Foo Bar=value\"   set Foo Bar to value (value is interpreted according to Foo Bar's type)\
 \n  -unset \"Foo Bar\"       disable Foo Bar (as Unset Foo Bar. in a file)\
-\n  -time                  display the time taken by each command\
+\n  -time                  display the time taken by each command, outputting to the message system (stdout for coqc)\
+\n  -time-file f           display the time taken by each command, outputting to file f\
 \n  -profile-ltac          display the time taken by each (sub)tactic\
 \n  -m, --memory           display total heap size at program exit\
 \n                         (use environment variable\
