@@ -13,7 +13,7 @@ type loc = {
   loc_end : Lexing.position;
 }
 
-type code = { code : string; loc : loc; }
+type code = { code : string; loc : loc option; }
 
 type user_symbol =
 | Ulist1 of user_symbol
@@ -112,6 +112,7 @@ type vernac_rule = {
   vernac_class : code option;
   vernac_depr : bool;
   vernac_body : code;
+  vernac_synterp : (string * code) option;
 }
 
 type vernac_ext = {
@@ -149,7 +150,7 @@ type argument_ext = {
 type node =
 | Code of code
 | Comment of string
-| DeclarePlugin of string
+| DeclarePlugin of string option
 | GramExt of grammar_ext
 | VernacExt of vernac_ext
 | VernacArgumentExt of vernac_argument_ext
