@@ -6,7 +6,7 @@
 
 set -e
 
-BENCH_DEBUG=1
+BENCH_DEBUG=
 
 r='\033[0m'          # reset (all attributes off)
 b='\033[1m'          # bold
@@ -16,10 +16,27 @@ bt='`'               # backtick
 start_code_block='```'
 end_code_block='```'
 
+# We put local binaries such as opam in .bin and extend PATH
+BIN=$(pwd)/.bin
+mkdir "$BIN"
+wget https://github.com/ocaml/opam/releases/download/2.1.3/opam-2.1.3-x86_64-linux -O "$BIN"/opam
+chmod +x "$BIN"/opam
+
+export PATH="$BIN":$PATH
+
+echo "Global env info:"
+echo "----------------"
+echo "pwd: $PWD"
+echo "path: $PATH"
+echo "opam version: `opam --version`"
+
 number_of_processors=$(cat /proc/cpuinfo | grep '^processor *' | wc -l)
 
 program_name="$0"
 program_path=$(readlink -f "${program_name%/*}")
+render_results="dune exec --no-print-directory --root $program_path/../.. -- dev/bench/render_results.exe"
+render_line_results="dune exec --no-print-directory --root $program_path/../.. -- dev/bench/render_line_results.exe"
+timelog2html="dune exec --no-print-directory --root $program_path/../.. -- dev/bench/timelog2html.exe"
 
 coqbot_url_prefix="https://coqbot.herokuapp.com/pendulum/"
 
@@ -33,26 +50,30 @@ check_variable () {
   fi
 }
 
-echo $PWD
-
 #check_variable "BUILD_ID"
 #check_variable "BUILD_URL"
 #check_variable "JOB_NAME"
 #check_variable "JENKINS_URL"
-check_variable "CI_JOB_URL"
+#check_variable "CI_JOB_URL"
+#check_variable "CI_REPOSITORY_URL"
 
 : "${coq_pr_number:=}"
 : "${coq_pr_comment_id:=}"
-: "${new_ocaml_switch:=ocaml-base-compiler.4.07.1}"
-: "${old_ocaml_switch:=ocaml-base-compiler.4.07.1}"
-: "${new_coq_repository:=https://gitlab.com/coq/coq.git}"
-: "${old_coq_repository:=https://gitlab.com/coq/coq.git}"
+: "${new_ocaml_version:=4.09.1}"
+: "${old_ocaml_version:=4.09.1}"
+: "${new_coq_repository:=$CI_REPOSITORY_URL}"
+: "${old_coq_repository:=$CI_REPOSITORY_URL}"
 : "${new_coq_opam_archive_git_uri:=https://github.com/coq/opam-coq-archive.git}"
 : "${old_coq_opam_archive_git_uri:=https://github.com/coq/opam-coq-archive.git}"
 : "${new_coq_opam_archive_git_branch:=master}"
 : "${old_coq_opam_archive_git_branch:=master}"
 : "${num_of_iterations:=1}"
-: "${coq_opam_packages:=coq-bignums coq-hott coq-performance-tests-lite coq-engine-bench-lite coq-mathcomp-ssreflect coq-mathcomp-fingroup coq-mathcomp-algebra coq-mathcomp-solvable coq-mathcomp-field coq-mathcomp-character coq-mathcomp-odd-order coq-math-classes coq-corn coq-flocq coq-compcert coq-geocoq coq-color coq-coqprime coq-coqutil coq-bedrock2 coq-rewriter coq-fiat-core coq-fiat-parsers coq-fiat-crypto coq-unimath coq-coquelicot coq-lambda-rust coq-verdi coq-verdi-raft coq-fourcolor coq-rewriter-perf-SuperFast coq-perennial coq-vst}"
+: "${timeout:=3h}"
+: "${coq_opam_packages:=coq-bignums coq-hott coq-performance-tests-lite coq-engine-bench-lite coq-mathcomp-ssreflect coq-mathcomp-fingroup coq-mathcomp-algebra coq-mathcomp-solvable coq-mathcomp-field coq-mathcomp-character coq-mathcomp-odd-order coq-math-classes coq-corn coq-compcert coq-equations coq-metacoq-template coq-metacoq-pcuic coq-metacoq-safechecker coq-metacoq-erasure coq-metacoq-translations coq-geocoq coq-color coq-coqprime coq-coqutil coq-bedrock2 coq-rewriter coq-fiat-core coq-fiat-parsers coq-fiat-crypto-with-bedrock coq-unimath coq-coquelicot coq-iris-examples coq-verdi coq-verdi-raft coq-fourcolor coq-rewriter-perf-SuperFast coq-perennial coq-vst coq-category-theory}"
+: "${coq_native:=}"
+
+new_ocaml_switch=ocaml-base-compiler.$new_ocaml_version
+old_ocaml_switch=ocaml-base-compiler.$old_ocaml_version
 
 new_coq_commit=$(git rev-parse HEAD^2)
 old_coq_commit=$(git merge-base HEAD^1 $new_coq_commit)
@@ -72,26 +93,25 @@ working_dir="$PWD/${bench_dirname}"
 
 log_dir=$working_dir/logs
 mkdir "$log_dir"
+export COQ_LOG_DIR=$log_dir
 
-if [ ! -z "$BENCH_DEBUG" ]
-then
-   echo "DEBUG: ocaml -version = $(ocaml -version)"
-   echo "DEBUG: working_dir = $working_dir"
-   echo "DEBUG: new_ocaml_switch = $new_ocaml_switch"
-   echo "DEBUG: new_coq_repository = $new_coq_repository"
-   echo "DEBUG: new_coq_commit = $new_coq_commit"
-   echo "DEBUG: new_coq_opam_archive_git_uri = $new_coq_opam_archive_git_uri"
-   echo "DEBUG: new_coq_opam_archive_git_branch = $new_coq_opam_archive_git_branch"
-   echo "DEBUG: old_ocaml_switch = $old_ocaml_switch"
-   echo "DEBUG: old_coq_repository = $old_coq_repository"
-   echo "DEBUG: old_coq_commit = $old_coq_commit"
-   echo "DEBUG: old_coq_opam_archive_git_uri = $old_coq_opam_archive_git_uri"
-   echo "DEBUG: old_coq_opam_archive_git_branch = $old_coq_opam_archive_git_branch"
-   echo "DEBUG: num_of_iterations = $num_of_iterations"
-   echo "DEBUG: coq_opam_packages = $coq_opam_packages"
-   echo "DEBUG: coq_pr_number = $coq_pr_number"
-   echo "DEBUG: coq_pr_comment_id = $coq_pr_comment_id"
-fi
+echo "DEBUG: ocaml -version = $(ocaml -version)"
+echo "DEBUG: working_dir = $working_dir"
+echo "DEBUG: new_ocaml_switch = $new_ocaml_switch"
+echo "DEBUG: new_coq_repository = $new_coq_repository"
+echo "DEBUG: new_coq_commit = $new_coq_commit"
+echo "DEBUG: new_coq_opam_archive_git_uri = $new_coq_opam_archive_git_uri"
+echo "DEBUG: new_coq_opam_archive_git_branch = $new_coq_opam_archive_git_branch"
+echo "DEBUG: old_ocaml_switch = $old_ocaml_switch"
+echo "DEBUG: old_coq_repository = $old_coq_repository"
+echo "DEBUG: old_coq_commit = $old_coq_commit"
+echo "DEBUG: old_coq_opam_archive_git_uri = $old_coq_opam_archive_git_uri"
+echo "DEBUG: old_coq_opam_archive_git_branch = $old_coq_opam_archive_git_branch"
+echo "DEBUG: num_of_iterations = $num_of_iterations"
+echo "DEBUG: coq_opam_packages = $coq_opam_packages"
+echo "DEBUG: coq_pr_number = $coq_pr_number"
+echo "DEBUG: coq_pr_comment_id = $coq_pr_comment_id"
+echo "DEBUG: coq_native = $coq_native"
 
 # --------------------------------------------------------------------------------
 
@@ -215,8 +235,8 @@ if [[ $ZULIP_BENCH_BOT ]]; then
     pr_num=${pr_full%%:*}
     pr_msg=${pr_full#*:}
     zulip_header="Bench at $CI_JOB_URL
-For packages $coq_opam_packages
 Testing [$pr_msg](https://github.com/coq/coq/pull/$pr_num)
+On packages $coq_opam_packages
 "
 
     # 24008 is the "github notifications" stream
@@ -237,6 +257,7 @@ Testing [$pr_msg](https://github.com/coq/coq/pull/$pr_num)
 fi
 
 zulip_edit() {
+    if ! [[ $zulip_post ]]; then return; fi
     ending=$1
     if [[ $rendered_results ]]; then
         msg="$zulip_header
@@ -255,7 +276,7 @@ $ending
     fi
     curl -sSX PATCH https://coq.zulipchat.com/api/v1/messages/"$zulip_post" \
          -u "$ZULIP_BENCH_BOT" \
-         --data-urlencode content="$msg" >/dev/null
+         --data-urlencode content="$msg" >/dev/null || echo "Failed to edit zulip post" >&2
 }
 zulip_autofail() {
     code=$?
@@ -264,6 +285,9 @@ zulip_autofail() {
 }
 if [[ $zulip_post ]]; then trap zulip_autofail ERR; fi
 
+# see https://github.com/coq/coq/pull/15807
+ulimit -S -s $((2 * $(ulimit -s)))
+
 # Clone the indicated git-repository.
 
 coq_dir="$working_dir/coq"
@@ -295,15 +319,20 @@ initial_opam_packages="num ocamlfind dune"
 # $2 = compiler name
 # $3 = git hash of Coq to be installed
 # $4 = directory of coq opam archive
+# $5 = use flambda if nonempty
 create_opam() {
 
     local RUNNER="$1"
     local OPAM_DIR="$working_dir/opam.$RUNNER"
-    local OPAM_COMP="$2"
+    local OCAML_VER="$2"
     local COQ_HASH="$3"
     local OPAM_COQ_DIR="$4"
+    local USE_FLAMBDA="$5"
+
+    local OPAM_COMP=ocaml-base-compiler.$OCAML_VER
 
     export OPAMROOT="$OPAM_DIR"
+    export COQ_RUNNER="$RUNNER"
 
     opam init --disable-sandboxing -qn -j$number_of_processors --bare
     # Allow beta compiler switches
@@ -313,12 +342,17 @@ create_opam() {
     # Rest of default switches
     opam repo add -q --set-default iris-dev "https://gitlab.mpi-sws.org/FP/opam-dev.git"
 
-    opam switch create -qy -j$number_of_processors "$OPAM_COMP"
+    if [[ $USE_FLAMBDA ]];
+    then flambda=--packages=ocaml-variants.${OCAML_VER}+options,ocaml-option-flambda
+    else flambda=
+    fi
+
+    opam switch create -qy -j$number_of_processors "ocaml-$RUNNER" "$OPAM_COMP" $flambda
     eval $(opam env)
 
     # For some reason opam guesses an incorrect upper bound on the
     # number of jobs available on Travis, so we set it here manually:
-    opam config set-global jobs $number_of_processors
+    opam var --global jobs=$number_of_processors >/dev/null
     if [ ! -z "$BENCH_DEBUG" ]; then opam config list; fi
 
     opam repo add -q --this-switch coq-extra-dev "$OPAM_COQ_DIR/extra-dev"
@@ -331,20 +365,29 @@ create_opam() {
     if [ ! -z "$BENCH_DEBUG" ]; then opam repo list; fi
 
     cd "$coq_dir"
-    if [ ! -z "$BENCH_DEBUG" ]; then echo "DEBUG: $1_coq_commit = $COQ_HASH"; fi
+    echo "$1_coq_commit = $COQ_HASH"
+
+    echo "wrap-build-commands: [\"$program_path/wrapper.sh\"]" >> "$OPAM_DIR/config"
 
     git checkout -q $COQ_HASH
     COQ_HASH_LONG=$(git log --pretty=%H | head -n 1)
 
     echo "$1_coq_commit_long = $COQ_HASH_LONG"
 
-    for package in coq-core coq-stdlib coq; do
+    if [ ! -z "$coq_native" ]; then opam install coq-native; fi
+
+    for package in coq-core coq-stdlib coqide-server coq; do
+        export COQ_OPAM_PACKAGE=$package
+        export COQ_ITERATION=1
+
+        # build stdlib with -j 1 to get nicer timings
+        local this_nproc=$number_of_processors
+        if [ "$package" = coq-stdlib ]; then this_nproc=1; fi
+
         _RES=0
-        /usr/bin/time -o "$log_dir/$package.$RUNNER.1.time" --format="%U %M %F" \
-                      perf stat -e instructions:u,cycles:u -o "$log_dir/$package.$RUNNER.1.perf" \
-                      opam pin add -y -b -j "$number_of_processors" --kind=path $package.dev . \
-                      3>$log_dir/$package.$RUNNER.opam_install.1.stdout.log 1>&3 \
-                      4>$log_dir/$package.$RUNNER.opam_install.1.stderr.log 2>&4 || \
+        opam pin add -y -b -j "$this_nproc" --kind=path $package.dev . \
+             3>$log_dir/$package.$RUNNER.opam_install.1.stdout.log 1>&3 \
+             4>$log_dir/$package.$RUNNER.opam_install.1.stderr.log 2>&4 || \
             _RES=$?
         if [ $_RES = 0 ]; then
             echo "$package ($RUNNER) installed successfully"
@@ -365,48 +408,81 @@ create_opam() {
 }
 
 # Create an OPAM-root to which we will install the NEW version of Coq.
-create_opam "NEW" "$new_ocaml_switch" "$new_coq_commit" "$new_coq_opam_archive_dir"
+create_opam "NEW" "$new_ocaml_version" "$new_coq_commit" "$new_coq_opam_archive_dir"
 new_coq_commit_long="$COQ_HASH_LONG"
 
 # Create an OPAM-root to which we will install the OLD version of Coq.
-create_opam "OLD" "$old_ocaml_switch" "$old_coq_commit" "$old_coq_opam_archive_dir"
+create_opam "OLD" "$old_ocaml_version" "$old_coq_commit" "$old_coq_opam_archive_dir"
 old_coq_commit_long="$COQ_HASH_LONG"
 
 # Packages which appear in the rendered table
 # Deliberately don't include the dummy "coq" package
 installable_coq_opam_packages="coq-core coq-stdlib"
 
-echo "DEBUG: $program_path/render_results $log_dir $num_of_iterations $new_coq_commit_long $old_coq_commit_long 0 user_time_pdiff $installable_coq_opam_packages"
-rendered_results="$($program_path/render_results "$log_dir" $num_of_iterations $new_coq_commit_long $old_coq_commit_long 0 user_time_pdiff $installable_coq_opam_packages)"
+echo "DEBUG: $render_results $log_dir $num_of_iterations $new_coq_commit_long $old_coq_commit_long 0 user_time_pdiff $installable_coq_opam_packages"
+rendered_results="$($render_results "$log_dir" $num_of_iterations $new_coq_commit_long $old_coq_commit_long 0 user_time_pdiff $installable_coq_opam_packages)"
 echo "${rendered_results}"
 zulip_edit "Benching continues..."
 
+# HTML for stdlib
+# NB: unlike coq_makefile packages, stdlib produces foo.timing not foo.v.timing
+new_base_path=$new_opam_root/ocaml-NEW/.opam-switch/build/coq-stdlib.dev/_build/default/theories/
+old_base_path=$old_opam_root/ocaml-OLD/.opam-switch/build/coq-stdlib.dev/_build/default/theories/
+for vo in $(cd $new_base_path/; find . -name '*.vo'); do
+    if [ -e $old_base_path/$vo ]; then
+        echo "$coq_opam_package/$vo $(stat -c%s $old_base_path/$vo) $(stat -c%s $new_base_path/$vo)" >> "$log_dir/vosize.log"
+    fi
+    if [ -e $old_base_path/${vo%%.vo}.timing ] && \
+           [ -e $new_base_path/${vo%%.vo}.timing ]; then
+        mkdir -p $working_dir/html/coq-stdlib/$(dirname $vo)/
+        # NB: sometimes randomly fails
+        $timelog2html $new_base_path/${vo%%o} \
+                      $old_base_path/${vo%%.vo}.timing \
+                      $new_base_path/${vo%%.vo}.timing > \
+                      $working_dir/html/coq-stdlib/${vo%%o}.html ||
+            echo "Failed (code $?):" $timelog2html $new_base_path/${vo%%o} \
+                 $old_base_path/${vo%%.vo}.timing \
+                 $new_base_path/${vo%%.vo}.timing
+    fi
+done
+
 # --------------------------------------------------------------------------------
 # Measure the compilation times of the specified OPAM packages in both switches
 
 # Sort the opam packages
 sorted_coq_opam_packages=$("${program_path}/sort-by-deps.sh" ${coq_opam_packages})
-if [ ! -z "$BENCH_DEBUG" ]
-then
-   echo "DEBUG: sorted_coq_opam_packages = ${sorted_coq_opam_packages}"
-fi
+echo "sorted_coq_opam_packages = ${sorted_coq_opam_packages}"
+
+failed_packages=
+skipped_packages=
 
 # Generate per line timing info in devs that use coq_makefile
 export TIMING=1
 
 for coq_opam_package in $sorted_coq_opam_packages; do
 
+    export COQ_OPAM_PACKAGE=$coq_opam_package
     if [ ! -z "$BENCH_DEBUG" ]; then
         opam list
-        echo "DEBUG: coq_opam_package = $coq_opam_package"
-        opam show $coq_opam_package || continue 2
+        opam show $coq_opam_package || {
+            failed_packages="$failed_packages
+$coq_opam_package (unknown package)"
+            continue
+            }
     else
         # cause to skip with error if unknown package
-        opam show $coq_opam_package >/dev/null || continue 2
+        opam show $coq_opam_package >/dev/null || {
+            failed_packages="$failed_packages
+$coq_opam_package (unknown package)"
+            continue
+            }
     fi
+    echo "coq_opam_package = $coq_opam_package"
 
     for RUNNER in NEW OLD; do
 
+        export COQ_RUNNER=$RUNNER
+
         # perform measurements for the NEW/OLD commit (provided by the user)
         if [ $RUNNER = "NEW" ]; then
             export OPAMROOT="$new_opam_root"
@@ -421,27 +497,40 @@ for coq_opam_package in $sorted_coq_opam_packages; do
         # If a given OPAM-package was already installed (as a
         # dependency of some OPAM-package that we have benchmarked
         # before), remove it.
-        opam uninstall -q $coq_opam_package
+        opam uninstall -q $coq_opam_package >/dev/null 2>&1
+
+        for dep in $(opam install --show-actions "$coq_opam_package" | grep -o '∗\s*install\s*[^ ]*' | sed 's/∗\s*install\s*//g'); do
+            # show-actions will print transitive deps
+            # so we don't need to look at the skipped_packages
+            if echo "$failed_packages" | grep -q "$dep"; then
+                skipped_packages="$skipped_packages
+$coq_opam_package (dependency $dep failed)"
+                continue 3
+            fi
+        done
 
         # OPAM 2.0 likes to ignore the -j when it feels like :S so we
         # workaround that here.
-        opam config set-global jobs $number_of_processors
+        opam var --global jobs=$number_of_processors >/dev/null
 
         opam install $coq_opam_package -v -b -j$number_of_processors --deps-only -y \
              3>$log_dir/$coq_opam_package.$RUNNER.opam_install.deps_only.stdout.log 1>&3 \
-             4>$log_dir/$coq_opam_package.$RUNNER.opam_install.deps_only.stderr.log 2>&4 || continue 2
+             4>$log_dir/$coq_opam_package.$RUNNER.opam_install.deps_only.stderr.log 2>&4 || {
+            failed_packages="$failed_packages
+$coq_opam_package (dependency install failed in $RUNNER)"
+            continue 2
+            }
 
-        opam config set-global jobs 1
+        opam var --global jobs=1 >/dev/null
 
         if [ ! -z "$BENCH_DEBUG" ]; then ls -l $working_dir; fi
 
         for iteration in $(seq $num_of_iterations); do
+            export COQ_ITERATION=$iteration
             _RES=0
-            /usr/bin/time -o "$log_dir/$coq_opam_package.$RUNNER.$iteration.time" --format="%U %M %F" \
-                 perf stat -e instructions:u,cycles:u -o "$log_dir/$coq_opam_package.$RUNNER.$iteration.perf" \
-                    opam install -v -b -j1 $coq_opam_package \
-                     3>$log_dir/$coq_opam_package.$RUNNER.opam_install.$iteration.stdout.log 1>&3 \
-                     4>$log_dir/$coq_opam_package.$RUNNER.opam_install.$iteration.stderr.log 2>&4 || \
+            timeout "$timeout" opam install -v -b -j1 $coq_opam_package \
+                 3>$log_dir/$coq_opam_package.$RUNNER.opam_install.$iteration.stdout.log 1>&3 \
+                 4>$log_dir/$coq_opam_package.$RUNNER.opam_install.$iteration.stderr.log 2>&4 || \
                 _RES=$?
             if [ $_RES = 0 ];
             then
@@ -459,6 +548,8 @@ for coq_opam_package in $sorted_coq_opam_packages; do
             else
                 # "opam install" failed.
                 echo $_RES > $log_dir/$coq_opam_package.$RUNNER.opam_install.$iteration.exit_status
+                failed_packages="$failed_packages
+$coq_opam_package"
                 continue 3
             fi
         done
@@ -467,6 +558,8 @@ for coq_opam_package in $sorted_coq_opam_packages; do
     installable_coq_opam_packages="$installable_coq_opam_packages $coq_opam_package"
 
     # --------------------------------------------------------------
+    cat $log_dir/$coq_opam_package.$RUNNER.1.*.time || true
+    cat $log_dir/$coq_opam_package.$RUNNER.1.*.perf || true
 
     # Print the intermediate results after we finish benchmarking each OPAM package
     if [ "$coq_opam_package" = "$(echo $sorted_coq_opam_packages | sed 's/ /\n/g' | tail -n 1)" ]; then
@@ -478,15 +571,17 @@ for coq_opam_package in $sorted_coq_opam_packages; do
         :
     else
 
-        echo "DEBUG: $program_path/render_results "$log_dir" $num_of_iterations $new_coq_commit_long $old_coq_commit_long 0 user_time_pdiff $installable_coq_opam_packages"
-        if [ ! -z "$BENCH_DEBUG" ]; then
-            cat $log_dir/$coq_opam_package.$RUNNER.1.time || true
-            cat $log_dir/$coq_opam_package.$RUNNER.1.perf || true
-        fi
-        rendered_results="$($program_path/render_results "$log_dir" $num_of_iterations $new_coq_commit_long $old_coq_commit_long 0 user_time_pdiff $installable_coq_opam_packages)"
+        echo "DEBUG: $render_results "$log_dir" $num_of_iterations $new_coq_commit_long $old_coq_commit_long 0 user_time_pdiff $installable_coq_opam_packages"
+        rendered_results="$($render_results "$log_dir" $num_of_iterations $new_coq_commit_long $old_coq_commit_long 0 user_time_pdiff $installable_coq_opam_packages)"
         echo "${rendered_results}"
         # update the comment
         coqbot_update_comment "" "${rendered_results}" ""
+        msg="Benching continues..."
+        if [ -n "$failed_packages" ]; then
+            msg="$msg
+Failed: $failed_packages
+$skipped_packages"
+        fi
         zulip_edit "Benching continues..."
     fi
 
@@ -495,8 +590,8 @@ for coq_opam_package in $sorted_coq_opam_packages; do
     # N.B. Not all packages end in .dev, e.g., coq-lambda-rust uses .dev.timestamp.
     # So we use a wildcard to catch such packages.  This will have to be updated if
     # ever there is a package that uses some different naming scheme.
-    new_base_path=$new_opam_root/$new_ocaml_switch/.opam-switch/build/$coq_opam_package.dev*/
-    old_base_path=$old_opam_root/$old_ocaml_switch/.opam-switch/build/$coq_opam_package.dev*/
+    new_base_path=$new_opam_root/ocaml-NEW/.opam-switch/build/$coq_opam_package.dev*/
+    old_base_path=$old_opam_root/ocaml-OLD/.opam-switch/build/$coq_opam_package.dev*/
     for vo in $(cd $new_base_path/; find . -name '*.vo'); do
         if [ -e $old_base_path/$vo ]; then
           echo "$coq_opam_package/$vo $(stat -c%s $old_base_path/$vo) $(stat -c%s $new_base_path/$vo)" >> "$log_dir/vosize.log"
@@ -505,11 +600,11 @@ for coq_opam_package in $sorted_coq_opam_packages; do
                [ -e $new_base_path/${vo%%o}.timing ]; then
             mkdir -p $working_dir/html/$coq_opam_package/$(dirname $vo)/
             # NB: sometimes randomly fails
-            $program_path/timelog2html $new_base_path/${vo%%o} \
+            $timelog2html $new_base_path/${vo%%o} \
                                        $old_base_path/${vo%%o}.timing \
                                        $new_base_path/${vo%%o}.timing > \
                                        $working_dir/html/$coq_opam_package/${vo%%o}.html ||
-                echo "Failed (code $?):" $program_path/timelog2html $new_base_path/${vo%%o} \
+                echo "Failed (code $?):" $timelog2html $new_base_path/${vo%%o} \
                      $old_base_path/${vo%%o}.timing \
                      $new_base_path/${vo%%o}.timing
         fi
@@ -535,6 +630,23 @@ du -ha "$working_dir" > "$working_dir/files.listing"
 #
 # The next script processes all these files and prints results in a table.
 
+# timings data
+timings=$working_dir/timings
+mkdir -p $timings
+
+# Print line by line slow downs and speed ups
+if [ -d "$working_dir/html" ]; then # might not exist if all jobs failed
+cd "$working_dir/html"
+$render_line_results
+# Move line timing files to timings folder (they will become artifacts)
+mv fast_table slow_table timings_table $timings
+
+# html tables don't get generated if the bench is run locally ie without CI variables
+for f in fast_table.html slow_table.html timings_table.html; do
+    if [ -f "$f" ]; then mv "$f" $timings; fi
+done
+fi
+
 echo "INFO: workspace = ${CI_JOB_URL}/artifacts/browse/${bench_dirname}"
 
 # Print the final results.
@@ -546,9 +658,10 @@ if [ -z "$installable_coq_opam_packages" ]; then
     exit 1
 fi
 
-echo "DEBUG: $program_path/render_results $log_dir $num_of_iterations $new_coq_commit_long $old_coq_commit_long 0 user_time_pdiff $installable_coq_opam_packages"
-rendered_results="$($program_path/render_results "$log_dir" $num_of_iterations $new_coq_commit_long $old_coq_commit_long 0 user_time_pdiff $installable_coq_opam_packages)"
+echo "DEBUG: $render_results $log_dir $num_of_iterations $new_coq_commit_long $old_coq_commit_long 0 user_time_pdiff $installable_coq_opam_packages"
+rendered_results="$($render_results "$log_dir" $num_of_iterations $new_coq_commit_long $old_coq_commit_long 0 user_time_pdiff $installable_coq_opam_packages)"
 echo "${rendered_results}"
+echo "${rendered_results}" > $timings/bench_summary
 
 echo "INFO: per line timing: ${CI_JOB_URL}/artifacts/browse/${bench_dirname}/html/"
 
@@ -558,14 +671,23 @@ git log -n 1 "$old_coq_commit"
 echo INFO: New Coq version
 git log -n 1 "$new_coq_commit"
 
-not_installable_coq_opam_packages=$(comm -23 <(echo $sorted_coq_opam_packages | sed 's/ /\n/g' | sort | uniq) <(echo $installable_coq_opam_packages | sed 's/ /\n/g' | sort | uniq) | sed 's/\t//g')
+if [ -n "$failed_packages" ]; then
+    not_installable_coq_opam_packages=$failed_packages
+    if [ -n "$skipped_packages" ]; then
+        not_installable_coq_opam_packages="$not_installable_coq_opam_packages
+$skipped_packages"
+    fi
+else # in case the failed package detection is bugged
+    not_installable_coq_opam_packages=$(comm -23 <(echo $sorted_coq_opam_packages | sed 's/ /\n/g' | sort | uniq) <(echo $installable_coq_opam_packages | sed 's/ /\n/g' | sort | uniq) | sed 's/\t//g')
+fi
 
 coqbot_update_comment "done" "${rendered_results}" "${not_installable_coq_opam_packages}"
 
+touch $timings/bench_failures
 if [ -n "$not_installable_coq_opam_packages" ]; then
     # Tell the user that some of the provided OPAM-package(s)
     # is/are not installable.
-    printf '\n\nINFO: failed to install %s\n' "$not_installable_coq_opam_packages"
+    printf '\n\nINFO: failed to install %s\n' "$not_installable_coq_opam_packages" | tee $timings/bench_failures
     zulip_edit "Bench complete, failed to install packages:
 $not_installable_coq_opam_packages"
     exit 1
