@@ -67,29 +67,51 @@ let read_timing_lines file =
   in
   close_in ic; lines
 
-let list_timing_data (time1, time2, diff, pdiff, line_num, file) =
+type html_data = { link_prefix : string }
+
+let get_html_data () =
+  match Sys.getenv_opt "CI_PROJECT_NAMESPACE",
+        Sys.getenv_opt "CI_PROJECT_NAME",
+        Sys.getenv_opt "CI_JOB_ID" with
+  | Some ns, Some project, Some id ->
+    Some { link_prefix =
+             Printf.sprintf "https://%s.gitlab.io/-/%s/-/jobs/%s/artifacts/_bench/html/"
+               ns project id }
+  | None, _, _ | _, None, _ | _, _, None -> None
+
+let html_str ?html lnum s = match html with
+  | None -> Table.raw_str s
+  | Some html ->
+    let size = String.length s in
+    let s = Printf.sprintf "<a href=\"%s%s#L%d\">%s</a>" html.link_prefix s lnum s in
+    { Table.str = s; size }
+
+let list_timing_data ?html (time1, time2, diff, pdiff, line_num, file) =
   let str_time1 = Printf.sprintf "%.4f" time1 in
   let str_time2 = Printf.sprintf "%.4f" time2 in
   let str_diff  = Printf.sprintf "%.4f" diff in
   let str_pdiff = Printf.sprintf "%3.2f%%" pdiff in
   let str_line_num = string_of_int line_num in
-  [ str_time1; str_time2; str_diff; str_pdiff; str_line_num; file ]
+  List.append
+    (List.map Table.raw_str [ str_time1; str_time2; str_diff; str_pdiff; str_line_num])
+    [ html_str ?html line_num file ]
 
 let render_table ?(reverse=false) title num table =
   let open Table.Align in
-  let headers = [title] in
-  let top = [["OLD"; "NEW"; "DIFF"; "%DIFF"; "Ln"; "FILE"]] in
+  let headers = [Table.raw_str title] in
+  let top = Table.raw_row [["OLD"; "NEW"; "DIFF"; "%DIFF"; "Ln"; "FILE"]] in
   let align_top = [[Middle; Middle; Middle; Middle; Middle; MidLeft]] in
   let align_rows = [[Right; Right; Right; Right; Right; Left]] in
   (if reverse then CList.rev table else table)
-  |> Array.of_list
-  |> fun x -> Array.sub x 0 num
-  |> Array.to_list
+  |> CList.firstn num
   |> fun x -> Table.print headers top x ~align_top ~align_rows ()
 
+let to_file fname fmt =
+  Printf.kfprintf close_out (open_out fname) fmt
+
 let main () =
   let () = Printexc.record_backtrace true in
-  let table =
+  let data =
     Unix.getcwd ()
     |> list_html_files
     |> CList.filter_map read_timing_lines
@@ -97,6 +119,9 @@ let main () =
     (* Do we want to do a unique sort? *)
     (* |> CList.sort_uniq (fun (_,_,x,_,_,_) (_,_,y,_,_,_) -> Float.compare x y) *)
     |> CList.sort (fun (_,_,x,_,_,_) (_,_,y,_,_,_) -> Float.compare x y)
+  in
+  let table =
+    data
     |> CList.map list_timing_data
     |> CList.map (fun x -> [ x ])
   in
@@ -108,18 +133,26 @@ let main () =
   let timings_table = render_table "Significant line time changes in bench" (CList.length table) table in
   (* Print tables to stdout *)
   Printf.printf "%s\n%s\n" slow_table fast_table;
-  (* Print slow table to slow_table file *)
-  let oc_slow = open_out "slow_table" in
-  Printf.fprintf oc_slow "%s\n" slow_table;
-  close_out oc_slow;
-  (* Print fast table to fast_table file *)
-  let oc_fast = open_out "fast_table" in
-  Printf.fprintf oc_fast "%s\n" fast_table;
-  close_out oc_fast;
-  (* Print all timing data to line_timings file *)
-  let oc_timings = open_out "timings_table" in
-  Printf.fprintf oc_timings "%s\n" timings_table;
-  close_out oc_timings;
-  ()
+  (* Print tables to files *)
+  to_file "slow_table" "%s\n" slow_table;
+  to_file "fast_table" "%s\n" fast_table;
+  to_file "timings_table" "%s\n" timings_table;
+  (* html tables *)
+  match get_html_data () with
+  | None -> ()
+  | Some html ->
+    let table =
+      data
+      |> CList.map (list_timing_data ~html)
+      |> CList.map (fun x -> [ x ])
+    in
+    let slow_table = render_table (Printf.sprintf "TOP %d SLOW DOWNS" num) ~reverse:true num table in
+    let fast_table = render_table (Printf.sprintf "TOP %d SPEED UPS" num) num table in
+    let timings_table = render_table "Significant line time changes in bench" (CList.length table) table in
+    to_file "slow_table.html" "<pre>%s</pre>\n" slow_table;
+    to_file "fast_table.html" "<pre>%s</pre>\n" fast_table;
+    to_file "timings_table.html" "<pre>%s</pre>\n" timings_table;
+    ()
+
 
 let () = main ()
