@@ -42,32 +42,48 @@ function project {
 ########################################################################
 # MathComp
 ########################################################################
-project mathcomp 'https://github.com/math-comp/math-comp' '10816c7d830a568d732c749831466ba8830ffe49'
+project mathcomp 'https://github.com/math-comp/math-comp' '6fcb8c35e9bd7ed0d74d6870893b2d3a69cbe67c'
 
-project fourcolor 'https://github.com/math-comp/fourcolor' '4581901330cce906c3c58ce5c24e5d061ea3ab39'
+project fourcolor "https://github.com/math-comp/fourcolor" "e831b0b00e264285f91938917a0a5ef64ec1a829"
+# put back master when testing MathComp 2
+# project fourcolor "https://github.com/math-comp/fourcolor" "master"
 
-project oddorder 'https://github.com/math-comp/odd-order' '91e3b9a75d12d142568985566e652988189a2ff4'
+project oddorder "https://github.com/math-comp/odd-order" "2c03794e64eef467442a4ea2ef1430b13b0faa97"
+# put back master when testing MathComp 2
+# project oddorder "https://github.com/math-comp/odd-order" "master"
 
-project mczify 'https://github.com/math-comp/mczify' 'a1cca0ee6aba5d0a932d43888cc814cdf620784d'
+project mczify "https://github.com/math-comp/mczify" "2046446984f7b8c8f5102df4df6076b60874e688"
+# put back master when testing MathComp 2
+# project mczify "https://github.com/math-comp/mczify" "master"
+
+project finmap "https://github.com/math-comp/finmap" "cea9f088c9cddea1173bc2f7c4c7ebda35081b60"
+# put back master when testing MathComp 2
+# project finmap "https://github.com/math-comp/finmap" "master"
+
+project bigenough 'https://github.com/math-comp/bigenough' 'ff71b25f31658d80fdae9657e8cf34e5e1052647'
+
+project analysis "https://github.com/math-comp/analysis" "9193f4a1278409cc13a1de739adf3620aa24a638"
+# put back master when testing MathComp 2
+# project analysis "https://github.com/math-comp/analysis" "master"
 
 ########################################################################
 # UniMath
 ########################################################################
-project unimath 'https://github.com/UniMath/UniMath' 'd0e7d748fb5cee00472c08abe1835022910ec7ff'
+project unimath 'https://github.com/UniMath/UniMath' '18f15a8eb58f20680672b074354c1b1b82aec4dc'
 
 ########################################################################
 # Unicoq + Mtac2
 ########################################################################
-project unicoq 'https://github.com/unicoq/unicoq' '5ed1cdf38a7ccc4d5f3f4d83c7855c41e0f855e8'
+project unicoq 'https://github.com/unicoq/unicoq' 'd9b0e1f5b1b51f06a8e2f21910c3c117e9233d6a'
 
-project mtac2 'https://github.com/Mtac2/Mtac2' 'c27161ba0e9c750976f343cae3245708580c8079'
+project mtac2 'https://github.com/Mtac2/Mtac2' '29dc0ae56db985ba3d052e59c84a7d67471d6e8f'
 
 ########################################################################
 # Mathclasses + Corn
 ########################################################################
-project math_classes 'https://github.com/coq-community/math-classes' 'b94b24aa387abb90740acf7cc24286bf27d0fb50'
+project math_classes 'https://github.com/coq-community/math-classes' '7a57556e489989cb38899decf61ea66f4394c60f'
 
-project corn 'https://github.com/coq-community/corn' 'c366d3f01ec1812b145117a4da940518b092d3a6'
+project corn 'https://github.com/coq-community/corn' 'c08a0418f97a04ea7a6cdc3a930561cc8fc84d82'
 
 ########################################################################
 # Iris
@@ -79,238 +95,297 @@ project stdpp "https://gitlab.mpi-sws.org/iris/stdpp" ""
 
 project iris "https://gitlab.mpi-sws.org/iris/iris" ""
 
-project autosubst 'https://github.com/coq-community/autosubst' 'd1a90313f321a06ca96b7f453b3de8296dc914dc'
+project autosubst 'https://github.com/coq-community/autosubst' '495d547fa2a3e40da0f7d2a31469ac8caab058d2'
 
-project iris_examples 'https://gitlab.mpi-sws.org/iris/examples' 'ee6a36b3e803ebec465bf44bb5bcbf9b19fe6b62'
+project iris_examples 'https://gitlab.mpi-sws.org/iris/examples' 'f899efbd7c49ed4ead30ac67b1595602f0b091ec'
 
 ########################################################################
 # HoTT
 ########################################################################
-project hott 'https://github.com/HoTT/HoTT' '8f17cbee7274578a27606953b6da576b22c1ea0d'
+project hott 'https://github.com/HoTT/HoTT' '832aef3e6fff0f5b953ed170522e1a3d6288a4bb'
 
 ########################################################################
 # CoqHammer
 ########################################################################
-project coqhammer 'https://github.com/lukaszcz/coqhammer' '1079879790f671389e282a3606c0eda996b6c16a'
+project coqhammer 'https://github.com/lukaszcz/coqhammer' '35b1ddb346ff4739d4b7d0aaa2ee34ebb1de22ae'
 
 ########################################################################
 # GeoCoq
 ########################################################################
-project geocoq 'https://github.com/GeoCoq/GeoCoq' '1786a9a33dbca21e466e3ba57afaaa78876d1721'
+project geocoq 'https://github.com/GeoCoq/GeoCoq' 'b0f922bdf06bc175155a0a489a451fb73c1e403a'
 
 ########################################################################
 # Flocq
 ########################################################################
-project flocq 'https://gitlab.inria.fr/flocq/flocq' '239727fc0669d76ee60a45efb176bda2fe261f58'
+project flocq 'https://gitlab.inria.fr/flocq/flocq' '2c499dc6fd577ee13ac84ee6a0bad5024055aa29'
 
 ########################################################################
 # coq-performance-tests
 ########################################################################
-project coq_performance_tests 'https://github.com/coq-community/coq-performance-tests' 'ae8385b9471409387d0f47f01e38b866ba70bda1'
+project coq_performance_tests 'https://github.com/coq-community/coq-performance-tests' '4aaef74e5742fe3ad87c5d18d735e82b1284ecc6'
 
 ########################################################################
 # coq-tools
 ########################################################################
-project coq_tools 'https://github.com/JasonGross/coq-tools' '366c217bf1d7b158d0c1721d7e8f254f025f89a4'
+project coq_tools 'https://github.com/JasonGross/coq-tools' '7562bafc9b21cb34d74f869632455e67d11f5bc3'
 
 ########################################################################
 # Coquelicot
 ########################################################################
-project coquelicot 'https://gitlab.inria.fr/coquelicot/coquelicot' '09eb630f233c6887bef8f6764c7145bc45f24951'
-
-########################################################################
-# Gappa stand alone tool
-########################################################################
-project gappa_tool 'https://gitlab.inria.fr/gappa/gappa' '6c97a36257369d89ff32c9877c0e83681bfd3df9'
-
-########################################################################
-# Gappa plugin
-########################################################################
-project gappa 'https://gitlab.inria.fr/gappa/coq' 'f32711d1395a590492d9419a59ce0f9707765c4b'
+project coquelicot 'https://gitlab.inria.fr/coquelicot/coquelicot' 'df29052d3607adaf9c7d1e663a13c36149035e8f'
 
 ########################################################################
 # CompCert
 ########################################################################
-project compcert 'https://github.com/AbsInt/CompCert' 'a0ad5ff6f9c7603610a7448935b36c9ed22c6435'
+project compcert 'https://github.com/AbsInt/CompCert' '222bd6d812b726be939eb8afc67b935a08799bfa'
 
 ########################################################################
 # VST
 ########################################################################
-# todo: 2021 03 11: switch back to master once vst merges the compcert3.9 branch
-project vst 'https://github.com/PrincetonUniversity/VST' '0da1e909bdbe458af24148c5d681be1be09942e2'
+project vst 'https://github.com/PrincetonUniversity/VST' 'cd6f85a813069ab7742d3f6e1510960230a86e2e'
 
 ########################################################################
 # cross-crypto
 ########################################################################
-project cross_crypto 'https://github.com/mit-plv/cross-crypto' 'c7cc46de53cf6b2363631a3b17b2ab12ce42e2f8'
+project cross_crypto 'https://github.com/mit-plv/cross-crypto' '964dc89064da3232ffb202474bf9b775c581ee5c'
 
 ########################################################################
 # rewriter
 ########################################################################
-project rewriter 'https://github.com/mit-plv/rewriter' '7553a6d910b7846e778fd648adc14f7395d2b3c9'
+project rewriter 'https://github.com/mit-plv/rewriter' 'c79bbc9148fb8386e188267c684a882d8634293b'
 
 ########################################################################
 # fiat_parsers
 ########################################################################
-project fiat_parsers 'https://github.com/mit-plv/fiat' '490532505f20d3c93e0936f5770d2358c6effaaf'
+project fiat_parsers 'https://github.com/mit-plv/fiat' '2bb593d87a7005cca74dd5e6810d311f2872b4b4'
 
 ########################################################################
 # fiat_crypto
 ########################################################################
-project fiat_crypto 'https://github.com/mit-plv/fiat-crypto' 'f549083118b817dca59d435c18b83179aa06bd9a'
+project fiat_crypto 'https://github.com/mit-plv/fiat-crypto' 'c371fb8945cf311aefc598bd7ae3ef0d4123b91e'
 
 ########################################################################
 # fiat_crypto_legacy
 ########################################################################
-project fiat_crypto_legacy 'https://github.com/mit-plv/fiat-crypto' '093ea19d246a3a79d019b5a78606559b261cf7c1'
+project fiat_crypto_legacy 'https://github.com/mit-plv/fiat-crypto' '20ad035d5212531f5efb82f9fff32f26b80c39e6'
 
 ########################################################################
 # coq_dpdgraph
 ########################################################################
-project coq_dpdgraph 'https://github.com/Karmaki/coq-dpdgraph' '3f8a02e32487b66d168a59f5c24450531ee3a632'
+project coq_dpdgraph 'https://github.com/Karmaki/coq-dpdgraph' '8b4291573cc12a03d7e703e00dd82e5e80e5d446'
 
 ########################################################################
 # CoLoR
 ########################################################################
-project color 'https://github.com/fblanqui/color' '1e7b26553c1ca94c787ad5a1b938acabb8d47f2f'
+project color 'https://github.com/fblanqui/color' 'f2ef98f7d13c5d71dd2a614ed2e6721703a34532'
 
 ########################################################################
 # TLC
 ########################################################################
-project tlc 'https://github.com/charguer/tlc' 'f5390a5ec9106b15b8f8a4434958d4959b15a295'
+project tlc 'https://github.com/charguer/tlc' '198142f9db4427063c381450ad6a25839d7fd35c'
 
 ########################################################################
 # Bignums
 ########################################################################
-project bignums 'https://github.com/coq/bignums' 'a419f649a7a34e28049f3f18a5ee4068c2cd7fff'
+project bignums 'https://github.com/coq/bignums' 'a8c50a569a971d6e36e8df88ca16f9f8958f9543'
 
 ########################################################################
 # coqprime
 ########################################################################
-project coqprime 'https://github.com/thery/coqprime' 'd4bcfa8f7d7bdbe879d842ecfe575406f622c8ae'
+project coqprime 'https://github.com/thery/coqprime' '431d7a66877cbe8688fc8864ef892369401440e1'
 
 ########################################################################
 # bbv
 ########################################################################
-project bbv 'https://github.com/mit-plv/bbv' '1386458cdf0b87bd9deaa2cdc47d3d732f96f5b6'
+project bbv 'https://github.com/mit-plv/bbv' '6144e214583b72182a4ed89e4328ea62ba766e2e'
 
 ########################################################################
 # bedrock2
 ########################################################################
-project bedrock2 'https://github.com/mit-plv/bedrock2' '76ad54ba0704cc589d7c121b0e16b53ad3ea8b75'
+project bedrock2 'https://github.com/mit-plv/bedrock2' '0893c1f7d9ac2ef60c9b62cbbf54ec797d98728b'
+
+########################################################################
+# coq-lsp
+########################################################################
+project coq_lsp 'https://github.com/ejgallego/coq-lsp' 'c1faca2ed1d4c1432bad43be05e7d03176c246dc'
 
 ########################################################################
 # Equations
 ########################################################################
-project equations 'https://github.com/mattam82/Coq-Equations' '3b48ad15c64fe1f3ec81978679960f265a7877b5'
+project equations 'https://github.com/mattam82/Coq-Equations' '21a3306a7e13b3037d20492bfdb5da75463929fb'
 
 ########################################################################
 # Elpi + Hierarchy Builder
 ########################################################################
-project elpi 'https://github.com/LPCIC/coq-elpi' '60db122c8009b04f1bc44c2466a737e55ebe4b6a'
+project elpi 'https://github.com/LPCIC/coq-elpi' 'd39c6ca16ac987490eedefdcddd09ca25202e862'
 
-project hierarchy_builder 'https://github.com/math-comp/hierarchy-builder' 'bd589b63bf48cf9a00812bdaa0c9a8192e78b83b'
+project hierarchy_builder 'https://github.com/math-comp/hierarchy-builder' 'c8e542e7994f033151a47e77306e2a52929c9a2c'
 
 ########################################################################
 # Engine-Bench
 ########################################################################
-project engine_bench 'https://github.com/mit-plv/engine-bench' '45f456e4c3b37704c6ff83ef072dc1260fef8f43'
+project engine_bench 'https://github.com/mit-plv/engine-bench' '8c0b15abc38b1d3f7dc606f4eeef9fba1a986b05'
 
 ########################################################################
 # fcsl-pcm
 ########################################################################
-project fcsl_pcm 'https://github.com/imdea-software/fcsl-pcm' 'cbd73130489169d1bb236a0888fceacaa77f0d70'
+project fcsl_pcm 'https://github.com/imdea-software/fcsl-pcm' '35cb0a2235adb656c2696e33512f0c4555ad753b'
 
 ########################################################################
 # ext-lib
 ########################################################################
-project ext_lib 'https://github.com/coq-community/coq-ext-lib' '968fb49a1f1044adf87f8e427cbe2d83d525d37d'
+project ext_lib 'https://github.com/coq-community/coq-ext-lib' '995151f6df3768625922c3b2be5d9f4801b3dd3f'
 
 ########################################################################
 # simple-io
 ########################################################################
-project simple_io 'https://github.com/Lysxia/coq-simple-io' 'f7536c629dd02883abb5a2d64865becafed3af5b'
+project simple_io 'https://github.com/Lysxia/coq-simple-io' '403db47e0b96222cd3c89c48ed5f8da10319da77'
 
 ########################################################################
 # quickchick
 ########################################################################
-project quickchick 'https://github.com/QuickChick/QuickChick' '80f4c88e949f25b6167993ce6a7e34ed29ca44cb'
+project quickchick 'https://github.com/QuickChick/QuickChick' '6cf6918a5add0e10512f4529719f8fd8a4b605fa'
 
 ########################################################################
 # reduction-effects
 ########################################################################
-project reduction_effects 'https://github.com/coq-community/reduction-effects' '7cf41546f2d15ea3a4eeeb68695b288db2d9e09a'
+project reduction_effects 'https://github.com/coq-community/reduction-effects' '604528fa0e8472cd1d1264782ffcbfecb8571b48'
 
 ########################################################################
 # menhirlib
 ########################################################################
 # Note: menhirlib is now in subfolder coq-menhirlib of menhir
-project menhirlib "https://gitlab.inria.fr/fpottier/menhir" "20210310"
+project menhirlib "https://gitlab.inria.fr/fpottier/menhir" "20220210"
 
 ########################################################################
 # aac_tactics
 ########################################################################
-project aac_tactics 'https://github.com/coq-community/aac-tactics' '16d30aba89ec4f5a7663ff31ac1b81212dd22e9a'
+project aac_tactics 'https://github.com/coq-community/aac-tactics' '0f2de225cb6d42d93c8aa61891510e5c6ac6726c'
+
+########################################################################
+# paco
+########################################################################
+project paco 'https://github.com/snu-sf/paco' '5c5693f46c8957f36a2349a0d906e911366136de'
+
+########################################################################
+# coq-itree
+########################################################################
+project itree 'https://github.com/DeepSpec/InteractionTrees' 'eede8c539985e358e7149ab77b5c058def37571d'
 
 ########################################################################
 # paramcoq
 ########################################################################
-project paramcoq 'https://github.com/coq-community/paramcoq' '1d9409b9eb56cb3f60e176e1f9194d6c4e09afef'
+project paramcoq 'https://github.com/coq-community/paramcoq' '9d7f66fffbaf44e034f35c4ada50dac14824fe0b'
 
 ########################################################################
 # relation_algebra
 ########################################################################
-project relation_algebra 'https://github.com/damien-pous/relation-algebra' '7c37b9284848f39091770a60920fcc90a0019b47'
+project relation_algebra 'https://github.com/damien-pous/relation-algebra' '1c37f34e53a0a0c523fbd813f9ae10e18f6e41f2'
 
 ########################################################################
 # StructTact + InfSeqExt + Cheerios + Verdi + Verdi Raft
 ########################################################################
-project struct_tact 'https://github.com/uwplse/StructTact' '179bd5312e9d8b63fc3f4071c628cddfc496d741'
+project struct_tact 'https://github.com/uwplse/StructTact' '2f2ff253be29bb09f36cab96d036419b18a95b00'
 
-project inf_seq_ext 'https://github.com/DistributedComponents/InfSeqExt' 'd7a56b2ba532ac0a78a3e3ad0080d223186838de'
+project inf_seq_ext 'https://github.com/DistributedComponents/InfSeqExt' '601e89ec019501c48c27fcfc14b9a3c70456e408'
 
-project cheerios 'https://github.com/uwplse/cheerios' '9c7f66e57b91f706d70afa8ed99d64ed98ab367d'
+project cheerios 'https://github.com/uwplse/cheerios' 'bad8ad2476e14df6b5a819b7aaddc27a7c53fb69'
 
-project verdi 'https://github.com/uwplse/verdi' '54597d8ac7ab7dd4dae683f651237644bf77701e'
+project verdi 'https://github.com/uwplse/verdi' '76833a7b2188e99e358b8439ea6b4f38401c962a'
 
-project verdi_raft 'https://github.com/uwplse/verdi-raft' 'dbf3084f8a741d56b5d05a87e39705329753bb6e'
+project verdi_raft 'https://github.com/uwplse/verdi-raft' '7c8e4d53d27f7264ec4d3de72944dc0368e065f0'
 
 ########################################################################
 # stdlib2
 ########################################################################
-project stdlib2 'https://github.com/coq/stdlib2' '54e057ea7023d058e57169a3bceaab708a5f7d26'
+project stdlib2 'https://github.com/coq/stdlib2' '9dfc4b760ba4ec2ed08732d1fccc904b1f9189dc'
 
 ########################################################################
 # argosy
 ########################################################################
-project argosy 'https://github.com/mit-pdos/argosy' '63dd6e6d046e3dde072ae1b6862639a8f62b1213'
+project argosy 'https://github.com/mit-pdos/argosy' 'a6a5aa0d3868efd4ada0b40927b5748e5d8967d3'
 
 ########################################################################
 # perennial
 ########################################################################
-project perennial 'https://github.com/mit-pdos/perennial' '8d8e50d7c2cdf587bded3a3dbca9af26333dcf5f'
+project perennial 'https://github.com/mit-pdos/perennial' '42d4197a1990b27c9f9abc533a5cde9f5813115b'
 
 ########################################################################
 # metacoq
 ########################################################################
-project metacoq 'https://github.com/MetaCoq/metacoq' '52950970c23eec013e91346e9237e693b6debec3'
+project metacoq 'https://github.com/MetaCoq/metacoq' 'b34d3f32a8d73e4fea721df2cddb6be42509b3f0'
 
 ########################################################################
 # SF suite
 ########################################################################
-project sf 'https://github.com/DeepSpec/sf' '41c4ddd76dbde6a7aede634d7b003a64b5b11e98'
+project sf 'https://github.com/DeepSpec/sf' 'd98b8968483d7ee30dc3a502cd7d58f1490bfa9b'
 
 ########################################################################
 # Coqtail
 ########################################################################
-project coqtail 'https://github.com/whonore/Coqtail' 'aa817559be68d9e90316bc7e5d3205fe2ffcffbe'
+project coqtail 'https://github.com/whonore/Coqtail' 'ec80f3d48dcbf19209ef51d6020838cda5a1d46e'
 
 ########################################################################
 # Deriving
 ########################################################################
-project deriving 'https://github.com/arthuraa/deriving' 'fa62caa5bd55431d42c3d6ed88bfc00899413dce'
+project deriving 'https://github.com/arthuraa/deriving' 'bf967106a0b61a281c3b4ababfe67f80de7fba47'
 
 ########################################################################
 # VsCoq
 ########################################################################
-project vscoq 'https://github.com/maximedenes/vscoq' 'f6f61c52d7e01b18dd4fea19604ee0a2bf1be22c'
+project vscoq 'https://github.com/coq-community/vscoq' 'ed40eda53d69dd17c14dce1542c83ddc81cab3e9'
+
+########################################################################
+# category-theory
+########################################################################
+project category_theory 'https://github.com/jwiegley/category-theory' '5376e32a4eeace4a84674820083bc2985a2a593f'
+
+########################################################################
+# itauto
+########################################################################
+project itauto 'https://gitlab.inria.fr/fbesson/itauto' '04b02439e1a7383e8834caacb1a5031d873c1de4'
+
+########################################################################
+# Mathcomp-word
+########################################################################
+project mathcomp_word 'https://github.com/jasmin-lang/coqword' 'cb8b06aa14ab13b07e04011ae376b3a133cc911d'
+
+########################################################################
+# Jasmin
+########################################################################
+project jasmin 'https://github.com/jasmin-lang/jasmin' 'c8f25be1829036d23d486a8e9210f990a9a0c298'
+
+########################################################################
+# Lean Importer
+########################################################################
+project lean_importer 'https://github.com/SkySkimmer/coq-lean-import' 'c7dd746178afe675463a32d6550fed29fb011640'
+
+########################################################################
+# SerAPI
+########################################################################
+project serapi 'https://github.com/ejgallego/coq-serapi' '70ddb488161ae3447c91d811192101ac3cf97abb'
+
+########################################################################
+# SMTCoq
+########################################################################
+project smtcoq 'https://github.com/smtcoq/smtcoq' '4a29a8cdb77ec38a77e42c2dd3ef366ecda8834e'
+
+########################################################################
+# Stalmarck
+########################################################################
+project stalmarck 'https://github.com/coq-community/stalmarck' '9e6cd57df21f991ca5cdd54800707b96fba16ced'
+
+########################################################################
+# coq-library-undecidability
+########################################################################
+project coq_library_undecidability 'https://github.com/uds-psl/coq-library-undecidability' '0e29426de0e6749b851514968f3982db2e9e8ccb'
+
+########################################################################
+# Tactician
+########################################################################
+project tactician 'https://github.com/coq-tactician/coq-tactician' 'd01f50ae2e9eff55e10c6c2e6ce61df9970f04b6'
+
+########################################################################
+# Ltac2 compiler
+########################################################################
+project ltac2_compiler 'https://github.com/SkySkimmer/coq-ltac2-compiler' '064771bd53e1fd8f61fd27ae16ab84f262d37177'
