@@ -5,12 +5,13 @@ set -e
 ci_dir="$(dirname "$0")"
 . "${ci_dir}/ci-common.sh"
 
-FORCE_GIT=1
+WITH_SUBMODULES=1
 git_download bedrock2
 
+if [ "$DOWNLOAD_ONLY" ]; then exit 0; fi
+
 export COQEXTRAFLAGS='-native-compiler no'
 ( cd "${CI_BUILD_DIR}/bedrock2"
-  git submodule update --init --recursive
   COQMF_ARGS='-arg "-async-proofs-tac-j 1"' make
   make install
 )
