@@ -7,9 +7,11 @@ ci_dir="$(dirname "$0")"
 
 git_download equations
 
+if [ "$DOWNLOAD_ONLY" ]; then exit 0; fi
+
 export COQEXTRAFLAGS='-native-compiler no'
 ( cd "${CI_BUILD_DIR}/equations"
-  ./configure.sh coq
-  make ci
+  [ -e Makefile.coq ] || ./configure.sh coq
+  make
   make install
 )
