@@ -5,9 +5,11 @@ set -e
 ci_dir="$(dirname "$0")"
 . "${ci_dir}/ci-common.sh"
 
-FORCE_GIT=1
+WITH_SUBMODULES=1
 git_download fiat_crypto
 
+if [ "$DOWNLOAD_ONLY" ]; then exit 0; fi
+
 # We need a larger stack size to not overflow ocamlopt+flambda when
 # building the executables.
 # c.f. https://github.com/coq/coq/pull/8313#issuecomment-416650241
@@ -18,10 +20,10 @@ stacksize=32768
 # version of other developments
 make_args=(EXTERNAL_REWRITER=1 EXTERNAL_COQPRIME=1)
 
-export COQEXTRAFLAGS='-native-compiler no'
 ( cd "${CI_BUILD_DIR}/fiat_crypto"
-  git submodule update --init --recursive
   ulimit -s $stacksize
-  make "${make_args[@]}" pre-standalone-extracted printlite lite
-  make -j 1 "${make_args[@]}" all-except-compiled
+  make "${make_args[@]}" pre-standalone-extracted printlite lite ||
+    make -j 1 "${make_args[@]}" pre-standalone-extracted printlite lite
+  make "${make_args[@]}" all-except-compiled ||
+    make -j 1 "${make_args[@]}" all-except-compiled
 )
