@@ -6,40 +6,33 @@ ci_dir="$(dirname "$0")"
 . "${ci_dir}/ci-common.sh"
 
 git_download struct_tact
-
-( cd "${CI_BUILD_DIR}/struct_tact"
-  ./configure
-  make
-  make install
-)
-
 git_download inf_seq_ext
-
-( cd "${CI_BUILD_DIR}/inf_seq_ext"
-  ./configure
-  make
-  make install
-)
-
 git_download cheerios
-
-( cd "${CI_BUILD_DIR}/cheerios"
-  ./configure
-  make
-  make install
-)
-
 git_download verdi
-
-( cd "${CI_BUILD_DIR}/verdi"
-  ./configure
-  make
-  make install
-)
-
 git_download verdi_raft
 
+if [ "$DOWNLOAD_ONLY" ]; then exit 0; fi
+
+( cd "${CI_BUILD_DIR}/struct_tact"
+  make
+  make install
+)
+
+( cd "${CI_BUILD_DIR}/inf_seq_ext"
+  make
+  make install
+)
+
+( cd "${CI_BUILD_DIR}/cheerios"
+  make
+  make install
+)
+
+( cd "${CI_BUILD_DIR}/verdi"
+  make
+  make install
+)
+
 ( cd "${CI_BUILD_DIR}/verdi_raft"
-  ./configure
   make
 )
