@@ -23,9 +23,11 @@ open Context
 open Genarg
 open Clenv
 
-let _ = Detyping.print_evar_arguments := true
-let _ = Detyping.print_universes := true
-let _ = Goptions.set_bool_option_value ["Printing";"Matching"] false
+let () = Flags.in_debugger := true
+let () = Goptions.set_bool_option_value ["Printing";"Existential";"Instances"] true
+let () = Detyping.print_universes := true
+let () = Goptions.set_bool_option_value ["Printing";"Matching"] false
+let () = Goptions.set_bool_option_value ["Printing";"Sort";"Qualities"] true
 
 let with_env_evm f x =
   let env = Global.env() in
@@ -46,6 +48,7 @@ let ppmbid mbid = pp (str (MBId.debug_to_string mbid))
 let ppdir dir = pp (DirPath.print dir)
 let ppmp mp = pp(str (ModPath.debug_to_string mp))
 let ppcon con = pp(Constant.debug_print con)
+let ppprojrepr con = pp(Constant.debug_print (Projection.Repr.constant con))
 let ppproj con = pp(Constant.debug_print (Projection.constant con))
 let ppkn kn = pp(str (KerName.to_string kn))
 let ppmind kn = pp(MutInd.debug_print kn)
@@ -83,6 +86,16 @@ let pppattern = (fun x -> pp(envpp pr_constr_pattern_env x))
 let pptype = (fun x -> try pp(envpp (fun env evm t -> pr_ltype_env env evm t) x) with e -> pp (str (Printexc.to_string e)))
 let ppfconstr c = ppconstr (CClosure.term_of_fconstr c)
 
+let ppuint63 i = pp (str (Uint63.to_string i))
+
+let pp_parray pr a =
+  let a, def = Parray.to_array a in
+  let a = Array.to_list a in
+  pp (str "[|" ++ prlist_with_sep (fun () -> str ";" ++ spc()) pr a ++ spc() ++ str "|" ++ spc() ++ pr def ++ str "|]")
+
+let pp_constr_parray = pp_parray pr_constr
+let pp_fconstr_parray = pp_parray (fun f -> pr_constr (CClosure.term_of_fconstr f))
+
 let ppfsubst s =
   let (s, k) = Esubst.Internal.repr s in
   let sep () = str ";" ++ spc () in
@@ -113,6 +126,28 @@ let pridmapgen l = prmapgen Id.print (Id.Set.elements (Id.Map.domain l))
 let ppidmapgen l = pp (pridmapgen l)
 let printmapgen l = prmapgen int (Int.Set.elements (Int.Map.domain l))
 let ppintmapgen l = pp (printmapgen l)
+let prmodidmapgen l = prmapgen Id.print (ModIdset.elements (ModIdmap.domain l))
+let ppmodidmapgen l = pp (prmodidmapgen l)
+
+let ppmpmapgen l =
+  pp (prmapgen
+        (fun mp -> str (ModPath.debug_to_string mp))
+        (MPset.elements (MPmap.domain l)))
+
+let ppdpmapgen l =
+  pp (prmapgen
+        (fun mp -> str (DirPath.to_string mp))
+        (DPset.elements (DPmap.domain l)))
+
+let ppconmapenvgen l =
+  pp (prmapgen
+        (fun mp -> str (Constant.debug_to_string mp))
+        (Cset_env.elements (Cmap_env.domain l)))
+
+let ppmindmapenvgen l =
+  pp (prmapgen
+        (fun mp -> str (MutInd.debug_to_string mp))
+        (Mindmap_env.Set.elements (Mindmap_env.domain l)))
 
 let ppevarsubst = ppidmap (fun id0 -> prset (fun (c,copt,id) ->
   hov 0
@@ -191,14 +226,13 @@ let ppexistentialset evars =
 let ppexistentialfilter filter = match Evd.Filter.repr filter with
 | None -> pp (Pp.str "ø")
 | Some f -> pp (prlist_with_sep spc bool f)
+let pr_goal e = Pp.(str "GOAL:" ++ int (Evar.repr e))
 let ppclenv clenv = pp(pr_clenv clenv)
-let ppgoalgoal gl = pp(Goal.pr_goal gl)
-let ppgoal g = pp(Printer.pr_goal g)
-let ppgoalsigma g = pp(Printer.pr_goal g ++ Termops.pr_evar_map None (Global.env ()) (Tacmach.project g))
+let ppgoal g = pp(Printer.Debug.pr_goal g)
 let pphintdb db = pp(envpp Hints.pr_hint_db_env db)
 let ppproofview p =
   let gls,sigma = Proofview.proofview p in
-  pp(pr_enum Goal.pr_goal gls ++ fnl () ++ Termops.pr_evar_map (Some 1) (Global.env ()) sigma)
+  pp(pr_enum pr_goal gls ++ fnl () ++ Termops.pr_evar_map (Some 1) (Global.env ()) sigma)
 
 let ppopenconstr (x : Evd.open_constr) =
   let (evd,c) = x in pp (Termops.pr_evar_map (Some 2) (Global.env ()) evd ++ envpp pr_econstr_env c)
@@ -220,38 +254,40 @@ let pppftreestate p = pp(print_pftreestate p)
 
 let pproof p = pp(Proof.pr_proof p)
 
-let ppuni u = pp(Universe.pr u)
-let ppuni_level u = pp (Level.pr u)
+let ppuni u = pp(Universe.raw_pr u)
+let ppuni_level u = pp (Level.raw_pr u)
+let ppqvar q = pp (QVar.pr q)
+let ppesorts s = pp (Sorts.debug_print (Evd.MiniEConstr.ESorts.unsafe_to_sorts s))
 
-let prlev = UnivNames.pr_with_global_universes Id.Map.empty
-let ppuniverse_set l = pp (LSet.pr prlev l)
+let prlev l = UnivNames.pr_with_global_universes l
+let ppuniverse_set l = pp (Level.Set.pr prlev l)
 let ppuniverse_instance l = pp (Instance.pr prlev l)
 let ppuniverse_context l = pp (pr_universe_context prlev l)
 let ppuniverse_context_set l = pp (pr_universe_context_set prlev l)
-let ppuniverse_subst l = pp (Univ.pr_universe_subst l)
-let ppuniverse_opt_subst l = pp (UnivSubst.pr_universe_opt_subst l)
-let ppuniverse_level_subst l = pp (Univ.pr_universe_level_subst l)
+let ppuniverse_subst l = pp (UnivSubst.pr_universe_subst Level.raw_pr l)
+let ppuniverse_opt_subst l = pp (UState.pr_universe_opt_subst Level.raw_pr l)
+let ppuniverse_level_subst l = pp (Univ.pr_universe_level_subst Level.raw_pr l)
 let ppevar_universe_context l = pp (Termops.pr_evar_universe_context l)
-let ppconstraints c = pp (pr_constraints Level.pr c)
+let ppconstraints c = pp (pr_constraints Level.raw_pr c)
 let ppuniverseconstraints c = pp (UnivProblem.Set.pr c)
 let ppuniverse_context_future c =
   let ctx = Future.force c in
     ppuniverse_context ctx
-let ppuniverses u = pp (UGraph.pr_universes Level.pr (UGraph.repr u))
+let ppuniverses u = pp (UGraph.pr_universes Level.raw_pr (UGraph.repr u))
 let ppnamedcontextval e =
   let env = Global.env () in
   let sigma = Evd.from_env env in
   pp (pr_named_context env sigma (named_context_of_val e))
 
 let ppaucontext auctx =
-  let nas = AUContext.names auctx in
+  let nas = AbstractContext.names auctx in
   let prlev l = match Level.var_index l with
     | Some n -> (match nas.(n) with
         | Anonymous -> prlev l
         | Name id -> Id.print id)
     | None -> prlev l
   in
-  pp (pr_universe_context prlev (AUContext.repr auctx))
+  pp (pr_universe_context prlev (AbstractContext.repr auctx))
 
 
 let ppenv e = pp
@@ -277,7 +313,6 @@ let cast_kind_display k =
   match k with
   | VMcast -> "VMcast"
   | DEFAULTcast -> "DEFAULTcast"
-  | REVERTcast -> "REVERTcast"
   | NATIVEcast -> "NATIVEcast"
 
 let constr_display csr =
@@ -296,7 +331,11 @@ let constr_display csr =
       "LetIn("^(name_display na)^","^(term_display b)^","
       ^(term_display t)^","^(term_display c)^")"
   | App (c,l) -> "App("^(term_display c)^","^(array_display l)^")\n"
-  | Evar (e,l) -> "Evar("^(Pp.string_of_ppcmds (Evar.print e))^","^(array_display (Array.of_list l))^")"
+  | Evar (e,l) ->
+    let l = SList.to_list l in
+    let map = function None -> "?" | Some t -> term_display t in
+    let l = List.map map l in
+    "Evar("^(Pp.string_of_ppcmds (Evar.print e))^", [|"^(String.concat "; " l)^"|])"
   | Const (c,u) -> "Const("^(Constant.to_string c)^","^(universes_display u)^")"
   | Ind ((sp,i),u) ->
       "MutInd("^(MutInd.to_string sp)^","^(string_of_int i)^","^(universes_display u)^")"
@@ -333,10 +372,10 @@ let constr_display csr =
        v "")^"|]"
 
   and univ_display u =
-    incr cnt; pp (str "with " ++ int !cnt ++ str" " ++ pr_uni u ++ fnl ())
+    incr cnt; pp (str "with " ++ int !cnt ++ str" " ++ Universe.raw_pr u ++ fnl ())
 
   and level_display u =
-    incr cnt; pp (str "with " ++ int !cnt ++ str" " ++ Level.pr u ++ fnl ())
+    incr cnt; pp (str "with " ++ int !cnt ++ str" " ++ Level.raw_pr u ++ fnl ())
 
   and sort_display = function
     | SProp -> "SProp"
@@ -344,6 +383,7 @@ let constr_display csr =
     | Prop -> "Prop"
     | Type u -> univ_display u;
         "Type("^(string_of_int !cnt)^")"
+    | QSort (q, u) -> univ_display u; Printf.sprintf "QSort(%s, %i)" (Sorts.QVar.to_string q) !cnt
 
   and universes_display l =
     Array.fold_right (fun x i -> level_display x; (string_of_int !cnt)^(if not(i="")
@@ -393,7 +433,8 @@ let print_pure_constr csr =
       Array.iter (fun x -> print_space (); box_display x) l;
       print_string ")"
   | Evar (e,l) -> print_string "Evar#"; print_int (Evar.repr e); print_string "{";
-      List.iter (fun x -> print_space (); box_display x) l;
+      let iter = function None -> print_space (); print_string "?" | Some t -> print_space (); box_display t in
+      List.iter iter (SList.to_list l);
       print_string"}"
   | Const (c,u) -> print_string "Cons(";
       sp_con_display c;
@@ -485,14 +526,16 @@ let print_pure_constr csr =
   and box_display c = open_hovbox 1; term_display c; close_box()
 
   and universes_display u =
-    Array.iter (fun u -> print_space (); pp (Level.pr u)) (Instance.to_array u)
+    Array.iter (fun u -> print_space (); pp (Level.raw_pr u)) (Instance.to_array u)
 
   and sort_display = function
     | SProp -> print_string "SProp"
     | Set -> print_string "Set"
     | Prop -> print_string "Prop"
     | Type u -> open_hbox();
-        print_string "Type("; pp (pr_uni u); print_string ")"; close_box()
+        print_string "Type("; pp (Universe.raw_pr u); print_string ")"; close_box()
+    | QSort (q, u) -> open_hbox();
+        print_string "QSort("; pp (QVar.pr q); print_string ", "; pp (Universe.raw_pr u); print_string ")"; close_box()
 
   and name_display x = match x.binder_name with
     | Name id -> print_string (Id.to_string id)
@@ -569,23 +612,23 @@ VERNAC COMMAND EXTEND PrintConstr
 END
 *)
 
-let _ =
+let () =
   let open Vernacextend in
   let ty_constr = Extend.TUentry (get_arg_tag Stdarg.wit_constr) in
   let cmd_sig = TyTerminal("PrintConstr", TyNonTerminal(ty_constr, TyNil)) in
-  let cmd_fn c ?loc:_ ~atts () = VtDefault (fun () -> in_current_context econstr_display c) in
+  let cmd_fn c ?loc:_ ~atts () = vtdefault (fun () -> in_current_context econstr_display c) in
   let cmd_class _ = VtQuery in
   let cmd : ty_ml = TyML (false, cmd_sig, cmd_fn, Some cmd_class) in
-  vernac_extend ~command:"PrintConstr" [cmd]
+  static_vernac_extend ~plugin:None ~command:"PrintConstr" [cmd]
 
-let _ =
+let () =
   let open Vernacextend in
   let ty_constr = Extend.TUentry (get_arg_tag Stdarg.wit_constr) in
   let cmd_sig = TyTerminal("PrintPureConstr", TyNonTerminal(ty_constr, TyNil)) in
-  let cmd_fn c ?loc:_ ~atts () = VtDefault (fun () -> in_current_context print_pure_econstr c) in
+  let cmd_fn c ?loc:_ ~atts () = vtdefault (fun () -> in_current_context print_pure_econstr c) in
   let cmd_class _ = VtQuery in
   let cmd : ty_ml = TyML (false, cmd_sig, cmd_fn, Some cmd_class) in
-  vernac_extend ~command:"PrintPureConstr" [cmd]
+  static_vernac_extend ~plugin:None ~command:"PrintPureConstr" [cmd]
 
 (* Setting printer of unbound global reference *)
 open Names
@@ -601,14 +644,14 @@ let encode_path ?loc prefix mpdir suffix id =
 
 let raw_string_of_ref ?loc _ = let open GlobRef in function
   | ConstRef cst ->
-      let (mp,id) = Constant.repr2 cst in
+      let (mp,id) = KerName.repr (Constant.user cst) in
       encode_path ?loc "CST" (Some mp) [] (Label.to_id id)
   | IndRef (kn,i) ->
-      let (mp,id) = MutInd.repr2 kn in
+      let (mp,id) = KerName.repr (MutInd.user kn) in
       encode_path ?loc "IND" (Some mp) [Label.to_id id]
         (Id.of_string ("_"^string_of_int i))
   | ConstructRef ((kn,i),j) ->
-      let (mp,id) = MutInd.repr2 kn in
+      let (mp,id) = KerName.repr (MutInd.user kn) in
       encode_path ?loc "CSTR" (Some mp)
         [Label.to_id id;Id.of_string ("_"^string_of_int i)]
         (Id.of_string ("_"^string_of_int j))
@@ -630,6 +673,5 @@ let short_string_of_ref ?loc _ = let open GlobRef in function
 (* Anticipate that printers can be used from ocamldebug and that
    pretty-printer should not make calls to the global env since ocamldebug
    runs in a different process and does not have the proper env at hand *)
-let _ = Flags.in_debugger := true
-let _ = Constrextern.set_extern_reference
+let () = Constrextern.set_extern_reference
   (if !rawdebug then raw_string_of_ref else short_string_of_ref)
