@@ -14,11 +14,33 @@ open Names
 open Constr
 open Context
 
+module NamedDecl = Context.Named.Declaration
+
 module ESorts = struct
   include Evd.MiniEConstr.ESorts
 
   let equal sigma s1 s2 =
     Sorts.equal (kind sigma s1) (kind sigma s2)
+
+  let is_small sigma s = Sorts.is_small (kind sigma s)
+  let is_prop sigma s = Sorts.is_prop (kind sigma s)
+  let is_sprop sigma s = Sorts.is_sprop (kind sigma s)
+  let is_set sigma s = Sorts.is_set (kind sigma s)
+
+  let prop = make Sorts.prop
+  let sprop = make Sorts.sprop
+  let set = make Sorts.set
+  let type1 = make Sorts.type1
+
+  let super sigma s =
+    make (Sorts.super (kind sigma s))
+
+  let relevance_of_sort sigma s =
+    let r = Sorts.relevance_of_sort (unsafe_to_sorts s) in
+    UState.nf_relevance (Evd.evar_universe_context sigma) r
+
+  let family sigma s = Sorts.family (kind sigma s)
+
 end
 
 module EInstance = struct
@@ -42,10 +64,12 @@ type case = (t, t, EInstance.t) pcase
 type fixpoint = (t, t) pfixpoint
 type cofixpoint = (t, t) pcofixpoint
 type unsafe_judgment = (constr, types) Environ.punsafe_judgment
-type unsafe_type_judgment = types Environ.punsafe_type_judgment
+type unsafe_type_judgment = (types, ESorts.t) Environ.punsafe_type_judgment
 type named_declaration = (constr, types) Context.Named.Declaration.pt
 type rel_declaration = (constr, types) Context.Rel.Declaration.pt
+type compacted_declaration = (constr, types) Context.Compacted.Declaration.pt
 type named_context = (constr, types) Context.Named.pt
+type compacted_context = compacted_declaration list
 type rel_context = (constr, types) Context.Rel.pt
 
 type 'a puniverses = 'a * EInstance.t
@@ -60,18 +84,15 @@ let mkRel n = of_kind (Rel n)
 let mkVar id = of_kind (Var id)
 let mkMeta n = of_kind (Meta n)
 let mkEvar e = of_kind (Evar e)
-let mkSort s = of_kind (Sort (ESorts.make s))
+let mkSort s = of_kind (Sort s)
 let mkCast (b, k, t) = of_kind (Cast (b, k, t))
 let mkProd (na, t, u) = of_kind (Prod (na, t, u))
 let mkLambda (na, t, c) = of_kind (Lambda (na, t, c))
 let mkLetIn (na, b, t, c) = of_kind (LetIn (na, b, t, c))
 let mkApp (f, arg) = of_kind (App (f, arg))
 let mkConstU pc = of_kind (Const pc)
-let mkConst c = of_kind (Const (in_punivs c))
 let mkIndU pi = of_kind (Ind pi)
-let mkInd i = of_kind (Ind (in_punivs i))
 let mkConstructU pc = of_kind (Construct pc)
-let mkConstruct c = of_kind (Construct (in_punivs c))
 let mkConstructUi ((ind,u),i) = of_kind (Construct ((ind,i),u))
 let mkCase (ci, u, pms, c, iv, r, p) = of_kind (Case (ci, u, pms, c, iv, r, p))
 let mkFix f = of_kind (Fix f)
@@ -89,7 +110,9 @@ let mkRef (gr,u) = let open GlobRef in match gr with
   | ConstructRef c -> mkConstructU (c,u)
   | VarRef x -> mkVar x
 
-let type1 = mkSort Sorts.type1
+let mkLEvar = Evd.MiniEConstr.mkLEvar
+
+let type1 = mkSort ESorts.type1
 
 let applist (f, arg) = mkApp (f, Array.of_list arg)
 let applistc f arg = mkApp (f, Array.of_list arg)
@@ -128,15 +151,19 @@ let isRef sigma c = match kind sigma c with
   | Const _ | Ind _ | Construct _ | Var _ -> true
   | _ -> false
 
-let isRefX sigma x c =
+let isRefX env sigma x c =
   let open GlobRef in
   match x, kind sigma c with
-  | ConstRef c, Const (c', _) -> Constant.CanOrd.equal c c'
-  | IndRef i, Ind (i', _) -> Ind.CanOrd.equal i i'
-  | ConstructRef i, Construct (i', _) -> Construct.CanOrd.equal i i'
+  | ConstRef c, Const (c', _) -> Environ.QConstant.equal env c c'
+  | IndRef i, Ind (i', _) -> Environ.QInd.equal env i i'
+  | ConstructRef i, Construct (i', _) -> Environ.QConstruct.equal env i i'
   | VarRef id, Var id' -> Id.equal id id'
   | _ -> false
 
+let is_lib_ref env sigma x c =
+  match Coqlib.lib_ref_opt x with
+  | Some x -> isRefX env sigma x c
+  | None -> false
 
 let destRel sigma c = match kind sigma c with
 | Rel p -> p
@@ -214,11 +241,16 @@ let destRef sigma c = let open GlobRef in match kind sigma c with
   | _ -> raise DestKO
 
 let decompose_app sigma c =
+  match kind sigma c with
+  | App (f,cl) -> (f, cl)
+  | _ -> (c,[||])
+
+let decompose_app_list sigma c =
   match kind sigma c with
     | App (f,cl) -> (f, Array.to_list cl)
     | _ -> (c,[])
 
-let decompose_lam sigma c =
+let decompose_lambda sigma c =
   let rec lamdec_rec l c = match kind sigma c with
     | Lambda (x,t,c) -> lamdec_rec ((x,t)::l) c
     | Cast (c,_,_)     -> lamdec_rec l c
@@ -226,7 +258,7 @@ let decompose_lam sigma c =
   in
   lamdec_rec [] c
 
-let decompose_lam_assum sigma c =
+let decompose_lambda_decls sigma c =
   let open Rel.Declaration in
   let rec lamdec_rec l c =
     match kind sigma c with
@@ -237,10 +269,10 @@ let decompose_lam_assum sigma c =
   in
   lamdec_rec Context.Rel.empty c
 
-let decompose_lam_n_assum sigma n c =
+let decompose_lambda_n_assum sigma n c =
   let open Rel.Declaration in
   if n < 0 then
-    anomaly Pp.(str "decompose_lam_n_assum: integer parameter must be positive.");
+    anomaly Pp.(str "decompose_lambda_n_assum: integer parameter must be positive.");
   let rec lamdec_rec l n c =
     if Int.equal n 0 then l,c
     else
@@ -248,14 +280,14 @@ let decompose_lam_n_assum sigma n c =
       | Lambda (x,t,c)  -> lamdec_rec (Context.Rel.add (LocalAssum (x,t)) l) (n-1) c
       | LetIn (x,b,t,c) -> lamdec_rec (Context.Rel.add (LocalDef (x,b,t)) l) n c
       | Cast (c,_,_)    -> lamdec_rec l n c
-      | c -> anomaly Pp.(str "decompose_lam_n_assum: not enough abstractions.")
+      | c -> anomaly Pp.(str "decompose_lambda_n_assum: not enough abstractions.")
   in
   lamdec_rec Context.Rel.empty n c
 
-let decompose_lam_n_decls sigma n =
+let decompose_lambda_n_decls sigma n =
   let open Rel.Declaration in
   if n < 0 then
-    anomaly Pp.(str "decompose_lam_n_decls: integer parameter must be positive.");
+    anomaly Pp.(str "decompose_lambda_n_decls: integer parameter must be positive.");
   let rec lamdec_rec l n c =
     if Int.equal n 0 then l,c
     else
@@ -263,20 +295,10 @@ let decompose_lam_n_decls sigma n =
       | Lambda (x,t,c)  -> lamdec_rec (Context.Rel.add (LocalAssum (x,t)) l) (n-1) c
       | LetIn (x,b,t,c) -> lamdec_rec (Context.Rel.add (LocalDef (x,b,t)) l) (n-1) c
       | Cast (c,_,_)    -> lamdec_rec l n c
-      | c -> anomaly Pp.(str "decompose_lam_n_decls: not enough abstractions.")
+      | c -> anomaly Pp.(str "decompose_lambda_n_decls: not enough abstractions.")
   in
   lamdec_rec Context.Rel.empty n
 
-let lamn n env b =
-  let rec lamrec = function
-    | (0, env, b)        -> b
-    | (n, ((v,t)::l), b) -> lamrec (n-1,  l, mkLambda (v,t,b))
-    | _ -> assert false
-  in
-  lamrec (n,env,b)
-
-let compose_lam l b = lamn (List.length l) l b
-
 let rec to_lambda sigma n prod =
   if Int.equal n 0 then
     prod
@@ -294,7 +316,7 @@ let decompose_prod sigma c =
   in
   proddec_rec [] c
 
-let decompose_prod_assum sigma c =
+let decompose_prod_decls sigma c =
   let open Rel.Declaration in
   let rec proddec_rec l c =
     match kind sigma c with
@@ -305,10 +327,10 @@ let decompose_prod_assum sigma c =
   in
   proddec_rec Context.Rel.empty c
 
-let decompose_prod_n_assum sigma n c =
+let decompose_prod_n_decls sigma n c =
   let open Rel.Declaration in
   if n < 0 then
-    anomaly Pp.(str "decompose_prod_n_assum: integer parameter must be positive.");
+    anomaly Pp.(str "decompose_prod_n_decls: integer parameter must be positive.");
   let rec prodec_rec l n c =
     if Int.equal n 0 then l,c
     else
@@ -316,10 +338,12 @@ let decompose_prod_n_assum sigma n c =
       | Prod (x,t,c)    -> prodec_rec (Context.Rel.add (LocalAssum (x,t)) l) (n-1) c
       | LetIn (x,b,t,c) -> prodec_rec (Context.Rel.add (LocalDef (x,b,t)) l) (n-1) c
       | Cast (c,_,_)    -> prodec_rec l n c
-      | c -> anomaly Pp.(str "decompose_prod_n_assum: not enough declarations.")
+      | c -> anomaly Pp.(str "decompose_prod_n_decls: not enough assumptions.")
   in
   prodec_rec Context.Rel.empty n c
 
+let prod_decls sigma t = fst (decompose_prod_decls sigma t)
+
 let existential_type = Evd.existential_type
 
 let lift n c = of_constr (Vars.lift n (unsafe_to_constr c))
@@ -347,17 +371,62 @@ let map_return_predicate f p =
   let f c = unsafe_to_constr (f (of_constr c)) in
   of_return (Constr.map_return_predicate f (unsafe_to_return p))
 
+let map_instance sigma f evk args =
+  let rec map ctx args = match ctx, SList.view args with
+  | [], None -> SList.empty
+  | decl :: ctx, Some (Some c, rem) ->
+    let c' = f c in
+    let rem' = map ctx rem in
+    if c' == c && rem' == rem then args
+    else if Constr.isVarId (NamedDecl.get_id decl) c' then SList.default rem'
+    else SList.cons c' rem'
+  | decl :: ctx, Some (None, rem) ->
+    let c = Constr.mkVar (NamedDecl.get_id decl) in
+    let c' = f c in
+    let rem' = map ctx rem in
+    if c' == c && rem' == rem then args
+    else SList.cons c' rem'
+  | [], Some _ | _ :: _, None -> assert false
+  in
+  let EvarInfo evi = Evd.find sigma evk in
+  let ctx = Evd.evar_filtered_context evi in
+  map ctx args
+
 let map sigma f c =
   let f c = unsafe_to_constr (f (of_constr c)) in
-  of_constr (Constr.map f (unsafe_to_constr (whd_evar sigma c)))
+  let c = unsafe_to_constr @@ whd_evar sigma c in
+  match Constr.kind c with
+  | Evar (evk, args) ->
+    let args' = map_instance sigma f evk args in
+    if args' == args then of_constr c else of_constr @@ Constr.mkEvar (evk, args')
+  | _ -> of_constr (Constr.map f c)
 
 let map_with_binders sigma g f l c =
   let f l c = unsafe_to_constr (f l (of_constr c)) in
-  of_constr (Constr.map_with_binders g f l (unsafe_to_constr (whd_evar sigma c)))
+  let c = unsafe_to_constr @@ whd_evar sigma c in
+  match Constr.kind c with
+  | Evar (evk, args) ->
+    let args' = map_instance sigma (fun c -> f l c) evk args in
+    if args' == args then of_constr c else of_constr @@ Constr.mkEvar (evk, args')
+  | _ -> of_constr (Constr.map_with_binders g f l c)
+
+let map_existential sigma f ((evk, args) as ev : existential) =
+  let f c = unsafe_to_constr (f (of_constr c)) in
+  let args : Constr.t SList.t = match Evd.MiniEConstr.unsafe_eq with Refl -> args in
+  let args' = map_instance sigma f evk args in
+  if args' == args then ev
+  else
+    let args' : t SList.t = match Evd.MiniEConstr.unsafe_eq with Refl -> args' in
+    (evk, args')
 
 let iter sigma f c =
   let f c = f (of_constr c) in
-  Constr.iter f (unsafe_to_constr (whd_evar sigma c))
+  let c = unsafe_to_constr @@ whd_evar sigma c in
+  match Constr.kind c with
+  | Evar ((evk, _) as ev) ->
+    let args = Evd.expand_existential0 sigma ev in
+    List.iter (fun c -> f c) args
+  | _ -> Constr.iter f c
 
 let expand_case env _sigma (ci, u, pms, p, iv, c, bl) =
   let u = EInstance.unsafe_to_instance u in
@@ -379,9 +448,9 @@ let annotate_case env sigma (ci, u, pms, p, iv, c, bl as case) =
     (* Too bad we need to fetch this data in the environment, should be in the
       case_info instead. *)
     let (_, mip) = Inductive.lookup_mind_specif env ci.ci_ind in
-    decompose_lam_n_decls sigma (mip.Declarations.mind_nrealdecls + 1) p
+    decompose_lambda_n_decls sigma (mip.Declarations.mind_nrealdecls + 1) p
   in
-  let mk_br c n = decompose_lam_n_decls sigma n c in
+  let mk_br c n = decompose_lambda_n_decls sigma n c in
   let bl = Array.map2 mk_br bl ci.ci_cstr_ndecls in
   (ci, u, pms, p, iv, c, bl)
 
@@ -422,7 +491,9 @@ let iter_with_full_binders env sigma g f n c =
   | Lambda (na,t,c) -> f n t; f (g (LocalAssum (na, t)) n) c
   | LetIn (na,b,t,c) -> f n b; f n t; f (g (LocalDef (na, b, t)) n) c
   | App (c,l) -> f n c; Array.Fun1.iter f n l
-  | Evar (_,l) -> List.iter (fun c -> f n c) l
+  | Evar ((_,l) as ev) ->
+    let l = Evd.expand_existential sigma ev in
+    List.iter (fun c -> f n c) l
   | Case (ci,u,pms,p,iv,c,bl) ->
     let (ci, _, pms, p, iv, c, bl) = annotate_case env sigma (ci, u, pms, p, iv, c, bl) in
     let f_ctx (ctx, c) = f (List.fold_right g ctx n) c in
@@ -440,32 +511,56 @@ let iter_with_full_binders env sigma g f n c =
 
 let iter_with_binders sigma g f n c =
   let f l c = f l (of_constr c) in
-  Constr.iter_with_binders g f n (unsafe_to_constr (whd_evar sigma c))
+  let c = unsafe_to_constr @@ whd_evar sigma c in
+  match Constr.kind c with
+  | Evar ((evk, _) as ev) ->
+    let args = Evd.expand_existential0 sigma ev in
+    List.iter (fun c -> f n c) args
+  | _ -> Constr.iter_with_binders g f n c
 
 let fold sigma f acc c =
   let f acc c = f acc (of_constr c) in
-  Constr.fold f acc (unsafe_to_constr (whd_evar sigma c))
+  let c = unsafe_to_constr @@ whd_evar sigma c in
+  match Constr.kind c with
+  | Evar ((evk, _) as ev) ->
+    let args = Evd.expand_existential0 sigma ev in
+    List.fold_left f acc args
+  | _ -> Constr.fold f acc c
 
-let fold_with_binders sigma g f acc e c =
+let fold_with_binders sigma g f e acc c =
   let f e acc c = f e acc (of_constr c) in
-  Constr.fold_constr_with_binders g f acc e (unsafe_to_constr (whd_evar sigma c))
+  let c = unsafe_to_constr @@ whd_evar sigma c in
+  match Constr.kind c with
+  | Evar ((evk, _) as ev) ->
+    let args = Evd.expand_existential0 sigma ev in
+    List.fold_left (fun acc c -> f e acc c) acc args
+  | _ -> Constr.fold_constr_with_binders g f e acc c
 
-let compare_gen k eq_inst eq_sort eq_constr nargs c1 c2 =
-  (c1 == c2) || Constr.compare_head_gen_with k k eq_inst eq_sort eq_constr nargs c1 c2
+let compare_gen k eq_inst eq_sort eq_constr eq_evars nargs c1 c2 =
+  (c1 == c2) || Constr.compare_head_gen_with k k eq_inst eq_sort eq_constr eq_evars nargs c1 c2
+
+let eq_existential sigma eq (evk1, args1) (evk2, args2) =
+  if Evar.equal evk1 evk2 then
+    let args1 = Evd.expand_existential sigma (evk1, args1) in
+    let args2 = Evd.expand_existential sigma (evk2, args2) in
+    List.equal eq args1 args2
+  else false
 
 let eq_constr sigma c1 c2 =
   let kind c = kind sigma c in
   let eq_inst _ i1 i2 = EInstance.equal sigma i1 i2 in
   let eq_sorts s1 s2 = ESorts.equal sigma s1 s2 in
+  let eq_existential eq e1 e2 = eq_existential sigma (eq 0) e1 e2 in
   let rec eq_constr nargs c1 c2 =
-    compare_gen kind eq_inst eq_sorts eq_constr nargs c1 c2
+    compare_gen kind eq_inst eq_sorts (eq_existential eq_constr) eq_constr nargs c1 c2
   in
   eq_constr 0 c1 c2
 
 let eq_constr_nounivs sigma c1 c2 =
   let kind c = kind sigma c in
+  let eq_existential eq e1 e2 = eq_existential sigma (eq 0) e1 e2 in
   let rec eq_constr nargs c1 c2 =
-    compare_gen kind (fun _ _ _ -> true) (fun _ _ -> true) eq_constr nargs c1 c2
+    compare_gen kind (fun _ _ _ -> true) (fun _ _ -> true) (eq_existential eq_constr) eq_constr nargs c1 c2
   in
   eq_constr 0 c1 c2
 
@@ -473,8 +568,9 @@ let compare_constr sigma cmp c1 c2 =
   let kind c = kind sigma c in
   let eq_inst _ i1 i2 = EInstance.equal sigma i1 i2 in
   let eq_sorts s1 s2 = ESorts.equal sigma s1 s2 in
+  let eq_existential eq e1 e2 = eq_existential sigma (eq 0) e1 e2 in
   let cmp nargs c1 c2 = cmp c1 c2 in
-  compare_gen kind eq_inst eq_sorts cmp 0 c1 c2
+  compare_gen kind eq_inst eq_sorts (eq_existential cmp) cmp 0 c1 c2
 
 let compare_cumulative_instances cv_pb nargs_ok variances u u' cstrs =
   let open UnivProblem in
@@ -488,8 +584,8 @@ let compare_cumulative_instances cv_pb nargs_ok variances u u' cstrs =
          | Irrelevant -> Set.add (UWeak (u,u')) cstrs
          | Covariant ->
            (match cv_pb with
-            | Reduction.CONV -> Set.add (UEq (make u, make u')) cstrs
-            | Reduction.CUMUL -> Set.add (ULe (make u, make u')) cstrs)
+            | Conversion.CONV -> Set.add (UEq (make u, make u')) cstrs
+            | Conversion.CUMUL -> Set.add (ULe (make u, make u')) cstrs)
          | Invariant ->
            Set.add (UEq (make u, make u')) cstrs)
       cstrs variances (Univ.Instance.to_array u) (Univ.Instance.to_array u')
@@ -499,7 +595,7 @@ let cmp_inductives cv_pb (mind,ind as spec) nargs u1 u2 cstrs =
   match mind.Declarations.mind_variance with
   | None -> enforce_eq_instances_univs false u1 u2 cstrs
   | Some variances ->
-    let num_param_arity = Reduction.inductive_cumulativity_arguments spec in
+    let num_param_arity = Conversion.inductive_cumulativity_arguments spec in
     compare_cumulative_instances cv_pb (Int.equal num_param_arity nargs) variances u1 u2 cstrs
 
 let cmp_constructors (mind, ind, cns as spec) nargs u1 u2 cstrs =
@@ -507,7 +603,7 @@ let cmp_constructors (mind, ind, cns as spec) nargs u1 u2 cstrs =
   match mind.Declarations.mind_variance with
   | None -> enforce_eq_instances_univs false u1 u2 cstrs
   | Some _ ->
-    let num_cnstr_args = Reduction.constructor_cumulativity_arguments spec in
+    let num_cnstr_args = Conversion.constructor_cumulativity_arguments spec in
     if not (Int.equal num_cnstr_args nargs)
     then enforce_eq_instances_univs false u1 u2 cstrs
     else
@@ -543,8 +639,8 @@ let test_constr_universes env sigma leq ?(nargs=0) m n =
   if m == n then Some Set.empty
   else
     let cstrs = ref Set.empty in
-    let cv_pb = if leq then Reduction.CUMUL else Reduction.CONV in
-    let eq_universes refargs l l' = eq_universes env sigma cstrs Reduction.CONV refargs l l'
+    let cv_pb = if leq then Conversion.CUMUL else Conversion.CONV in
+    let eq_universes refargs l l' = eq_universes env sigma cstrs Conversion.CONV refargs l l'
     and leq_universes refargs l l' = eq_universes env sigma cstrs cv_pb refargs l l' in
     let eq_sorts s1 s2 =
       let s1 = ESorts.kind sigma s1 in
@@ -563,16 +659,17 @@ let test_constr_universes env sigma leq ?(nargs=0) m n =
            (ULe (s1, s2)) !cstrs;
          true)
     in
-    let rec eq_constr' nargs m n = compare_gen kind eq_universes eq_sorts eq_constr' nargs m n in
+    let eq_existential eq e1 e2 = eq_existential sigma (eq 0) e1 e2 in
+    let rec eq_constr' nargs m n = compare_gen kind eq_universes eq_sorts (eq_existential eq_constr') eq_constr' nargs m n in
     let res =
       if leq then
         let rec compare_leq nargs m n =
-          Constr.compare_head_gen_leq_with kind kind leq_universes leq_sorts
+          Constr.compare_head_gen_leq_with kind kind leq_universes leq_sorts (eq_existential eq_constr')
             eq_constr' leq_constr' nargs m n
         and leq_constr' nargs m n = m == n || compare_leq nargs m n in
         compare_leq nargs m n
       else
-        Constr.compare_head_gen_with kind kind eq_universes eq_sorts eq_constr' nargs m n
+        Constr.compare_head_gen_with kind kind eq_universes eq_sorts (eq_existential eq_constr') eq_constr' nargs m n
     in
     if res then Some !cstrs else None
 
@@ -581,7 +678,7 @@ let eq_constr_universes env sigma ?nargs m n =
 let leq_constr_universes env sigma ?nargs m n =
   test_constr_universes env sigma true ?nargs m n
 
-let compare_head_gen_proj env sigma equ eqs eqc' nargs m n =
+let compare_head_gen_proj env sigma equ eqs eqev eqc' nargs m n =
   let kind c = kind sigma c in
   match kind m, kind n with
   | Proj (p, c), App (f, args)
@@ -593,14 +690,14 @@ let compare_head_gen_proj env sigma equ eqs eqc' nargs m n =
             eqc' 0 c args.(npars)
           else false
       | _ -> false)
-  | _ -> Constr.compare_head_gen_with kind kind equ eqs eqc' nargs m n
+  | _ -> Constr.compare_head_gen_with kind kind equ eqs eqev eqc' nargs m n
 
 let eq_constr_universes_proj env sigma m n =
   let open UnivProblem in
   if m == n then Some Set.empty
   else
     let cstrs = ref Set.empty in
-    let eq_universes ref l l' = eq_universes env sigma cstrs Reduction.CONV ref l l' in
+    let eq_universes ref l l' = eq_universes env sigma cstrs Conversion.CONV ref l l' in
     let eq_sorts s1 s2 =
       let s1 = ESorts.kind sigma s1 in
       let s2 = ESorts.kind sigma s2 in
@@ -610,8 +707,9 @@ let eq_constr_universes_proj env sigma m n =
            (UEq (s1, s2)) !cstrs;
          true)
     in
+    let eq_existential eq e1 e2 = eq_existential sigma (eq 0) e1 e2 in
     let rec eq_constr' nargs m n =
-      m == n || compare_head_gen_proj env sigma eq_universes eq_sorts eq_constr' nargs m n
+      m == n || compare_head_gen_proj env sigma eq_universes eq_sorts (eq_existential eq_constr') eq_constr' nargs m n
     in
     let res = eq_constr' 0 m n in
     if res then Some !cstrs else None
@@ -630,7 +728,7 @@ let universes_of_constr sigma c =
       else
         Level.Set.fold Level.Set.add (Sorts.levels sort) s
     | Evar (k, args) ->
-      let concl = Evd.evar_concl (Evd.find sigma k) in
+      let concl = Evd.evar_concl (Evd.find_undefined sigma k) in
       fold sigma aux (aux s concl) c
     | Array (u,_,_,_) ->
       let s = Level.Set.fold Level.Set.add (Instance.levels (EInstance.kind sigma u)) s in
@@ -647,9 +745,6 @@ open Environ
 let cast_list : type a b. (a,b) eq -> a list -> b list =
   fun Refl x -> x
 
-let cast_list_snd : type a b. (a,b) eq -> ('c * a) list -> ('c * b) list =
-  fun Refl x -> x
-
 let cast_vect : type a b. (a,b) eq -> a array -> b array =
   fun Refl x -> x
 
@@ -696,11 +791,22 @@ let substnl_decl subst n d = of_rel_decl (Vars.substnl_decl (cast_list unsafe_eq
 let substl_decl subst d = of_rel_decl (Vars.substl_decl (cast_list unsafe_eq subst) (to_rel_decl d))
 let subst1_decl c d = of_rel_decl (Vars.subst1_decl (to_constr c) (to_rel_decl d))
 
-let replace_vars subst c =
-  of_constr (Vars.replace_vars (cast_list_snd unsafe_eq subst) (to_constr c))
-let substn_vars n subst c = of_constr (Vars.substn_vars n subst (to_constr c))
-let subst_vars subst c = of_constr (Vars.subst_vars subst (to_constr c))
-let subst_var subst c = of_constr (Vars.subst_var subst (to_constr c))
+type substituend = Vars.substituend
+let make_substituend c = Vars.make_substituend (unsafe_to_constr c)
+let lift_substituend n s = of_constr (Vars.lift_substituend n s)
+
+let replace_vars = replace_vars
+
+(* (subst_var str t) substitute (Var str) by (Rel 1) in t *)
+let subst_var sigma str t = replace_vars sigma [(str, mkRel 1)] t
+
+(* (subst_vars [id1;...;idn] t) substitute (Var idj) by (Rel j) in t *)
+let substn_vars sigma p vars c =
+  let _,subst =
+    List.fold_left (fun (n,l) var -> ((n+1),(var, mkRel n)::l)) (p,[]) vars
+  in replace_vars sigma (List.rev subst) c
+
+let subst_vars sigma subst c = substn_vars sigma 1 subst c
 
 let subst_univs_level_constr subst c =
   of_constr (Vars.subst_univs_level_constr subst (to_constr c))
@@ -715,6 +821,7 @@ let subst_instance_constr subst c =
 let noccurn sigma n term =
   let rec occur_rec n c = match kind sigma c with
     | Rel m -> if Int.equal m n then raise LocalOccur
+    | Evar (_, l) -> SList.Skip.iter (fun c -> occur_rec n c) l
     | _ -> iter_with_binders sigma succ occur_rec n c
   in
   try occur_rec n term; true with LocalOccur -> false
@@ -722,6 +829,7 @@ let noccurn sigma n term =
 let noccur_between sigma n m term =
   let rec occur_rec n c = match kind sigma c with
     | Rel p -> if n<=p && p<n+m then raise LocalOccur
+    | Evar (_, l) -> SList.Skip.iter (fun c -> occur_rec n c) l
     | _        -> iter_with_binders sigma succ occur_rec n c
   in
   try occur_rec n term; true with LocalOccur -> false
@@ -729,6 +837,7 @@ let noccur_between sigma n m term =
 let closedn sigma n c =
   let rec closed_rec n c = match kind sigma c with
     | Rel m -> if m>n then raise LocalOccur
+    | Evar (_, l) -> SList.Skip.iter (fun c -> closed_rec n c) l
     | _ -> iter_with_binders sigma succ closed_rec n c
   in
   try closed_rec n c; true with LocalOccur -> false
@@ -766,11 +875,6 @@ let esubst : (int -> 'a -> t) -> 'a Esubst.subs -> t -> t =
 match unsafe_eq with
 | Refl -> Vars.esubst
 
-type substituend = Vars.substituend
-let make_substituend c = Vars.make_substituend (unsafe_to_constr c)
-let lift_substituend n s = of_constr (Vars.lift_substituend n s)
-
-
 end
 
 let rec isArity sigma c =
@@ -795,37 +899,70 @@ let destArity sigma =
   in
   prodec_rec []
 
+(* Constructs either [forall x:t, c] or [let x:=b:t in c] *)
 let mkProd_or_LetIn decl c =
   let open Context.Rel.Declaration in
   match decl with
   | LocalAssum (na,t) -> mkProd (na, t, c)
   | LocalDef (na,b,t) -> mkLetIn (na, b, t, c)
 
+(* Constructs either [forall x:t, c] or [c] in which [x] is replaced by [b] *)
+let mkProd_wo_LetIn decl c =
+  let open Context.Rel.Declaration in
+  match decl with
+  | LocalAssum (na,t) -> mkProd (na, t, c)
+  | LocalDef (_,b,_) -> Vars.subst1 b c
+
 let mkLambda_or_LetIn decl c =
   let open Context.Rel.Declaration in
   match decl with
   | LocalAssum (na,t) -> mkLambda (na, t, c)
   | LocalDef (na,b,t) -> mkLetIn (na, b, t, c)
 
-let mkNamedProd id typ c = mkProd (map_annot Name.mk_name id, typ, Vars.subst_var id.binder_name c)
-let mkNamedLambda id typ c = mkLambda (map_annot Name.mk_name id, typ, Vars.subst_var id.binder_name c)
-let mkNamedLetIn id c1 t c2 = mkLetIn (map_annot Name.mk_name id, c1, t, Vars.subst_var id.binder_name c2)
+let mkLambda_wo_LetIn decl c =
+  let open Context.Rel.Declaration in
+  match decl with
+  | LocalAssum (na,t) -> mkLambda (na, t, c)
+  | LocalDef (_,b,_) -> Vars.subst1 b c
 
-let mkNamedProd_or_LetIn decl c =
+let mkNamedProd sigma id typ c = mkProd (map_annot Name.mk_name id, typ, Vars.subst_var sigma id.binder_name c)
+let mkNamedLambda sigma id typ c = mkLambda (map_annot Name.mk_name id, typ, Vars.subst_var sigma id.binder_name c)
+let mkNamedLetIn sigma id c1 t c2 = mkLetIn (map_annot Name.mk_name id, c1, t, Vars.subst_var sigma id.binder_name c2)
+
+let mkNamedProd_or_LetIn sigma decl c =
   let open Context.Named.Declaration in
   match decl with
-    | LocalAssum (id,t) -> mkNamedProd id t c
-    | LocalDef (id,b,t) -> mkNamedLetIn id b t c
+  | LocalAssum (id,t) -> mkNamedProd sigma id t c
+  | LocalDef (id,b,t) -> mkNamedLetIn sigma id b t c
 
-let mkNamedLambda_or_LetIn decl c =
+let mkNamedLambda_or_LetIn sigma decl c =
   let open Context.Named.Declaration in
   match decl with
-    | LocalAssum (id,t) -> mkNamedLambda id t c
-    | LocalDef (id,b,t) -> mkNamedLetIn id b t c
+  | LocalAssum (id,t) -> mkNamedLambda sigma id t c
+  | LocalDef (id,b,t) -> mkNamedLetIn sigma id b t c
+
+let mkNamedProd_wo_LetIn sigma decl c =
+  let open Context.Named.Declaration in
+  match decl with
+  | LocalAssum (id,t) -> mkNamedProd sigma id t c
+  | LocalDef (id,b,t) -> Vars.subst1 b c
+
+let it_mkProd init = List.fold_left (fun c (n,t)  -> mkProd (n, t, c)) init
+let it_mkLambda init = List.fold_left (fun c (n,t)  -> mkLambda (n, t, c)) init
+
+let compose_lam l b = it_mkLambda b l
 
 let it_mkProd_or_LetIn t ctx = List.fold_left (fun c d -> mkProd_or_LetIn d c) t ctx
 let it_mkLambda_or_LetIn t ctx = List.fold_left (fun c d -> mkLambda_or_LetIn d c) t ctx
 
+let it_mkProd_wo_LetIn t ctx = List.fold_left (fun c d -> mkProd_wo_LetIn d c) t ctx
+let it_mkLambda_wo_LetIn t ctx = List.fold_left (fun c d -> mkLambda_wo_LetIn d c) t ctx
+
+let it_mkNamedProd_or_LetIn sigma t ctx = List.fold_left (fun c d -> mkNamedProd_or_LetIn sigma d c) t ctx
+let it_mkNamedLambda_or_LetIn sigma t ctx = List.fold_left (fun c d -> mkNamedLambda_or_LetIn sigma d c) t ctx
+
+let it_mkNamedProd_wo_LetIn sigma t ctx = List.fold_left (fun c d -> mkNamedProd_wo_LetIn sigma d c) t ctx
+
 let push_rel d e = push_rel (cast_rel_decl unsafe_eq d) e
 let push_rel_context d e = push_rel_context (cast_rel_context unsafe_eq d) e
 let push_rec_types d e = push_rec_types (cast_rec_decl unsafe_eq d) e
@@ -840,7 +977,7 @@ let val_of_named_context e = val_of_named_context (cast_named_context unsafe_eq
 let named_context_of_val e = cast_named_context (sym unsafe_eq) (named_context_of_val e)
 
 let of_existential : Constr.existential -> existential =
-  let gen : type a b. (a,b) eq -> 'c * b list -> 'c * a list = fun Refl x -> x in
+  let gen : type a b. (a,b) eq -> 'c * b SList.t -> 'c * a SList.t = fun Refl x -> x in
   gen unsafe_eq
 
 let lookup_rel i e = cast_rel_decl (sym unsafe_eq) (lookup_rel i e)
@@ -857,12 +994,12 @@ let map_rel_context_in_env f env sign =
   aux env [] (List.rev sign)
 
 let match_named_context_val :
-  named_context_val -> (named_declaration * lazy_val * named_context_val) option =
+  named_context_val -> (named_declaration * named_context_val) option =
   match unsafe_eq with
   | Refl -> match_named_context_val
 
-let identity_subst_val : named_context_val -> t list =
-  match unsafe_eq with Refl -> fun ctx -> ctx.env_named_var
+let identity_subst_val : named_context_val -> t SList.t = fun ctx ->
+  SList.defaultn (List.length ctx.Environ.env_named_ctx) SList.empty
 
 let fresh_global ?loc ?rigid ?names env sigma reference =
   let (evd,t) = Evd.fresh_global ?loc ?rigid ?names env sigma reference in
@@ -912,3 +1049,22 @@ let to_case_invert = unsafe_to_case_invert
 
 let eq = unsafe_eq
 end
+
+module UnsafeMonomorphic = struct
+  let mkConst c = of_kind (Const (in_punivs c))
+  let mkInd i = of_kind (Ind (in_punivs i))
+  let mkConstruct c = of_kind (Construct (in_punivs c))
+end
+
+(* deprecated *)
+
+let decompose_lambda_assum = decompose_lambda_decls
+let decompose_prod_assum = decompose_prod_decls
+let decompose_prod_n_assum = decompose_prod_n_decls
+let prod_assum = prod_decls
+let decompose_lam = decompose_lambda
+let decompose_lam_n_assum = decompose_lambda_n_assum
+let decompose_lam_n_decls = decompose_lambda_n_decls
+let decompose_lam_assum = decompose_lambda_assum
+
+include UnsafeMonomorphic
