@@ -86,6 +86,8 @@ let init () =
 \n    <menuitem action='Display all low-level contents' />\
 \n    <menuitem action='Display unfocused goals' />\
 \n    <menuitem action='Display goal names' />\
+\n    <menuitem action='Use record syntax' />\
+\n    <menuitem action='Hide synthesizable types' />\
 \n    <separator/>\
 \n    <menuitem action='Unset diff' />\
 \n    <menuitem action='Set diff' />\
@@ -153,7 +155,6 @@ let init () =
 \n    <menuitem action='Browse Coq Library' />\
 \n    <menuitem action='Help for keyword' />\
 \n    <menuitem action='Help for μPG mode' />\
-\n    <menuitem action='Memory usage' />\
 \n    <separator />\
 \n    <menuitem name='Abt' action='About Coq' />\
 \n  </menu>\
@@ -172,7 +173,7 @@ let init () =
 \n  <toolitem action='Next' />\
 \n</toolbar>\
 \n</ui>"
-    (if Coq_config.gtk_platform <> `QUARTZ then "<menuitem action='Quit' />" else "")
+    (if Config.gtk_platform <> `QUARTZ then "<menuitem action='Quit' />" else "")
     (Buffer.contents (list_items "Template" Coq_commands.commands))
     (Buffer.contents (list_queries "User-Query" Preferences.user_queries#get))
  in
