@@ -370,7 +370,7 @@ let cmd_editor =
   new preference ~name:["cmd_editor"] ~init ~repr:Repr.(string)
 
 let text_font =
-  let init = match Coq_config.gtk_platform with
+  let init = match Config.gtk_platform with
   | `QUARTZ -> "Arial Unicode MS 11"
   | _ -> "Monospace 10"
   in
@@ -666,14 +666,60 @@ let tag_button () =
 let pref_file = Filename.concat (Minilib.coqide_config_home ()) "coqiderc"
 let accel_file = Filename.concat (Minilib.coqide_config_home ()) "coqide.keys"
 
+let accel_regex = Str.regexp {|\(; \|\)(gtk_accel_path "\([^""]*\)"|}
+
+exception UnknownFormat
+
+(* Attention: this only works with absolute normalized paths -
+   which can be assumed here since the path comes from  Glib.get_user_config_dir *)
+let rec mkdir_p path perms =
+  if not (Sys.file_exists path)
+  then
+    let parent = Filename.dirname path in
+    if not (Sys.file_exists parent) && parent<>path
+    then mkdir_p parent perms
+    else Unix.mkdir path perms
+  else ()
+
 let save_accel_pref () =
-  if not (Sys.file_exists (Minilib.coqide_config_home ()))
-  then Unix.mkdir (Minilib.coqide_config_home ()) 0o700;
-  GtkData.AccelMap.save accel_file
+  mkdir_p (Minilib.coqide_config_home ()) 0o700;
+  let tmp_file, fd = Filename.open_temp_file ?perms:(Some 0o644)
+    ?temp_dir:(Some (Filename.dirname accel_file))
+    "coqide.keys_" "" in
+  close_out fd;
+  GtkData.AccelMap.save tmp_file;
+  (* AccelMap.save writes entries in random order, so sort them: *)
+  let fd = open_in tmp_file in
+  let map = ref CString.Map.empty in
+  let top_lines = ref [] in
+  begin
+    try
+      while true do
+        let line = input_line fd in
+        if Str.string_match accel_regex line 0 then
+          let key = Str.matched_group 2 line in
+          map := CString.Map.add key line !map
+        else begin
+          if not (CString.Map.is_empty !map) then begin
+            Minilib.log ("Unknown format for coqide.keys; sorting skipped");
+            raise UnknownFormat
+          end;
+          top_lines := line :: !top_lines
+        end
+      done
+    with
+    | UnknownFormat -> close_in fd
+    | End_of_file ->
+      close_in fd;
+      let fd = open_out tmp_file in
+      List.iter (fun l -> Printf.fprintf fd "%s\n" l) (List.rev !top_lines);
+      CString.Map.iter (fun k l -> Printf.fprintf fd "%s\n" l) !map;
+      close_out fd
+  end;
+  Sys.rename tmp_file accel_file
 
 let save_rc_pref () =
-  if not (Sys.file_exists (Minilib.coqide_config_home ()))
-  then Unix.mkdir (Minilib.coqide_config_home ()) 0o700;
+  mkdir_p (Minilib.coqide_config_home ()) 0o700;
   let add = Util.String.Map.add in
   let fold key obj accu = add key (obj.get ()) accu in
   let prefs = Util.String.Map.fold fold !preferences Util.String.Map.empty in
@@ -749,7 +795,27 @@ let configure ?(apply=(fun () -> ())) parent =
   let cmd_coqdoc = pstring "     coqdoc" cmd_coqdoc in
   let cmd_print = pstring "   Print ps" cmd_print in
 
-  let config_font =
+  let config_font = match Coq_config.arch with
+  | "Darwin" ->
+    (* A poor man's font selection. Work around #16136, which is ultimately a GTK bug. *)
+    let box = GPack.vbox () in
+    let () = box#set_height_request 200 in
+    let () = box#set_height_request 300 in
+    let w = GEdit.entry ~text:text_font#get () in
+    let () = box#pack ~expand:false w#coerce in
+    let text = GMisc.label ~text:"Goal (∃n : nat, n ≤ 0)∧(∀x,y,z, x∈y⋃z↔x∈y∨x∈z)." () in
+    let () = text#set_ypad 10 in
+    let () = box#pack ~expand:false text#coerce in
+    let set_font () =
+      let fd = GPango.font_description_from_string w#text in
+      let () = text_font#set fd#to_string in
+      let () = text#misc#modify_font_by_name fd#to_string in
+      w#set_text fd#to_string
+    in
+    let () = text#misc#modify_font_by_name text_font#get in
+    let _ = w#connect#activate ~callback:set_font in
+    custom ~label:"Fonts for text" box set_font true
+  | _ ->
     let box = GPack.hbox () in
     let w = GMisc.font_selection () in
     w#set_preview_text
