@@ -69,6 +69,8 @@ sig
 
     val get_relevance : ('c, 't) pt -> Sorts.relevance
 
+    val set_annot : Name.t binder_annot -> ('c, 't) pt -> ('c, 't) pt
+
     (** Set the name that is bound by a given declaration. *)
     val set_name : Name.t -> ('c, 't) pt -> ('c, 't) pt
 
@@ -160,6 +162,9 @@ sig
       Outermost declarations are processed first. *)
   val fold_outside : (('c, 't) Declaration.pt -> 'a -> 'a) -> ('c, 't) pt -> init:'a -> 'a
 
+  (** Return the set of all named variables bound in a given rel-context. *)
+  val to_vars : ('c, 't) pt -> Id.Set.t
+
   (** Map a given rel-context to a list where each {e local
       assumption} is mapped to [true] and each {e local definition} is
       mapped to [false]. The resulting list is in reverse order
@@ -169,6 +174,14 @@ sig
   (** Turn all [LocalDef] into [LocalAssum], leave [LocalAssum] unchanged. *)
   val drop_bodies : ('c, 't) pt -> ('c, 't) pt
 
+  (** [chop_nhyps n Γ] returns [Γ'',Γ'] such that [Γ]=[Γ'Γ''], [Γ''] has
+      [n] hypotheses (i.e. [LocalAssum]), excluding local definitions
+      (i.e. [LocalDef]), and [Γ''], if [n] non zero, starts with an
+      hypothesis (i.e., [Γ''] has the form [x:A;Γ'''], i.e., local
+      definitions at the junction of the [n] hypotheses are put in
+      [Γ'] rather than in [Γ''] *)
+  val chop_nhyps : int -> ('c, 't) pt -> ('c, 't) pt * ('c, 't) pt
+
   (** [instance mk n Γ] builds an instance [args] such that [Γ,Δ ⊢ args:Γ]
       with n = |Δ| and with the {e local definitions} of [Γ] skipped in
       [args] where [mk] is used to build the corresponding variables.
