@@ -45,7 +45,7 @@ type link_info =
   | Linked of string
   | NotLinked
 
-type constant_key = Opaqueproof.opaque constant_body * (link_info ref * key)
+type constant_key = constant_body * (link_info ref * key)
 
 type mind_key = mutual_inductive_body * link_info ref
 
@@ -63,35 +63,15 @@ module Globals = struct
   let view x = x
 end
 
-type stratification = {
-  env_universes : UGraph.t;
-  env_sprop_allowed : bool;
-  env_universes_lbound : UGraph.Bound.t;
-  env_engagement : engagement
-}
-
-type val_kind =
-    | VKvalue of (Vmvalues.values * Id.Set.t) CEphemeron.key
-    | VKnone
-
-type lazy_val = val_kind ref
-
-let force_lazy_val vk = match !vk with
-| VKnone -> None
-| VKvalue v -> try Some (CEphemeron.get v) with CEphemeron.InvalidKey -> None
-
-let dummy_lazy_val () = ref VKnone
-let build_lazy_val vk key = vk := VKvalue (CEphemeron.create key)
-
 type named_context_val = {
   env_named_ctx : Constr.named_context;
-  env_named_map : (Constr.named_declaration * lazy_val) Id.Map.t;
-  env_named_var : Constr.t list;
+  env_named_map : Constr.named_declaration Id.Map.t;
+  env_named_idx : Constr.named_declaration Range.t;
 }
 
 type rel_context_val = {
   env_rel_ctx : Constr.rel_context;
-  env_rel_map : (Constr.rel_declaration * lazy_val) Range.t;
+  env_rel_map : Constr.rel_declaration Range.t;
 }
 
 type env = {
@@ -99,16 +79,18 @@ type env = {
   env_named_context : named_context_val; (* section variables *)
   env_rel_context   : rel_context_val;
   env_nb_rel        : int;
-  env_stratification : stratification;
+  env_universes : UGraph.t;
+  env_universes_lbound : UGraph.Bound.t;
+  irr_constants : Cset_env.t;
+  irr_inds : Indset_env.t;
   env_typing_flags  : typing_flags;
   retroknowledge : Retroknowledge.retroknowledge;
-  indirect_pterms : Opaqueproof.opaquetab;
 }
 
 let empty_named_context_val = {
   env_named_ctx = [];
   env_named_map = Id.Map.empty;
-  env_named_var = [];
+  env_named_idx = Range.empty;
 }
 
 let empty_rel_context_val = {
@@ -126,14 +108,12 @@ let empty_env = {
   env_named_context = empty_named_context_val;
   env_rel_context = empty_rel_context_val;
   env_nb_rel = 0;
-  env_stratification = {
-    env_universes = UGraph.initial_universes;
-    env_sprop_allowed = true;
-    env_universes_lbound = UGraph.Bound.Set;
-    env_engagement = PredicativeSet };
+  env_universes = UGraph.initial_universes;
+  env_universes_lbound = UGraph.Bound.Set;
+  irr_constants = Cset_env.empty;
+  irr_inds = Indset_env.empty;
   env_typing_flags = Declareops.safe_flags Conv_oracle.empty;
   retroknowledge = Retroknowledge.empty;
-  indirect_pterms = Opaqueproof.empty_opaquetab;
 }
 
 
@@ -141,15 +121,14 @@ let empty_env = {
 
 let push_rel_context_val d ctx = {
   env_rel_ctx = Context.Rel.add d ctx.env_rel_ctx;
-  env_rel_map = Range.cons (d, ref VKnone) ctx.env_rel_map;
+  env_rel_map = Range.cons d ctx.env_rel_map;
 }
 
 let match_rel_context_val ctx = match ctx.env_rel_ctx with
 | [] -> None
 | decl :: rem ->
-  let (_, lval) = Range.hd ctx.env_rel_map in
   let ctx = { env_rel_ctx = rem; env_rel_map = Range.tl ctx.env_rel_map } in
-  Some (decl, lval, ctx)
+  Some (decl, ctx)
 
 let push_rel d env =
     { env with
@@ -157,11 +136,7 @@ let push_rel d env =
       env_nb_rel = env.env_nb_rel + 1 }
 
 let lookup_rel n env =
-  try fst (Range.get env.env_rel_context.env_rel_map (n - 1))
-  with Invalid_argument _ -> raise Not_found
-
-let lookup_rel_val n env =
-  try snd (Range.get env.env_rel_context.env_rel_map (n - 1))
+  try Range.get env.env_rel_context.env_rel_map (n - 1)
   with Invalid_argument _ -> raise Not_found
 
 let rel_skipn n ctx = {
@@ -175,26 +150,27 @@ let env_of_rel n env =
     env_nb_rel = env.env_nb_rel - n
   }
 
+let set_rel_context_val v env =
+  { env with
+    env_rel_context = v;
+    env_nb_rel = Range.length v.env_rel_map; }
+
 (* Named context *)
 
-let push_named_context_val_val d rval ctxt =
+let push_named_context_val d ctxt =
 (*   assert (not (Id.Map.mem (NamedDecl.get_id d) ctxt.env_named_map)); *)
   {
     env_named_ctx = Context.Named.add d ctxt.env_named_ctx;
-    env_named_map = Id.Map.add (NamedDecl.get_id d) (d, rval) ctxt.env_named_map;
-    env_named_var = mkVar (NamedDecl.get_id d) :: ctxt.env_named_var;
+    env_named_map = Id.Map.add (NamedDecl.get_id d) d ctxt.env_named_map;
+    env_named_idx = Range.cons d ctxt.env_named_idx;
   }
 
-let push_named_context_val d ctxt =
-  push_named_context_val_val d (ref VKnone) ctxt
-
 let match_named_context_val c = match c.env_named_ctx with
 | [] -> None
 | decl :: ctx ->
-  let (_, v) = Id.Map.find (NamedDecl.get_id decl) c.env_named_map in
   let map = Id.Map.remove (NamedDecl.get_id decl) c.env_named_map in
-  let cval = { env_named_ctx = ctx; env_named_map = map; env_named_var = List.tl c.env_named_var } in
-  Some (decl, v, cval)
+  let cval = { env_named_ctx = ctx; env_named_map = map; env_named_idx = Range.tl c.env_named_idx } in
+  Some (decl, cval)
 
 let map_named_val f ctxt =
   let open Context.Named.Declaration in
@@ -202,25 +178,24 @@ let map_named_val f ctxt =
     let d' = f d in
     let accu =
       if d == d' then accu
-      else Id.Map.modify (get_id d) (fun _ (_, v) -> (d', v)) accu
+      else Id.Map.set (get_id d) d' accu
     in
     (accu, d')
   in
-  let map, ctx = List.fold_left_map fold ctxt.env_named_map ctxt.env_named_ctx in
+  let map, ctx = List.Smart.fold_left_map fold ctxt.env_named_map ctxt.env_named_ctx in
   if map == ctxt.env_named_map then ctxt
-  else { env_named_ctx = ctx; env_named_map = map; env_named_var = ctxt.env_named_var }
+  else
+    let idx = List.fold_right Range.cons ctx Range.empty in
+    { env_named_ctx = ctx; env_named_map = map; env_named_idx = idx }
 
 let push_named d env =
   {env with env_named_context = push_named_context_val d env.env_named_context}
 
 let lookup_named id env =
-  fst (Id.Map.find id env.env_named_context.env_named_map)
-
-let lookup_named_val id env =
-  snd(Id.Map.find id env.env_named_context.env_named_map)
+  Id.Map.find id env.env_named_context.env_named_map
 
 let lookup_named_ctxt id ctxt =
-  fst (Id.Map.find id ctxt.env_named_map)
+  Id.Map.find id ctxt.env_named_map
 
 let fold_constants f env acc =
   Cmap_env.fold (fun c (body,_) acc -> f c body acc) env.env_globals.Globals.constants acc
@@ -262,46 +237,36 @@ let set_oracle env o =
   let env_typing_flags = { env.env_typing_flags with conv_oracle = o } in
   { env with env_typing_flags }
 
-let engagement env = env.env_stratification.env_engagement
 let typing_flags env = env.env_typing_flags
 
-let is_impredicative_set env =
-  match engagement env with
-  | ImpredicativeSet -> true
-  | _ -> false
+let is_impredicative_set env = env.env_typing_flags.impredicative_set
 
 let is_impredicative_sort env = function
   | Sorts.SProp | Sorts.Prop -> true
   | Sorts.Set -> is_impredicative_set env
-  | Sorts.Type _ -> false
-
-let is_impredicative_univ env u = is_impredicative_sort env (Sorts.sort_of_univ u)
+  | Sorts.Type _ | Sorts.QSort _ -> false
 
 let is_impredicative_family env = function
   | Sorts.InSProp | Sorts.InProp -> true
   | Sorts.InSet -> is_impredicative_set env
-  | Sorts.InType -> false
+  | Sorts.InType | Sorts.InQSort -> false
 
 let type_in_type env = not (typing_flags env).check_universes
 let deactivated_guard env = not (typing_flags env).check_guarded
 
 let indices_matter env = env.env_typing_flags.indices_matter
 
-let universes env = env.env_stratification.env_universes
-let universes_lbound env = env.env_stratification.env_universes_lbound
+let universes env = env.env_universes
+let universes_lbound env = env.env_universes_lbound
 
 let set_universes g env =
-  {env with env_stratification = {env.env_stratification with env_universes=g}}
+  {env with env_universes=g}
 
-let set_universes_lbound env lbound =
-  let env_stratification = { env.env_stratification with env_universes_lbound = lbound } in
-  { env with env_stratification }
+let set_universes_lbound env lbound = { env with env_universes_lbound = lbound }
 
 let named_context env = env.env_named_context.env_named_ctx
 let named_context_val env = env.env_named_context
 let rel_context env = env.env_rel_context.env_rel_ctx
-let opaque_tables env = env.indirect_pterms
-let set_opaque_tables env indirect_pterms = { env with indirect_pterms }
 
 let empty_context env =
   match env.env_rel_context.env_rel_ctx, env.env_named_context.env_named_ctx with
@@ -324,7 +289,7 @@ let fold_rel_context f env ~init =
   let rec fold_right env =
     match match_rel_context_val env.env_rel_context with
     | None -> init
-    | Some (rd, _, rc) ->
+    | Some (rd, rc) ->
         let env =
           { env with
             env_rel_context = rc;
@@ -377,7 +342,7 @@ let pop_rel_context n env =
     if Int.equal n 0 then ctx
     else match match_rel_context_val ctx with
     | None -> invalid_arg "List.skipn"
-    | Some (_, _, ctx) -> skip (pred n) ctx
+    | Some (_, ctx) -> skip (pred n) ctx
   in
   let ctxt = env.env_rel_context in
   { env with
@@ -388,7 +353,7 @@ let fold_named_context f env ~init =
   let rec fold_right env =
     match match_named_context_val env.env_named_context with
     | None -> init
-    | Some (d, _v, rem) ->
+    | Some (d, rem) ->
         let env =
           reset_with_named_context rem env in
         f env d (fold_right env)
@@ -400,17 +365,14 @@ let fold_named_context_reverse f ~init env =
 
 (* Universe constraints *)
 
-let map_universes f env =
-  let s = env.env_stratification in
-    { env with env_stratification =
-         { s with env_universes = f s.env_universes } }
+let map_universes f env = set_universes (f env.env_universes) env
 
 let add_constraints c env =
-  if Univ.Constraint.is_empty c then env
+  if Univ.Constraints.is_empty c then env
   else map_universes (UGraph.merge_constraints c) env
 
 let check_constraints c env =
-  UGraph.check_constraints c env.env_stratification.env_universes
+  UGraph.check_constraints c env.env_universes
 
 let add_universes ~lbound ~strict ctx g =
   let g = Array.fold_left
@@ -423,7 +385,7 @@ let push_context ?(strict=false) ctx env =
   map_universes (add_universes ~lbound:(universes_lbound env) ~strict ctx) env
 
 let add_universes_set ~lbound ~strict ctx g =
-  let g = Univ.LSet.fold
+  let g = Univ.Level.Set.fold
             (* Be lenient, module typing reintroduces universes and constraints due to includes *)
             (fun v g -> try UGraph.add_universe ~lbound ~strict v g with UGraph.AlreadyDeclared -> g)
             (Univ.ContextSet.levels ctx) g
@@ -435,9 +397,9 @@ let push_context_set ?(strict=false) ctx env =
 let push_subgraph (levels,csts) env =
   let lbound = universes_lbound env in
   let add_subgraph g =
-    let newg = Univ.LSet.fold (fun v g -> UGraph.add_universe ~lbound ~strict:false v g) levels g in
+    let newg = Univ.Level.Set.fold (fun v g -> UGraph.add_universe ~lbound ~strict:false v g) levels g in
     let newg = UGraph.merge_constraints csts newg in
-    (if not (Univ.Constraint.is_empty csts) then
+    (if not (Univ.Constraints.is_empty csts) then
        let restricted = UGraph.constraints_for ~kept:(UGraph.domain g) newg in
        (if not (UGraph.check_constraints restricted g) then
           CErrors.anomaly Pp.(str "Local constraints imply new transitive constraints.")));
@@ -445,10 +407,6 @@ let push_subgraph (levels,csts) env =
   in
   map_universes add_subgraph env
 
-let set_engagement c env =
-  { env with env_stratification =
-    { env.env_stratification with env_engagement = c } }
-
 (* It's convenient to use [{flags with foo = bar}] so we're smart wrt to it. *)
 let same_flags {
      check_guarded;
@@ -459,7 +417,8 @@ let same_flags {
      share_reduction;
      enable_VM;
      enable_native_compiler;
-     cumulative_sprop;
+     impredicative_set;
+     sprop_allowed;
      allow_uip;
   } alt =
   check_guarded == alt.check_guarded &&
@@ -470,35 +429,33 @@ let same_flags {
   share_reduction == alt.share_reduction &&
   enable_VM == alt.enable_VM &&
   enable_native_compiler == alt.enable_native_compiler &&
-  cumulative_sprop == alt.cumulative_sprop &&
+  impredicative_set == alt.impredicative_set &&
+  sprop_allowed == alt.sprop_allowed &&
   allow_uip == alt.allow_uip
 [@warning "+9"]
 
-let set_cumulative_sprop b = map_universes (UGraph.set_cumulative_sprop b)
 let set_type_in_type b = map_universes (UGraph.set_type_in_type b)
 
 let set_typing_flags c env =
   if same_flags env.env_typing_flags c then env
   else
     let env = { env with env_typing_flags = c } in
-    let env = set_cumulative_sprop c.cumulative_sprop env in
     let env = set_type_in_type (not c.check_universes) env in
     env
 
 let update_typing_flags ?typing_flags env =
   Option.cata (fun flags -> set_typing_flags flags env) env typing_flags
 
-let set_cumulative_sprop b env =
-  set_typing_flags {env.env_typing_flags with cumulative_sprop=b} env
+let set_impredicative_set b env =
+  set_typing_flags {env.env_typing_flags with impredicative_set=b} env
 
 let set_type_in_type b env =
   set_typing_flags {env.env_typing_flags with check_universes=not b} env
 
 let set_allow_sprop b env =
-  { env with env_stratification =
-    { env.env_stratification with env_sprop_allowed = b } }
+  set_typing_flags {env.env_typing_flags with sprop_allowed=b} env
 
-let sprop_allowed env = env.env_stratification.env_sprop_allowed
+let sprop_allowed env = env.env_typing_flags.sprop_allowed
 
 (* Global constants *)
 
@@ -509,8 +466,13 @@ let add_constant_key kn cb linkinfo env =
     Cmap_env.add kn (cb,(ref linkinfo, ref None)) env.env_globals.Globals.constants in
   let new_globals =
     { env.env_globals with
-        Globals.constants = new_constants } in
-  { env with env_globals = new_globals }
+      Globals.constants = new_constants }
+  in
+  let irr_constants = if cb.const_relevance == Sorts.Irrelevant
+    then Cset_env.add kn env.irr_constants
+    else env.irr_constants
+  in
+  { env with irr_constants; env_globals = new_globals }
 
 let add_constant kn cb env =
   add_constant_key kn cb no_link_info env
@@ -519,7 +481,7 @@ let add_constant kn cb env =
 let constant_type env (kn,u) =
   let cb = lookup_constant kn env in
   let uctx = Declareops.constant_polymorphic_context cb in
-  let csts = Univ.AUContext.instantiate u uctx in
+  let csts = Univ.AbstractContext.instantiate u uctx in
   (subst_instance_constr u cb.const_type, csts)
 
 type const_evaluation_result =
@@ -532,7 +494,7 @@ exception NotEvaluableConst of const_evaluation_result
 let constant_value_and_type env (kn, u) =
   let cb = lookup_constant kn env in
   let uctx = Declareops.constant_polymorphic_context cb in
-  let cst = Univ.AUContext.instantiate u uctx in
+  let cst = Univ.AbstractContext.instantiate u uctx in
   let b' = match cb.const_body with
     | Def l_body -> Some (subst_instance_constr u l_body)
     | OpaqueDef _ -> None
@@ -598,8 +560,8 @@ let is_array_type env c =
   | Some c' -> Constant.CanOrd.equal c c'
 
 let is_primitive_type env c =
-  (* dummy match to force an update if we add a primitive type, seperated clauses to satisfy ocaml 4.05 *)
-  let _ = function CPrimitives.(PTE(PT_int63)) -> () | CPrimitives.(PTE(PT_float64)) -> () | CPrimitives.(PTE(PT_array)) -> () in
+  (* dummy match to force an update if we add a primitive type *)
+  let _ = function CPrimitives.(PTE(PT_int63)) | CPrimitives.(PTE(PT_float64)) | CPrimitives.(PTE(PT_array)) -> () in
   is_int63_type env c || is_float64_type env c || is_array_type env c
 
 let polymorphic_constant cst env =
@@ -657,12 +619,18 @@ let template_polymorphic_pind (ind,u) env =
   if not (Univ.Instance.is_empty u) then false
   else template_polymorphic_ind ind env
 
-let add_mind_key kn (_mind, _ as mind_key) env =
+let add_mind_key kn (mind, _ as mind_key) env =
   let new_inds = Mindmap_env.add kn mind_key env.env_globals.Globals.inductives in
   let new_globals =
     { env.env_globals with
-        Globals.inductives = new_inds; } in
-  { env with env_globals = new_globals }
+      Globals.inductives = new_inds; }
+  in
+  let irr_inds = Array.fold_left_i (fun i irr_inds mip ->
+      if mip.mind_relevance == Sorts.Irrelevant
+      then Indset_env.add (kn, i) irr_inds
+      else irr_inds) env.irr_inds mind.mind_packets
+  in
+  { env with irr_inds; env_globals = new_globals }
 
 let add_mind kn mib env =
   let li = ref no_link_info in add_mind_key kn (mib, li) env
@@ -688,7 +656,7 @@ let constant_context env c =
 let universes_of_global env r =
   let open GlobRef in
     match r with
-    | VarRef _ -> Univ.AUContext.empty
+    | VarRef _ -> Univ.AbstractContext.empty
     | ConstRef c -> constant_context env c
     | IndRef (mind,_) | ConstructRef ((mind,_),_) ->
       let mib = lookup_mind mind env in
@@ -784,11 +752,11 @@ let make_judge v tj =
 let j_val j = j.uj_val
 let j_type j = j.uj_type
 
-type 'types punsafe_type_judgment = {
+type ('types, 'sorts) punsafe_type_judgment = {
   utj_val : 'types;
-  utj_type : Sorts.t }
+  utj_type : 'sorts }
 
-type unsafe_type_judgment = types punsafe_type_judgment
+type unsafe_type_judgment = (types, Sorts.t) punsafe_type_judgment
 
 exception Hyp_not_found
 
@@ -796,33 +764,36 @@ let apply_to_hyp ctxt id f =
   let open Context.Named.Declaration in
   let rec aux rtail ctxt =
     match match_named_context_val ctxt with
-    | Some (d, v, ctxt) ->
+    | Some (d, ctxt) ->
         if Id.equal (get_id d) id then
-          push_named_context_val_val (f ctxt.env_named_ctx d rtail) v ctxt
+          push_named_context_val (f ctxt.env_named_ctx d rtail) ctxt
         else
           let ctxt' = aux (d::rtail) ctxt in
-          push_named_context_val_val d v ctxt'
+          push_named_context_val d ctxt'
     | None -> raise Hyp_not_found
   in aux [] ctxt
 
 (* To be used in Logic.clear_hyps *)
-let remove_hyps ids check_context check_value ctxt =
-  let rec remove_hyps ctxt = match match_named_context_val ctxt with
-  | None -> empty_named_context_val, false
-  | Some (d, v, rctxt) ->
-     let open Context.Named.Declaration in
-    let (ans, seen) = remove_hyps rctxt in
-    if Id.Set.mem (get_id d) ids then (ans, true)
-    else if not seen then ctxt, false
-    else
-      let rctxt' = ans in
-      let d' = check_context d in
-      let v' = check_value v in
-      if d == d' && v == v' && rctxt == rctxt' then
-        ctxt, true
-      else push_named_context_val_val d' v' rctxt', true
+let remove_hyps ids check_context ctxt =
+  let rec remove_hyps ids ctxt =
+    if Id.Set.is_empty ids then ctxt, false
+    else match match_named_context_val ctxt with
+    | None -> empty_named_context_val, false
+    | Some (d, rctxt) ->
+      let id0 = Context.Named.Declaration.get_id d in
+      let removed = Id.Set.mem id0 ids in
+      let ids = if removed then Id.Set.remove id0 ids else ids in
+      let (ans, seen) = remove_hyps ids rctxt in
+      if removed then (ans, true)
+      else if not seen then ctxt, false
+      else
+        let rctxt' = ans in
+        let d' = check_context d in
+        if d == d' && rctxt == rctxt' then
+          ctxt, true
+        else push_named_context_val d' rctxt', true
   in
-  fst (remove_hyps ctxt)
+  fst (remove_hyps ids ctxt)
 
 (* A general request *)
 
