@@ -73,7 +73,7 @@ let pp_sort s =
   | Sorts.InSet -> str "Set"
   | Sorts.InProp -> str "Prop"
   | Sorts.InSProp -> str "SProp"
-  | Sorts.InType -> str "Type"
+  | Sorts.InType | Sorts.InQSort -> str "Type"
 
 let pr_con sp = str(Names.Label.to_string (Constant.label sp))
 
@@ -574,7 +574,7 @@ let rec get_fix_struct env i t = match kind (Reduction.whd_all env t) with
 | Prod (na, dom, t) ->
   if Int.equal i 0 then
     let dom = Reduction.whd_all env dom in
-    let (dom, _) = decompose_appvect dom in
+    let (dom, _) = decompose_app dom in
     match kind dom with
     | Ind (ind, _) -> ind
     | _ -> assert false
@@ -612,7 +612,7 @@ let rec lambda_of_constr cache env sigma c =
       Lprod(ld,  Llam([|id|], lc))
 
   | Lambda _ ->
-      let params, body = Term.decompose_lam c in
+      let params, body = Term.decompose_lambda c in
       let fold (na, t) env = Environ.push_rel (RelDecl.LocalAssum (na, t)) env in
       let env = List.fold_right fold params env in
       let lb = lambda_of_constr cache env sigma body in
