@@ -11,4 +11,4 @@
 open Declarations
 open Environ
 
-val check_subtypes : 'a Reduction.universe_state -> env -> module_type_body -> module_type_body -> 'a
+val check_subtypes : 'a Conversion.universe_state -> env -> module_type_body -> module_type_body -> 'a
