@@ -60,6 +60,12 @@ module UGlobal = struct
       let c = DirPath.compare u1.library u2.library in
       if c <> 0 then c
       else String.compare u1.process u2.process
+
+  let to_string { library = d; process = s; uid = n } =
+    DirPath.to_string d ^
+    (if CString.is_empty s then "" else "." ^ s) ^
+    "." ^ string_of_int n
+
 end
 
 module RawLevel =
@@ -173,13 +179,12 @@ module Level = struct
   let to_string x =
     match data x with
     | Set -> "Set"
-    | UGlobal.(Level { library = d; process = s; uid = n }) ->
-      Names.DirPath.to_string d ^
-      (if CString.is_empty s then "" else "." ^ s) ^
-      "." ^ string_of_int n
+    | Level l -> UGlobal.to_string l
     | Var n -> "Var(" ^ string_of_int n ^ ")"
 
-  let pr u = str (to_string u)
+  let raw_pr u = str (to_string u)
+
+  let pr = raw_pr
 
   let vars = Array.init 20 (fun i -> make (Var i))
 
@@ -221,9 +226,9 @@ module Level = struct
         else add u v acc)
         ext empty
 
-    let pr f m =
+    let pr prl f m =
       h (prlist_with_sep fnl (fun (u, v) ->
-        pr u ++ f v) (bindings m))
+        prl u ++ f v) (bindings m))
 
   end
 
@@ -330,12 +335,6 @@ struct
         if Int.equal cmp 0 then SuperSame (n < n')
         else SuperDiff cmp
 
-    let to_string (v, n) =
-      if Int.equal n 0 then Level.to_string v
-      else Level.to_string v ^ "+" ^ string_of_int n
-
-    let pr x = str(to_string x)
-
     let pr_with f (v, n) =
       if Int.equal n 0 then f v
       else f v ++ str"+" ++ int n
@@ -381,20 +380,15 @@ struct
   let make l = tip (Expr.make l)
   let tip x = tip x
 
-  let pr l = match l with
-    | [u] -> Expr.pr u
-    | _ ->
-      str "max(" ++ hov 0
-        (prlist_with_sep pr_comma Expr.pr l) ++
-        str ")"
-
-  let pr_with f l = match l with
+  let pr f l = match l with
     | [u] -> Expr.pr_with f u
     | _ ->
       str "max(" ++ hov 0
         (prlist_with_sep pr_comma (Expr.pr_with f) l) ++
         str ")"
 
+  let raw_pr l = pr Level.raw_pr l
+
   let is_level l = match l with
     | [l] -> Expr.is_level l
     | _ -> false
@@ -804,6 +798,7 @@ struct
 
   let make names (univs, _ as x) =
     assert (Array.length names = Array.length univs);
+    assert (Array.for_all (fun u -> not (Level.is_small u)) univs);
     (names, x)
 
   (** Universe contexts (variables as a list) *)
@@ -893,7 +888,6 @@ struct
 
   let of_set s = (s, Constraints.empty)
   let singleton l = of_set (Level.Set.singleton l)
-  let of_instance i = of_set (Instance.levels i)
 
   let union (univs, cst as x) (univs', cst' as y) =
     if x == y then x
@@ -913,12 +907,6 @@ struct
   let add_constraints cst' (univs, cst) =
     univs, Constraints.union cst cst'
 
-  let add_instance inst (univs, cst) =
-    let v = Instance.to_array inst in
-    let fold accu u = Level.Set.add u accu in
-    let univs = Array.fold_left fold univs v in
-    (univs, cst)
-
   let sort_levels a =
     Array.sort Level.compare a; a
 
@@ -1029,8 +1017,8 @@ let pr_abstract_universe_context = AbstractContext.pr
 
 let pr_universe_context_set = ContextSet.pr
 
-let pr_universe_level_subst =
-  Level.Map.pr (fun u -> str" := " ++ Level.pr u ++ spc ())
+let pr_universe_level_subst prl =
+  Level.Map.pr prl (fun u -> str" := " ++ prl u ++ spc ())
 
 module Huniverse_set =
   Hashcons.Make(
@@ -1051,5 +1039,3 @@ let hcons_universe_context_set (v, c) =
   (hcons_universe_set v, hcons_constraints c)
 
 let hcons_univ x = Universe.hcons x
-
-type explanation = Level.t * (constraint_type * Level.t) list
