@@ -99,21 +99,35 @@ let lift_rel_context n =
 
 (* 1st : general case *)
 
-type info = Closed | Open | Unknown
+module IntTbl = Hashtbl.Make(Int)
+
+type info = Closed | Open of Constr.t IntTbl.t | Unknown
 type substituend = { mutable sinfo: info; sit: Constr.t }
 
 let lift_substituend depth s =
   match s.sinfo with
     | Closed -> s.sit
-    | Open -> lift depth s.sit
+    | Open cache ->
+      begin match IntTbl.find_opt cache depth with
+      | Some v -> v
+      | None ->
+        let v = lift depth s.sit in
+        let () = IntTbl.add cache depth v in
+        v
+      end
     | Unknown ->
       let sit = s.sit in
       if closed0 sit then
         let () = s.sinfo <- Closed in
         sit
       else
-        let () = s.sinfo <- Open in
-        lift depth sit
+        let v = lift depth sit in
+        let cache = IntTbl.create 13 in
+        let () = IntTbl.add cache depth v in
+        let () = s.sinfo <- Open cache in
+        v
+
+let lift_substituend depth s = if Int.equal depth 0 then s.sit else lift_substituend depth s
 
 let make_substituend c = { sinfo=Unknown; sit=c }
 
@@ -217,11 +231,7 @@ let rec thin_val = function
       else (id, make_substituend c) :: (thin_val tl)
     | _ -> (id, make_substituend c) :: (thin_val tl)
 
-let rec find_var id = function
-| [] -> raise_notrace Not_found
-| (idc, c) :: subst ->
-  if Id.equal id idc then c
-  else find_var id subst
+let find_var id vars = CList.assoc_f Id.equal id vars
 
 (* (replace_vars sigma M) applies substitution sigma to term M *)
 let replace_vars var_alist x =
