@@ -1,7 +1,7 @@
 open Util
 open Names
 open Environ
-open Reduction
+open Conversion
 open Vm
 open Values
 open Vmvalues
@@ -181,13 +181,13 @@ and conv_arguments env ?from:(from=0) k args1 args2 cu =
 
 let warn_bytecode_compiler_failed =
   let open Pp in
-  CWarnings.create ~name:"bytecode-compiler-failed" ~category:"bytecode-compiler"
+  CWarnings.create ~name:"bytecode-compiler-failed" ~category:CWarnings.CoreCategories.bytecode_compiler
          (fun () -> strbrk "Bytecode compiler failed, " ++
                       strbrk "falling back to standard conversion")
 
 let vm_conv_gen cv_pb sigma env univs t1 t2 =
   if not (typing_flags env).Declarations.enable_VM then
-    Reduction.generic_conv cv_pb ~l2r:false sigma.Genlambda.evars_val
+    Conversion.generic_conv cv_pb ~l2r:false sigma.Genlambda.evars_val
       TransparentState.full env univs t1 t2
   else
   try
@@ -196,7 +196,7 @@ let vm_conv_gen cv_pb sigma env univs t1 t2 =
     fst (conv_val env cv_pb (nb_rel env) v1 v2 univs)
   with Not_found | Invalid_argument _ ->
     warn_bytecode_compiler_failed ();
-    Reduction.generic_conv cv_pb ~l2r:false sigma.Genlambda.evars_val
+    Conversion.generic_conv cv_pb ~l2r:false sigma.Genlambda.evars_val
       TransparentState.full env univs t1 t2
 
 let vm_conv cv_pb env t1 t2 =
