@@ -68,13 +68,11 @@ val pr_info : Environ.env -> Evd.evar_map -> function_info -> Pp.t
 
 val pr_table : Environ.env -> Evd.evar_map -> Pp.t
 
-module New : sig
-  val observe_tac :
-       header:Pp.t
-    -> (Environ.env -> Evd.evar_map -> Pp.t)
-    -> unit Proofview.tactic
-    -> unit Proofview.tactic
-end
+val observe_tac :
+  header:Pp.t
+  -> (Environ.env -> Evd.evar_map -> Pp.t)
+  -> unit Proofview.tactic
+  -> unit Proofview.tactic
 
 (* val function_debug : bool ref  *)
 val observe : Pp.t -> unit
@@ -101,7 +99,7 @@ val evaluable_of_global_reference :
 
 val list_rewrite : bool -> (EConstr.constr * bool) list -> unit Proofview.tactic
 
-val decompose_lam_n :
+val decompose_lambda_n :
      Evd.evar_map
   -> int
   -> EConstr.t
