@@ -196,7 +196,7 @@ let string_of_genarg_arg (ArgumentType arg) =
 
   let pr_quantified_hypothesis = function
     | AnonHyp n -> int n
-    | NamedHyp id -> pr_id id
+    | NamedHyp id -> pr_id id.CAst.v
 
   let pr_clear_flag clear_flag pp x =
     match clear_flag with
@@ -492,7 +492,15 @@ let string_of_genarg_arg (ArgumentType arg) =
         (prlist_with_sep (fun () -> str",")
            (fun id -> spc () ++ pr_hyp_location pr_id id) l ++ pr_occs)
 
-  let pr_orient b = if b then mt () else str "<- "
+
+  (* When the [ssreflect.SsrSynax] module is imported, ssreflect operates
+     in reduced compatibility mode. During printing, we try to account for
+     this when this module is imported. See [plugins/ssr/ssrparser.mlg] for
+     the code that enables the reduced compatibility mode. *)
+  let ssr_loaded = ref (fun () -> false)
+  let ssr_loaded_hook f = ssr_loaded := f
+
+  let pr_orient b = if b then if !ssr_loaded () then str "-> " else mt () else str "<- "
 
   let pr_multi = let open Equality in function
     | Precisely 1 -> mt ()
@@ -801,16 +809,6 @@ let pr_goal_selector ~toplevel s =
                 else pr_pose_as_style (pr.pr_constr env sigma) na c) ++
               pr_opt (fun p -> pr_eqn_ipat p ++ spc ()) e ++
               pr_non_empty_arg (pr_clauses (Some b) pr.pr_name) cl)
-        (*  | TacInstantiate (n,c,ConclLocation ()) ->
-            hov 1 (str "instantiate" ++ spc() ++
-            hov 1 (str"(" ++ pr_arg int n ++ str" :=" ++
-            pr_lconstrarg c ++ str ")" ))
-            | TacInstantiate (n,c,HypLocation (id,hloc)) ->
-            hov 1 (str "instantiate" ++ spc() ++
-            hov 1 (str"(" ++ pr_arg int n ++ str" :=" ++
-            pr_lconstrarg c ++ str ")" )
-            ++ str "in" ++ pr_hyp_location pr.pr_name (id,[],(hloc,ref None)))
-        *)
 
         (* Derived basic tactics *)
         | TacInductionDestruct (isrec,ev,(l,el)) ->
@@ -982,11 +980,6 @@ let pr_goal_selector ~toplevel s =
                 keyword "progress" ++ spc ()
                 ++ pr_tac (LevelLe ltactical) t),
               ltactical
-            | TacShowHyps t ->
-              hov 1 (
-                keyword "infoH" ++ spc ()
-                ++ pr_tac (LevelLe ltactical) t),
-              ltactical
             | TacOr (t1,t2) ->
               hov 1 (
                 pr_tac (LevelLt lorelse) t1 ++ spc ()
@@ -1036,7 +1029,7 @@ let pr_goal_selector ~toplevel s =
               keyword "solve" ++ spc () ++ pr_seq_body (pr_tac ltop) tl, llet
             | TacComplete t ->
               pr_tac (LevelLe lcomplete) t, lcomplete
-            | TacSelect (s, tac) -> pr_goal_selector ~toplevel:false s ++ spc () ++ pr_tac ltop tac, latom
+            | TacSelect (s, tac) -> pr_goal_selector ~toplevel:false s ++ spc () ++ pr_tac ltop tac, ltactical
             | TacId l ->
               keyword "idtac" ++ prlist (pr_arg (pr_message_token pr.pr_name)) l, latom
             | TacAtom t ->
@@ -1352,7 +1345,7 @@ let () =
     (make_constr_printer Printer.pr_econstr_n_env)
   ;
   Genprint.register_print0
-    wit_red_expr
+    Redexpr.wit_red_expr
     (lift_env (fun env sigma -> pr_red_expr env sigma (pr_constr_expr, pr_lconstr_expr, pr_or_by_notation pr_qualid, pr_constr_pattern_expr)))
     (lift_env (fun env sigma -> pr_red_expr env sigma
                   ((fun env sigma -> pr_and_constr_expr @@ pr_glob_constr_pptac env sigma),
