@@ -43,6 +43,18 @@ let open_constr_no_classes_flags =
   polymorphic = false;
   }
 
+let preterm_flags =
+  let open Pretyping in
+  {
+  use_coercions = true;
+  use_typeclasses = Pretyping.NoUseTC;
+  solve_unification_constraints = true;
+  fail_evar = false;
+  expand_evars = false;
+  program_mode = false;
+  polymorphic = false;
+  }
+
 (** Standard values *)
 
 module Value = Tac2ffi
@@ -93,6 +105,7 @@ let v_blk = Valexpr.make_block
 let of_relevance = function
   | Sorts.Relevant -> ValInt 0
   | Sorts.Irrelevant -> ValInt 1
+  | Sorts.RelevanceVar q -> ValInt 0 (* FIXME ? *)
 
 let to_relevance = function
   | ValInt 0 -> Sorts.Relevant
@@ -170,9 +183,9 @@ let has_fatal_flag info = match Exninfo.get info fatal_flag with
   | Some () -> true
 
 let set_bt info =
-  if !Tac2interp.print_ltac2_backtrace then
-    Tac2interp.get_backtrace >>= fun bt ->
-    Proofview.tclUNIT (Exninfo.add info Tac2entries.backtrace bt)
+  if !Tac2bt.print_ltac2_backtrace then
+    Tac2bt.get_backtrace >>= fun bt ->
+    Proofview.tclUNIT (Exninfo.add info Tac2bt.backtrace bt)
   else Proofview.tclUNIT info
 
 let throw ?(info = Exninfo.null) e =
@@ -185,7 +198,7 @@ let fail ?(info = Exninfo.null) e =
   Proofview.tclZERO ~info e
 
 let return x = Proofview.tclUNIT x
-let pname s = { mltac_plugin = "coq-core.plugins.ltac2"; mltac_tactic = s }
+let pname ?(plugin=ltac2_plugin) s = { mltac_plugin = plugin; mltac_tactic = s }
 
 let wrap f =
   return () >>= fun () -> return (f ())
@@ -233,10 +246,12 @@ let pf_apply ?(catch_exceptions=false) f =
 
 (** Primitives *)
 
-let define_primitive name arity f =
-  Tac2env.define_primitive (pname name) (mk_closure arity f)
+let define_primitive ?plugin name arity f =
+  Tac2env.define_primitive (pname ?plugin name) (mk_closure_val arity f)
 
-let define0 name f = define_primitive name arity_one (fun _ -> f)
+let defineval ?plugin name v = Tac2env.define_primitive (pname ?plugin name) v
+
+let define0 ?plugin name f = define_primitive ?plugin name arity_one (fun _ -> f)
 
 let define1 name r0 f = define_primitive name arity_one begin fun x ->
   f (Value.repr_to r0 x)
@@ -653,7 +668,7 @@ let () = define1 "constr_make" valexpr begin fun knd ->
     EConstr.mkLEvar sigma (evk, Array.to_list args)
   | (4, [|s|]) ->
     let s = Value.to_ext Value.val_sort s in
-    EConstr.mkSort (EConstr.Unsafe.to_sorts s)
+    EConstr.mkSort s
   | (5, [|c; k; t|]) ->
     let c = Value.to_constr c in
     let k = Value.to_ext Value.val_cast k in
@@ -774,6 +789,10 @@ let () = define1 "constr_case" (repr_ext val_inductive) begin fun ind ->
     throw err_notfound
 end
 
+let () = defineval "constr_cast_default" (of_cast DEFAULTcast)
+let () = defineval "constr_cast_vm" (of_cast VMcast)
+let () = defineval "constr_cast_native" (of_cast NATIVEcast)
+
 let () = define2 "constr_constructor" (repr_ext val_inductive) int begin fun (ind, i) k ->
   Proofview.tclENV >>= fun env ->
   try
@@ -805,7 +824,7 @@ let () = define3 "constr_in_context" ident constr closure begin fun id t c ->
         let t_ty = Retyping.get_type_of env sigma t in
         (* If the user passed eg ['_] for the type we force it to indeed be a type *)
         let sigma, j = Typing.type_judgment env sigma {uj_val=t; uj_type=t_ty} in
-        sigma, Sorts.relevance_of_sort j.utj_type
+        sigma, EConstr.ESorts.relevance_of_sort sigma j.utj_type
       in
       let nenv = EConstr.push_named (LocalAssum (Context.make_annot id t_rel, t)) env in
       let (sigma, (evt, _)) = Evarutil.new_type_evar nenv sigma Evd.univ_flexible in
@@ -836,7 +855,7 @@ let () = define1 "constr_pretype" (repr_ext val_preterm) begin fun c ->
         ltac_constrs = closure.typed;
         ltac_uconstrs = closure.untyped;
         ltac_idents = closure.idents;
-        ltac_genargs = Id.Map.empty;
+        ltac_genargs = closure.genargs;
       } in
       let flags = constr_flags in
       let sigma, t = understand_ltac flags env sigma vars WithoutTypeConstraint term in
@@ -884,6 +903,7 @@ let () = define_equality "evar_equal" evar Evar.equal
 let () = define_equality "float_equal" float Float64.equal
 let () = define_equality "uint63_equal" uint63 Uint63.equal
 let () = define_equality "meta_equal" int Int.equal
+let () = define_equality "constr_cast_equal" cast Glob_ops.cast_kind_eq
 
 let () = define_equality "constant_equal" constant Constant.UserOrd.equal
 let () = define_equality "constr_case_equal" (repr_ext val_case) begin fun x y ->
@@ -892,6 +912,8 @@ end
 let () = define_equality "constructor_equal" (repr_ext val_constructor) Construct.UserOrd.equal
 let () = define_equality "projection_equal" (repr_ext val_projection) Projection.UserOrd.equal
 
+
+
 (** Patterns *)
 
 let empty_context = Constr_matching.empty_context
@@ -1136,7 +1158,7 @@ let () = define1 "progress" closure begin fun f ->
 end
 
 let () = define2 "abstract" (option ident) closure begin fun id f ->
-  Abstract.tclABSTRACT id (Proofview.tclIGNORE (thaw f)) >>= fun () ->
+    Abstract.tclABSTRACT id (Proofview.tclIGNORE (thaw f)) >>= fun () ->
   return v_unit
 end
 
@@ -1344,6 +1366,279 @@ let () = define1 "ltac1_to_list" ltac1 begin fun v ->
   return (Value.of_option (Value.of_list of_ltac1) (Tacinterp.Value.to_list v))
 end
 
+module MapTagDyn = Dyn.Make()
+
+type ('a,'set,'map) map_tag = ('a * 'set * 'map) MapTagDyn.tag
+
+type any_map_tag = Any : _ map_tag -> any_map_tag
+type tagged_set = TaggedSet : (_,'set,_) map_tag * 'set -> tagged_set
+type tagged_map = TaggedMap : (_,_,'map) map_tag * 'map -> tagged_map
+
+let map_tag_ext : any_map_tag Tac2dyn.Val.tag = Tac2dyn.Val.create "fmap_tag"
+let map_tag_repr = Value.repr_ext map_tag_ext
+
+let set_ext : tagged_set Tac2dyn.Val.tag = Tac2dyn.Val.create "fset"
+let set_repr = Value.repr_ext set_ext
+let tag_set tag s = Value.repr_of set_repr (TaggedSet (tag,s))
+
+let map_ext : tagged_map Tac2dyn.Val.tag = Tac2dyn.Val.create "fmap"
+let map_repr = Value.repr_ext map_ext
+let tag_map tag m = Value.repr_of map_repr (TaggedMap (tag,m))
+
+module type MapType = sig
+  (* to have less boilerplate we use S.elt rather than declaring a toplevel type t *)
+  module S : CSig.SetS
+  module M : CMap.ExtS with type key = S.elt and module Set := S
+  type valmap
+  val valmap_eq : (valmap, valexpr M.t) Util.eq
+  val repr : S.elt Value.repr
+end
+
+module MapTypeV = struct
+  type _ t = Map : (module MapType with type S.elt = 't and type S.t = 'set and type valmap = 'map)
+    -> ('t * 'set * 'map) t
+end
+
+module MapMap = MapTagDyn.Map(MapTypeV)
+
+let maps = ref MapMap.empty
+
+let register_map ?(plugin=ltac2_plugin) ~tag_name x =
+  let tag = MapTagDyn.create (plugin^":"^tag_name) in
+  let () = maps := MapMap.add tag (Map x) !maps in
+  let () = defineval ~plugin tag_name (repr_of map_tag_repr (Any tag)) in
+  tag
+
+let get_map (type t s m) (tag:(t,s,m) map_tag)
+  : (module MapType with type S.elt = t and type S.t = s and type valmap = m) =
+  let Map v = MapMap.find tag !maps in
+  v
+
+let map_tag_eq (type a b c a' b' c') (t1:(a,b,c) map_tag) (t2:(a',b',c') map_tag)
+  : (a*b*c,a'*b'*c') Util.eq option
+  = MapTagDyn.eq t1 t2
+
+let assert_map_tag_eq t1 t2 = match map_tag_eq t1 t2 with
+  | Some v -> v
+  | None -> assert false
+
+let ident_map_tag : _ map_tag = register_map ~tag_name:"fmap_ident_tag" (module struct
+    module S = Id.Set
+    module M = Id.Map
+    let repr = Value.ident
+    type valmap = valexpr M.t
+    let valmap_eq = Refl
+  end)
+
+let int_map_tag : _ map_tag = register_map ~tag_name:"fmap_int_tag" (module struct
+    module S = Int.Set
+    module M = Int.Map
+    let repr = Value.int
+    type valmap = valexpr M.t
+    let valmap_eq = Refl
+  end)
+
+let string_map_tag : _ map_tag = register_map ~tag_name:"fmap_string_tag" (module struct
+    module S = String.Set
+    module M = String.Map
+    let repr = Value.string
+    type valmap = valexpr M.t
+    let valmap_eq = Refl
+  end)
+
+let inductive_map_tag : _ map_tag = register_map ~tag_name:"fmap_inductive_tag" (module struct
+    module S = Indset_env
+    module M = Indmap_env
+    let repr = Value.(repr_ext val_inductive)
+    type valmap = valexpr M.t
+    let valmap_eq = Refl
+  end)
+
+let constructor_map_tag : _ map_tag = register_map ~tag_name:"fmap_constructor_tag" (module struct
+    module S = Constrset_env
+    module M = Constrmap_env
+    let repr = Value.(repr_ext val_constructor)
+    type valmap = valexpr M.t
+    let valmap_eq = Refl
+  end)
+
+let constant_map_tag : _ map_tag = register_map ~tag_name:"fmap_constant_tag" (module struct
+    module S = Cset_env
+    module M = Cmap_env
+    let repr = Value.(repr_ext val_constant)
+    type valmap = valexpr M.t
+    let valmap_eq = Refl
+  end)
+
+let () = define1 "fset_empty" map_tag_repr begin fun (Any tag) ->
+    let module V = (val get_map tag) in
+    let open V in
+    return (tag_set tag S.empty)
+  end
+
+let () = define1 "fset_is_empty" set_repr begin fun (TaggedSet (tag,s)) ->
+    let module V = (val get_map tag) in
+    let open V in
+    return (Value.of_bool (S.is_empty s))
+  end
+
+let () = define2 "fset_mem" valexpr set_repr begin fun x (TaggedSet (tag,s)) ->
+    let module V = (val get_map tag) in
+    let open V in
+    let x = repr_to repr x in
+    return (Value.of_bool (S.mem x s))
+  end
+
+let () = define2 "fset_add" valexpr set_repr begin fun x (TaggedSet (tag,s)) ->
+    let module V = (val get_map tag) in
+    let open V in
+    let x = repr_to repr x in
+    return (tag_set tag (S.add x s))
+  end
+
+let () = define2 "fset_remove" valexpr set_repr begin fun x (TaggedSet (tag,s)) ->
+    let module V = (val get_map tag) in
+    let open V in
+    let x = repr_to repr x in
+    return (tag_set tag (S.remove x s))
+  end
+
+let () = define2 "fset_union" set_repr set_repr
+    begin fun (TaggedSet (tag,s1)) (TaggedSet (tag',s2)) ->
+      let Refl = assert_map_tag_eq tag tag' in
+      let module V = (val get_map tag) in
+      let open V in
+      return (tag_set tag (S.union s1 s2))
+    end
+
+let () = define2 "fset_inter" set_repr set_repr
+    begin fun (TaggedSet (tag,s1)) (TaggedSet (tag',s2)) ->
+      let Refl = assert_map_tag_eq tag tag' in
+      let module V = (val get_map tag) in
+      let open V in
+      return (tag_set tag (S.inter s1 s2))
+    end
+
+let () = define2 "fset_diff" set_repr set_repr
+    begin fun (TaggedSet (tag,s1)) (TaggedSet (tag',s2)) ->
+      let Refl = assert_map_tag_eq tag tag' in
+      let module V = (val get_map tag) in
+      let open V in
+      return (tag_set tag (S.diff s1 s2))
+    end
+
+let () = define2 "fset_equal" set_repr set_repr
+    begin fun (TaggedSet (tag,s1)) (TaggedSet (tag',s2)) ->
+      let Refl = assert_map_tag_eq tag tag' in
+      let module V = (val get_map tag) in
+      let open V in
+      return (Value.of_bool (S.equal s1 s2))
+    end
+
+let () = define2 "fset_subset" set_repr set_repr
+    begin fun (TaggedSet (tag,s1)) (TaggedSet (tag',s2)) ->
+      let Refl = assert_map_tag_eq tag tag' in
+      let module V = (val get_map tag) in
+      let open V in
+      return (Value.of_bool (S.subset s1 s2))
+    end
+
+let () = define1 "fset_cardinal" set_repr begin fun (TaggedSet (tag,s)) ->
+    let module V = (val get_map tag) in
+    let open V in
+    return (Value.of_int (S.cardinal s))
+  end
+
+let () = define1 "fset_elements" set_repr begin fun (TaggedSet (tag,s)) ->
+    let module V = (val get_map tag) in
+    let open V in
+    return (Value.of_list (repr_of repr) (S.elements s))
+  end
+
+let () = define1 "fmap_empty" map_tag_repr begin fun (Any (tag)) ->
+    let module V = (val get_map tag) in
+    let open V in
+    let Refl = valmap_eq in
+    return (tag_map tag M.empty)
+  end
+
+let () = define1 "fmap_is_empty" map_repr begin fun (TaggedMap (tag,m)) ->
+    let module V = (val get_map tag) in
+    let open V in
+    let Refl = valmap_eq in
+    return (Value.of_bool (M.is_empty m))
+  end
+
+let () = define2 "fmap_mem" valexpr map_repr begin fun x (TaggedMap (tag,m)) ->
+    let module V = (val get_map tag) in
+    let open V in
+    let Refl = valmap_eq in
+    let x = repr_to repr x in
+    return (Value.of_bool (M.mem x m))
+  end
+
+let () = define3 "fmap_add" valexpr valexpr map_repr begin fun x v (TaggedMap (tag,m)) ->
+    let module V = (val get_map tag) in
+    let open V in
+    let Refl = valmap_eq in
+    let x = repr_to repr x in
+    return (tag_map tag (M.add x v m))
+  end
+
+let () = define2 "fmap_remove" valexpr map_repr begin fun x (TaggedMap (tag,m)) ->
+    let module V = (val get_map tag) in
+    let open V in
+    let Refl = valmap_eq in
+    let x = repr_to repr x in
+    return (tag_map tag (M.remove x m))
+  end
+
+let () = define2 "fmap_find_opt" valexpr map_repr begin fun x (TaggedMap (tag,m)) ->
+    let module V = (val get_map tag) in
+    let open V in
+    let Refl = valmap_eq in
+    let x = repr_to repr x in
+    return (Value.of_option (fun v -> v) (M.find_opt x m))
+  end
+
+let () = define2 "fmap_mapi" closure map_repr begin fun f (TaggedMap (tag,m)) ->
+    let module V = (val get_map tag) in
+    let open V in
+    let Refl = valmap_eq in
+    let module Monadic = M.Monad(Proofview.Monad) in
+    Monadic.mapi (fun k v -> apply f [repr_of repr k;v]) m >>= fun m ->
+    return (tag_map tag m)
+  end
+
+let () = define3 "fmap_fold" closure map_repr valexpr begin fun f (TaggedMap (tag,m)) acc ->
+    let module V = (val get_map tag) in
+    let open V in
+    let Refl = valmap_eq in
+    let module Monadic = M.Monad(Proofview.Monad) in
+    Monadic.fold (fun k v acc -> apply f [repr_of repr k;v;acc]) m acc
+  end
+
+let () = define1 "fmap_cardinal" map_repr begin fun (TaggedMap (tag,m)) ->
+    let module V = (val get_map tag) in
+    let open V in
+    let Refl = valmap_eq in
+    return (Value.of_int (M.cardinal m))
+  end
+
+let () = define1 "fmap_bindings" map_repr begin fun (TaggedMap (tag,m)) ->
+    let module V = (val get_map tag) in
+    let open V in
+    let Refl = valmap_eq in
+    return (Value.(of_list (of_pair (repr_of repr) identity) (M.bindings m)))
+  end
+
+let () = define1 "fmap_domain" map_repr begin fun (TaggedMap (tag,m)) ->
+    let module V = (val get_map tag) in
+    let open V in
+    let Refl = valmap_eq in
+    return (tag_set tag (M.domain m))
+  end
+
 (** ML types *)
 
 (** Embed all Ltac2 data into Values *)
@@ -1356,7 +1651,19 @@ let gtypref kn = GTypRef (Other kn, [])
 
 let intern_constr self ist c =
   let (_, (c, _)) = Genintern.intern Stdarg.wit_constr ist c in
-  (GlbVal c, gtypref t_constr)
+  let v = match DAst.get c with
+    | GGenarg (GenArg (Glbwit tag, v)) ->
+      begin match genarg_type_eq tag wit_ltac2_var_quotation with
+      | Some Refl ->
+        begin match (fst v) with
+        | ConstrVar -> GlbTacexpr (GTacVar (snd v))
+        | _ -> GlbVal c
+        end
+      | None -> GlbVal c
+      end
+    | _ -> GlbVal c
+  in
+  (v, gtypref t_constr)
 
 let interp_constr flags ist c =
   let open Pretyping in
@@ -1409,8 +1716,8 @@ let () =
   let intern self ist c =
     let env = ist.Genintern.genv in
     let sigma = Evd.from_env env in
-    let warn = if !Ltac_plugin.Tacintern.strict_check then fun x -> x else Constrintern.for_grammar in
-    let _, pat = warn (fun () ->Constrintern.intern_constr_pattern env sigma ~as_type:false c) () in
+    let strict_check = ist.Genintern.strict_check in
+    let _, pat = Constrintern.intern_constr_pattern env sigma ~strict_check ~as_type:false c in
     GlbVal pat, gtypref t_pattern
   in
   let subst subst c =
@@ -1431,21 +1738,34 @@ let () =
 let () =
   let intern self ist c =
     let (_, (c, _)) = Genintern.intern Stdarg.wit_constr ist c in
-    (GlbVal c, gtypref t_preterm)
+    (GlbVal (Id.Set.empty,c), gtypref t_preterm)
   in
-  let interp _ c =
+  let interp env (ids,c) =
     let open Ltac_pretype in
+    let get_preterm id = match Id.Map.find_opt id env.env_ist with
+      | Some (ValExt (tag, v)) ->
+        begin match Tac2dyn.Val.eq tag val_preterm with
+          | Some Refl -> (v:closed_glob_constr)
+          | None -> assert false
+        end
+      | _ -> assert false
+    in
     let closure = {
       idents = Id.Map.empty;
       typed = Id.Map.empty;
-      untyped = Id.Map.empty;
-      genargs = Id.Map.empty;
+      untyped = Id.Map.bind get_preterm ids;
+      genargs = Tac2interp.set_env env Id.Map.empty;
     } in
     let c = { closure; term = c } in
     return (Value.of_ext val_preterm c)
   in
-  let subst subst c = Detyping.subst_glob_constr (Global.env()) subst c in
-  let print env sigma c = str "preterm:(" ++ Printer.pr_lglob_constr_env env sigma c ++ str ")" in
+  let subst subst (ids,c) = ids, Detyping.subst_glob_constr (Global.env()) subst c in
+  let print env sigma (ids,c) =
+    let ppids = if Id.Set.is_empty ids then mt()
+      else prlist_with_sep spc Id.print (Id.Set.elements ids) ++ spc() ++ str "|-" ++ spc()
+    in
+    str "preterm:(" ++ ppids ++ Printer.pr_lglob_constr_env env sigma c ++ str ")"
+  in
   let obj = {
     ml_intern = intern;
     ml_interp = interp;
@@ -1506,8 +1826,7 @@ let () =
       let ist = Ltac_plugin.Tacinterp.default_ist () in
       let ist = { ist with Geninterp.lfun = lfun } in
       let tac = (Ltac_plugin.Tacinterp.eval_tactic_ist ist tac : unit Proofview.tactic) in
-      let wrap (e, info) = set_bt info >>= fun info -> Proofview.tclZERO ~info e in
-      Proofview.tclOR tac wrap >>= fun () ->
+      tac >>= fun () ->
       return v_unit
     in
     let len = List.length ids in
@@ -1600,38 +1919,78 @@ let () =
   in
   GlobEnv.register_constr_interp0 wit_ltac2_constr interp
 
-let () =
-  let interp ?loc ~poly env sigma tycon id =
-    let ist = Tac2interp.get_env @@ GlobEnv.lfun env in
-    let env = GlobEnv.renamed_env env in
-    let c = Id.Map.find id ist.env_ist in
-    let c = Value.to_constr c in
-    let t = Retyping.get_type_of env sigma c in
-    let j = { Environ.uj_val = c; uj_type = t } in
-    match tycon with
-    | None ->
-      j, sigma
-    | Some ty ->
-      let sigma =
-        try Evarconv.unify_leq_delay env sigma t ty
-        with Evarconv.UnableToUnify (sigma,e) ->
-          Pretype_errors.error_actual_type ?loc env sigma j ty e
-      in
-      {j with Environ.uj_type = ty}, sigma
+let interp_constr_var_as_constr ?loc env sigma tycon id =
+  let ist = Tac2interp.get_env @@ GlobEnv.lfun env in
+  let env = GlobEnv.renamed_env env in
+  let c = Id.Map.find id ist.env_ist in
+  let c = Value.to_constr c in
+  let t = Retyping.get_type_of env sigma c in
+  let j = { Environ.uj_val = c; uj_type = t } in
+  match tycon with
+  | None ->
+    j, sigma
+  | Some ty ->
+    let sigma =
+      try Evarconv.unify_leq_delay env sigma t ty
+      with Evarconv.UnableToUnify (sigma,e) ->
+        Pretype_errors.error_actual_type ?loc env sigma j ty e
+    in
+    {j with Environ.uj_type = ty}, sigma
+
+let interp_preterm_var_as_constr ?loc env sigma tycon id =
+  let open Ltac_pretype in
+  let ist = Tac2interp.get_env @@ GlobEnv.lfun env in
+  let env = GlobEnv.renamed_env env in
+  let c = Id.Map.find id ist.env_ist in
+  let {closure; term} = Value.to_ext Value.val_preterm c in
+  let vars = {
+    ltac_constrs = closure.typed;
+    ltac_uconstrs = closure.untyped;
+    ltac_idents = closure.idents;
+    ltac_genargs = closure.genargs;
+  }
   in
-  GlobEnv.register_constr_interp0 wit_ltac2_quotation interp
+  let flags = preterm_flags in
+  let tycon = let open Pretyping in match tycon with
+    | Some ty -> OfType ty
+    | None -> WithoutTypeConstraint
+  in
+  let sigma, t, ty = Pretyping.understand_ltac_ty flags env sigma vars tycon term in
+  Environ.make_judge t ty, sigma
 
 let () =
-  let pr_raw id = Genprint.PrinterBasic (fun _env _sigma -> mt ()) in
-  let pr_glb id = Genprint.PrinterBasic (fun _env _sigma -> str "$" ++ Id.print id) in
-  let pr_top _ = Genprint.TopPrinterBasic mt in
-  Genprint.register_print0 wit_ltac2_quotation pr_raw pr_glb pr_top
+  let interp ?loc ~poly env sigma tycon (kind,id) =
+    let f = match kind with
+      | ConstrVar -> interp_constr_var_as_constr
+      | PretermVar -> interp_preterm_var_as_constr
+    in
+    f ?loc env sigma tycon id
+  in
+  GlobEnv.register_constr_interp0 wit_ltac2_var_quotation interp
 
 let () =
-  let subs globs (ids, tac) =
+  let pr_raw (kind,id) = Genprint.PrinterBasic (fun _env _sigma ->
+      let ppkind =
+      match kind with
+        | None -> mt()
+        | Some kind -> Id.print kind.CAst.v ++ str ":"
+      in
+      str "$" ++ ppkind ++ Id.print id.CAst.v)
+  in
+  let pr_glb (kind,id) = Genprint.PrinterBasic (fun _env _sigma ->
+      let ppkind = match kind with
+        | ConstrVar -> mt()
+        | PretermVar -> str "preterm:"
+      in
+      str "$" ++ ppkind ++ Id.print id) in
+  let pr_top x = Util.Empty.abort x in
+  Genprint.register_print0 wit_ltac2_var_quotation pr_raw pr_glb pr_top
+
+let () =
+  let subs avoid globs (ids, tac) =
     (* Let-bind the notation terms inside the tactic *)
-    let fold id (c, _) (rem, accu) =
-      let c = GTacExt (Tac2quote.wit_preterm, c) in
+    let fold id c (rem, accu) =
+      let c = GTacExt (Tac2quote.wit_preterm, (avoid, c)) in
       let rem = Id.Set.remove id rem in
       rem, (Name id, c) :: accu
     in
@@ -1646,21 +2005,29 @@ let () =
           str ", probably an ill-typed expression")
     in
     let tac = if List.is_empty bnd then tac else GTacLet (false, bnd, tac) in
-    (Id.Set.empty, tac)
+    (avoid, tac)
   in
   Genintern.register_ntn_subst0 wit_ltac2_constr subs
 
 (** Ltac2 in Ltac1 *)
 
 let () =
-  let e = Tac2entries.Pltac.tac2expr_in_env in
-  let inject (loc, v) = Ltac_plugin.Tacexpr.TacGeneric (Some "ltac2", in_gen (rawwit wit_ltac2) v) in
-  Ltac_plugin.Tacentries.create_ltac_quotation ~plugin:ltac2_plugin "ltac2" inject (e, None)
+  let create name wit =
+    let e = Tac2entries.Pltac.tac2expr_in_env in
+    let inject (loc, v) = Ltac_plugin.Tacexpr.TacGeneric (Some name, in_gen (rawwit wit) v) in
+    Ltac_plugin.Tacentries.create_ltac_quotation ~plugin:ltac2_plugin name inject (e, None)
+  in
+  let () = create "ltac2" wit_ltac2in1 in
+  let () = create "ltac2val" wit_ltac2in1_val in
+  ()
 
 (* Ltac1 runtime representation of Ltac2 closures. *)
 let typ_ltac2 : valexpr Geninterp.Val.typ =
   Geninterp.Val.create "ltac2:ltac2_eval"
 
+let () = Genprint.register_val_print0 typ_ltac2 (fun v ->
+    TopPrinterBasic (fun () -> Pp.str "<ltac2 closure>"))
+
 let cast_typ (type a) (tag : a Geninterp.Val.typ) (v : Geninterp.Val.t) : a =
   let Geninterp.Val.Dyn (tag', v) = v in
   match Geninterp.Val.eq tag tag' with
@@ -1687,7 +2054,7 @@ let () =
   in
   let () = Geninterp.register_interp0 wit_ltac2_val interp_fun in
   define1 "ltac1_lambda" valexpr begin fun f ->
-    let body = Tacexpr.TacGeneric (Some "coq-core.plugins.ltac2", in_gen (glbwit wit_ltac2_val) ()) in
+    let body = Tacexpr.TacGeneric (Some ltac2_plugin, in_gen (glbwit wit_ltac2_val) ()) in
     let clos = CAst.make (Tacexpr.TacFun ([Name arg_id], CAst.make (Tacexpr.TacArg body))) in
     let f = Geninterp.Val.inject (Geninterp.Val.Base typ_ltac2) f in
     let lfun = Id.Map.singleton tac_id f in
@@ -1698,7 +2065,7 @@ let () =
 let ltac2_eval =
   let open Ltac_plugin in
   let ml_name = {
-    Tacexpr.mltac_plugin = "coq-core.plugins.ltac2";
+    Tacexpr.mltac_plugin = ltac2_plugin;
     mltac_tactic = "ltac2_eval";
   } in
   let eval_fun args ist = match args with
@@ -1723,8 +2090,9 @@ let () =
     let idtac = Value.of_closure { ist with lfun = Id.Map.empty }
         (CAst.make (Tacexpr.TacId [])) in
     let ist = { env_ist = Id.Map.empty } in
-    Tac2interp.interp ist tac >>= fun _ ->
-    Ftactic.return idtac
+    Tac2interp.interp ist tac >>= fun v ->
+    let v = idtac in
+    Ftactic.return v
   | _ :: _ ->
     (* Return a closure [@f := {blob} |- fun ids => ltac2_eval(f, ids) ] *)
     (* This name cannot clash with Ltac2 variables which are all lowercase *)
@@ -1740,10 +2108,19 @@ let () =
     let ist = { ist with lfun = Id.Map.singleton self_id self } in
     Ftactic.return (Value.of_closure ist clos)
   in
-  Geninterp.register_interp0 wit_ltac2 interp
+  Geninterp.register_interp0 wit_ltac2in1 interp
 
 let () =
-  let pr_raw _ = Genprint.PrinterBasic (fun _env _sigma -> mt ()) in
+  let interp ist tac =
+    let ist = { env_ist = Id.Map.empty } in
+    Tac2interp.interp ist tac >>= fun v ->
+    let v = repr_to ltac1 v in
+    Ftactic.return v
+  in
+  Geninterp.register_interp0 wit_ltac2in1_val interp
+
+let () =
+  let pr_raw _ = Genprint.PrinterBasic (fun _env _sigma -> assert false) in
   let pr_glb (ids, e) =
     let ids =
       if List.is_empty ids then mt ()
@@ -1751,8 +2128,21 @@ let () =
     in
     Genprint.PrinterBasic Pp.(fun _env _sigma -> ids ++ Tac2print.pr_glbexpr e)
   in
-  let pr_top _ = Genprint.TopPrinterBasic mt in
-  Genprint.register_print0 wit_ltac2 pr_raw pr_glb pr_top
+  let pr_top x = Util.Empty.abort x in
+  Genprint.register_print0 wit_ltac2in1 pr_raw pr_glb pr_top
+
+let () =
+  let pr_raw _ = Genprint.PrinterBasic (fun _env _sigma -> assert false) in
+  let pr_glb (ids, e) =
+    let ids =
+      let ids = Id.Set.elements ids in
+      if List.is_empty ids then mt ()
+      else pr_sequence Id.print ids ++ str " |- "
+    in
+    Genprint.PrinterBasic Pp.(fun _env _sigma -> ids ++ Tac2print.pr_glbexpr e)
+  in
+  let pr_top e = Util.Empty.abort e in
+  Genprint.register_print0 wit_ltac2_constr pr_raw pr_glb pr_top
 
 (** Built-in notation scopes *)
 
@@ -1796,7 +2186,7 @@ end
 
 let () = add_scope "terminal" begin function
 | [SexprStr {loc;v=s}] ->
-  let scope = Pcoq.Symbol.token (CLexer.terminal s) in
+  let scope = Pcoq.Symbol.token (Pcoq.terminal s) in
   Tac2entries.ScopeRule (scope, (fun _ -> q_unit))
 | arg -> scope_fail "terminal" arg
 end
@@ -1809,7 +2199,7 @@ let () = add_scope "list0" begin function
   Tac2entries.ScopeRule (scope, act)
 | [tok; SexprStr {v=str}] ->
   let Tac2entries.ScopeRule (scope, act) = Tac2entries.parse_scope tok in
-  let sep = Pcoq.Symbol.tokens [Pcoq.TPattern (CLexer.terminal str)] in
+  let sep = Pcoq.Symbol.tokens [Pcoq.TPattern (Pcoq.terminal str)] in
   let scope = Pcoq.Symbol.list0sep scope sep false in
   let act l = Tac2quote.of_list act l in
   Tac2entries.ScopeRule (scope, act)
@@ -1824,7 +2214,7 @@ let () = add_scope "list1" begin function
   Tac2entries.ScopeRule (scope, act)
 | [tok; SexprStr {v=str}] ->
   let Tac2entries.ScopeRule (scope, act) = Tac2entries.parse_scope tok in
-  let sep = Pcoq.Symbol.tokens [Pcoq.TPattern (CLexer.terminal str)] in
+  let sep = Pcoq.Symbol.tokens [Pcoq.TPattern (Pcoq.terminal str)] in
   let scope = Pcoq.Symbol.list1sep scope sep false in
   let act l = Tac2quote.of_list act l in
   Tac2entries.ScopeRule (scope, act)
