@@ -120,6 +120,11 @@ let of_open_constr ?delimiters c =
   let c = quote_constr ?delimiters c in
   inj_wit ?loc wit_open_constr c
 
+let of_preterm ?delimiters c =
+  let loc = Constrexpr_ops.constr_loc c in
+  let c = quote_constr ?delimiters c in
+  inj_wit ?loc wit_preterm c
+
 let of_bool ?loc b =
   let c = if b then coq_core "true" else coq_core "false" in
   constructor ?loc c []
@@ -129,6 +134,10 @@ let rec of_list ?loc f = function
 | e :: l ->
   constructor ?loc (coq_core "::") [f e; of_list ?loc f l]
 
+let array_of_list ?loc l =
+  let of_list = global_ref ?loc (kername array_prefix "of_list")  in
+  CAst.make ?loc @@ CTacApp (of_list, [l])
+
 let of_qhyp {loc;v=h} = match h with
 | QAnonHyp n -> std_constructor ?loc "AnonHyp" [of_int n]
 | QNamedHyp id -> std_constructor ?loc "NamedHyp" [of_ident id]
@@ -206,7 +215,7 @@ let of_hyp_location ?loc ((occs, id), flag) =
 let of_clause {loc;v=cl} =
   let hyps = of_option ?loc (fun l -> of_list ?loc of_hyp_location l) cl.q_onhyps in
   let concl = of_occurrences cl.q_concl_occs in
-  CAst.make ?loc @@ CTacRec ([
+  CAst.make ?loc @@ CTacRec (None, [
     std_proj "on_hyps", hyps;
     std_proj "on_concl", concl;
   ])
@@ -223,7 +232,7 @@ let of_induction_clause {loc;v=cl} =
   let eqn = of_option ?loc of_intro_pattern_naming cl.indcl_eqn in
   let as_ = of_option ?loc of_or_and_intro_pattern cl.indcl_as in
   let in_ = of_option ?loc of_clause cl.indcl_in in
-  CAst.make ?loc @@ CTacRec ([
+  CAst.make ?loc @@ CTacRec (None, [
     std_proj "indcl_arg", arg;
     std_proj "indcl_eqn", eqn;
     std_proj "indcl_as", as_;
@@ -251,7 +260,7 @@ let abstract_vars loc ?typ vars tac =
   | Name id -> Some (CAst.make ?loc:na.CAst.loc id)
   | Anonymous -> None
   in
-  let def = try Some (List.find_map get_name vars) with Not_found -> None in
+  let def = List.find_map get_name vars in
   let na, tac = match def with
   | None -> (Anonymous, tac)
   | Some id0 ->
@@ -315,7 +324,7 @@ let of_rewriting {loc;v=rew} =
   in
   let repeat = of_repeat rew.rew_repeat in
   let equatn = thunk (of_constr_with_bindings rew.rew_equatn) in
-  CAst.make ?loc @@ CTacRec ([
+  CAst.make ?loc @@ CTacRec (None, [
     std_proj "rew_orient", orient;
     std_proj "rew_repeat", repeat;
     std_proj "rew_equatn", equatn;
@@ -382,7 +391,7 @@ let of_reference r =
 let of_strategy_flag {loc;v=flag} =
   let open Genredexpr in
   let flag = make_red_flag flag in
-  CAst.make ?loc @@ CTacRec ([
+  CAst.make ?loc @@ CTacRec (None, [
     std_proj "rBeta", of_bool ?loc flag.rBeta;
     std_proj "rMatch", of_bool ?loc flag.rMatch;
     std_proj "rFix", of_bool ?loc flag.rFix;
@@ -452,27 +461,45 @@ let of_goal_matching {loc;v=gm} =
     let hyps_pats = p.q_goal_match_hyps in
     let (concl_ctx, concl_pat, concl_knd) = mk_pat concl_pat in
     let vars = pattern_vars concl_pat in
-    let map accu (na, pat) =
-      let (ctx, pat, knd) = mk_pat pat in
-      let vars = pattern_vars pat in
-      (Id.Map.fold Id.Map.add vars accu, (na, ctx, pat, knd))
+    let map accu (na, bod, pat) = match bod with
+      | None ->
+        let (ctx, pat, knd) = mk_pat pat in
+        let vars = pattern_vars pat in
+        (Id.Map.fold Id.Map.add vars accu, (na, None, (ctx, knd, pat)))
+      | Some bod ->
+        let bctx, bpat, bknd = mk_pat bod in
+        let (ctx, pat, knd) = mk_pat pat in
+        let bvars = pattern_vars bpat in
+        let vars = pattern_vars pat in
+        let accu = Id.Map.fold Id.Map.add bvars accu in
+        let accu = Id.Map.fold Id.Map.add vars accu in
+        (accu, (na, Some ((bctx, bknd, bpat)), (ctx, knd, pat)))
     in
     let (vars, hyps_pats) = List.fold_left_map map vars hyps_pats in
-    let map (_, _, pat, knd) = of_tuple [knd; of_pattern pat] in
     let concl = of_tuple [concl_knd; of_pattern concl_pat] in
-    let r = of_tuple [of_list ?loc map hyps_pats; concl] in
-    let hyps = List.map (fun (na, _, _, _) -> na) hyps_pats in
-    let hctx = List.map (fun (_, na, _, _) -> na) hyps_pats in
+    let of_hyp_pat (_, bpat, (_, knd, pat)) =
+      let bpat = Option.map (fun (_, a, b) -> (a, b)) bpat in
+      let of_mpat (a, b) = of_tuple [a; of_pattern b] in
+      of_tuple [of_option of_mpat bpat; of_mpat (knd, pat)]
+    in
+    let r = of_tuple [
+        of_list ?loc of_hyp_pat hyps_pats;
+        concl]
+    in
+    let hyps = List.map (fun (na, _, _) -> na) hyps_pats in
+    let hbctx = CList.filter_map (fun (_, b, _) -> Option.map pi1 b) hyps_pats in
+    let hctx = CList.map (fun (_, _, (na, _, _)) -> na) hyps_pats in
     (* Order of elements is crucial here! *)
     let vars = Id.Map.bindings vars in
     let subst = List.map (fun (id, loc) -> CAst.make ?loc (Name id)) vars in
-    (r, hyps, hctx, subst, concl_ctx)
+    (r, hyps, hbctx, hctx, subst, concl_ctx)
   in
   let map {loc;v=(pat, tac)} =
-    let (pat, hyps, hctx, subst, cctx) = mk_gpat pat in
+    let (pat, hyps, hbctx, hctx, subst, cctx) = mk_gpat pat in
     let tac = CAst.make ?loc @@ CTacFun ([CAst.make ?loc @@ CPatVar cctx.CAst.v], tac) in
     let tac = abstract_vars loc subst tac in
     let tac = abstract_vars loc hctx tac in
+    let tac = abstract_vars loc hbctx tac in
     let tac = abstract_vars loc hyps tac in
     of_tuple ?loc [pat; tac]
   in
