@@ -29,8 +29,11 @@ open Mutils
 
 (* If set to some [file], arithmetic goals are dumped in [file].v *)
 
-let dump_file =
-  Goptions.declare_stringopt_option_and_ref ~depr:false ~key:["Dump"; "Arith"]
+let { Goptions.get = dump_file } =
+  Goptions.declare_stringopt_option_and_ref
+    ~key:["Dump"; "Arith"]
+    ~value:None
+    ()
 
 type ('prf, 'model) res = Prf of 'prf | Model of 'model | Unknown
 type zres = (Mc.zArithProof, int * Mc.z list) res
