@@ -257,9 +257,9 @@ module VNativeEntries =
       let (ct,cf) = get_bool_constructors env in
       CONSTR(Univ.in_punivs (if b then ct else cf), [||])
 
-    let int_ty env = VAL(0, mkConst @@ get_int_type env)
+    let int_ty env = VAL(0, UnsafeMonomorphic.mkConst @@ get_int_type env)
 
-    let float_ty env = VAL(0, mkConst @@ get_float_type env)
+    let float_ty env = VAL(0, UnsafeMonomorphic.mkConst @@ get_float_type env)
 
     let mkCarry env b e =
       let (c0,c1) = get_carry_constructors env in
@@ -500,17 +500,18 @@ let rec norm_head info env t stack =
       else
         (CBN(t,env), stack) (* Should we consider a commutative cut ? *)
 
-  | Evar ev ->
+  | Evar ((e, _) as ev) ->
       (match Evd.existential_opt_value0 info.sigma ev with
           Some c -> norm_head info env c stack
         | None ->
-          let e, xs = ev in
-          let xs' = List.map (apply_env env) xs in
-          (VAL(0, mkEvar (e,xs')), stack))
+          let ev = EConstr.of_existential ev in
+          let map c = EConstr.of_constr @@ apply_env env (EConstr.Unsafe.to_constr c) in
+          let ev' = EConstr.map_existential info.sigma map ev in
+          (VAL(0, EConstr.Unsafe.to_constr @@ EConstr.mkEvar ev'), stack))
 
   (* non-neutral cases *)
   | Lambda _ ->
-      let ctxt,b = Term.decompose_lam t in
+      let ctxt,b = Term.decompose_lambda t in
       (LAM(List.length ctxt, List.rev ctxt,b,env), stack)
   | Fix fix -> (FIXP(fix,env,[||]), stack)
   | CoFix cofix -> (COFIXP(cofix,env,[||]), stack)
@@ -689,9 +690,9 @@ let rec apply_stack info t = function
     let (_, ty, _, _, br) = Inductive.expand_case info.env (ci, u, pms, ty, iv, mkProp, br) in
     let ty =
       let (_, mip) = Inductive.lookup_mind_specif info.env ci.ci_ind in
-      Term.decompose_lam_n_decls (mip.Declarations.mind_nrealdecls + 1) ty
+      Term.decompose_lambda_n_decls (mip.Declarations.mind_nrealdecls + 1) ty
     in
-    let mk_br c n = Term.decompose_lam_n_decls n c in
+    let mk_br c n = Term.decompose_lambda_n_decls n c in
     let br = Array.map2 mk_br br ci.ci_cstr_ndecls in
     let map_ctx (nas, c) =
       let open Context.Rel.Declaration in
@@ -760,7 +761,7 @@ and cbv_norm_value info = function (* reduction under binders *)
 (* with profiling *)
 let cbv_norm infos constr =
   let constr = EConstr.Unsafe.to_constr constr in
-  EConstr.of_constr (with_stats (lazy (cbv_norm_term infos (subs_id 0) constr)))
+  EConstr.of_constr (cbv_norm_term infos (subs_id 0) constr)
 
 (* constant bodies are normalized at the first expansion *)
 let create_cbv_infos reds env sigma =
