@@ -34,11 +34,11 @@ open Evarconv
 open Evd
 open Globnames
 
-let get_use_typeclasses_for_conversion =
+let { Goptions.get = get_use_typeclasses_for_conversion } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["Typeclass"; "Resolution"; "For"; "Conversion"]
     ~value:true
+    ()
 
 (* Typing operations dealing with coercions *)
 exception NoCoercion
@@ -109,14 +109,14 @@ let app_opt env sigma f t =
 
 let pair_of_array a = (a.(0), a.(1))
 
-let disc_subset sigma x =
+let disc_subset env sigma x =
   match EConstr.kind sigma x with
   | App (c, l) ->
       (match EConstr.kind sigma c with
        Ind (i,_) ->
          let len = Array.length l in
          let sigty = delayed_force sig_typ in
-           if Int.equal len 2 && Ind.CanOrd.equal i (Globnames.destIndRef sigty)
+           if Int.equal len 2 && QInd.equal env i (Globnames.destIndRef sigty)
            then
              let (a, b) = pair_of_array l in
                Some (a, b)
@@ -148,8 +148,8 @@ let coerce ?loc env sigma (x : EConstr.constr) (y : EConstr.constr)
       Evarconv.UnableToUnify _ -> coerce' env sigma x y
   and coerce' env sigma x y : evar_map * (evar_map -> EConstr.constr -> evar_map * EConstr.constr) option =
     let subco sigma = subset_inferred env sigma x y in
-    let dest_prod c =
-      match Reductionops.splay_prod_n env sigma 1 c with
+    let hnf_decompose_prod c =
+      match Reductionops.hnf_decompose_prod_n_decls env sigma 1 c with
       | [LocalAssum (na,t) | LocalDef (na,_,t)], c -> (na, t), c
       | _ -> raise NoSubtacCoercion
     in
@@ -160,13 +160,13 @@ let coerce ?loc env sigma (x : EConstr.constr) (y : EConstr.constr)
           let hdx = l.(i) and hdy = l'.(i) in
           try
             let sigma = unify_leq_delay env sigma hdx hdy in
-            let (n, eqT), restT = dest_prod typ in
-            let (n', eqT'), restT' = dest_prod typ' in
+            let (n, eqT), restT = hnf_decompose_prod typ in
+            let (n', eqT'), restT' = hnf_decompose_prod typ' in
             aux sigma (hdx :: tele) (subst1 hdx restT) (subst1 hdy restT') (succ i) co
           with UnableToUnify _ as exn ->
             let _, info = Exninfo.capture exn in
-            let (n, eqT), restT = dest_prod typ in
-            let (n', eqT'), restT' = dest_prod typ' in
+            let (n, eqT), restT = hnf_decompose_prod typ in
+            let (n', eqT'), restT' = hnf_decompose_prod typ' in
             let sigma =
               try
                 unify_leq_delay env sigma eqT eqT'
@@ -237,10 +237,10 @@ let coerce ?loc env sigma (x : EConstr.constr) (y : EConstr.constr)
          let sigT = delayed_force sigT_typ in
          let prod = delayed_force prod_typ in
          (* Sigma types *)
-         if Int.equal len (Array.length l') && Int.equal len 2 && Ind.CanOrd.equal i i'
-            && (Ind.CanOrd.equal i (destIndRef sigT) || Ind.CanOrd.equal i (destIndRef prod))
+         if Int.equal len (Array.length l') && Int.equal len 2 && QInd.equal env i i'
+            && (QInd.equal env i (destIndRef sigT) || QInd.equal env i (destIndRef prod))
          then
-           if Ind.CanOrd.equal i (destIndRef sigT)
+           if QInd.equal env i (destIndRef sigT)
            then
              begin
                let (a, pb), (a', pb') =
@@ -300,7 +300,7 @@ let coerce ?loc env sigma (x : EConstr.constr) (y : EConstr.constr)
                      papp env sigma prod_intro [| a'; b'; x ; y |])
              end
          else
-         if Ind.CanOrd.equal i i' && Int.equal len (Array.length l') then
+         if QInd.equal env i i' && Int.equal len (Array.length l') then
            (try subco sigma
             with NoSubtacCoercion ->
               let sigma, typ = Typing.type_of env sigma c in
@@ -318,7 +318,7 @@ let coerce ?loc env sigma (x : EConstr.constr) (y : EConstr.constr)
     | _, _ ->  subco sigma
 
   and subset_inferred env sigma x y =
-    match disc_subset sigma x with
+    match disc_subset env sigma x with
       Some (u, p) ->
       let sigma, c = coerce_unify env sigma u y in
       let f sigma x =
@@ -326,7 +326,7 @@ let coerce ?loc env sigma (x : EConstr.constr) (y : EConstr.constr)
         app_opt env sigma c t
       in sigma, Some f
     | None ->
-      match disc_subset sigma y with
+      match disc_subset env sigma y with
         Some (u, p) ->
         let sigma, c = coerce_unify env sigma x u in
         sigma, Some
@@ -353,7 +353,7 @@ let coerce_itf ?loc env sigma v t c1 =
 
 let saturate_evd env sigma =
   Typeclasses.resolve_typeclasses
-    ~filter:Typeclasses.no_goals ~split:true ~fail:false env sigma
+    ~filter:Typeclasses.no_goals ~fail:false env sigma
 
 type coercion_trace =
   | IdCoe
@@ -422,7 +422,7 @@ let apply_coercion env sigma p h hty =
 let remove_subset env sigma t =
   let rec aux v =
     let v' = hnf env sigma v in
-    match disc_subset sigma v' with
+    match disc_subset env sigma v' with
     | Some (u, p) ->
       aux u
     | None -> v
@@ -432,7 +432,7 @@ let remove_subset env sigma t =
 let mu env sigma t =
   let rec aux v =
     let v' = hnf env sigma v in
-      match disc_subset sigma v' with
+      match disc_subset env sigma v' with
       | Some (u, p) ->
         let sigma, (f, ct, trace) = aux u in
         let p = hnf_nodelta env sigma p in
@@ -455,12 +455,14 @@ let unify_product env sigma typ =
     let sigma, _ = Evardefine.define_evar_as_product env sigma ev in
     Inl sigma
   | _ ->
-    let sigma, (domain_hole, _) = Evarutil.new_type_evar env sigma Evd.univ_flexible in
-    let env' = push_rel Context.(Rel.Declaration.LocalAssum (anonR, domain_hole)) env in
+    let sigma, (domain_hole, s) = Evarutil.new_type_evar env sigma Evd.univ_flexible in
+    let r = ESorts.relevance_of_sort sigma s in
+    let na = make_annot Anonymous r in
+    let env' = push_rel Context.(Rel.Declaration.LocalAssum (na, domain_hole)) env in
     let sigma, (codomain_hole, _) = Evarutil.new_type_evar env' sigma Evd.univ_flexible in
-    let prod = mkProd (Context.anonR, domain_hole, codomain_hole) in
+    let prod = mkProd (na, domain_hole, codomain_hole) in
     (* NB: unification needs the un-reduced type to do heuristics like canonical structures *)
-    try Inl (Evarconv.unify env sigma Reduction.CUMUL typ prod)
+    try Inl (Evarconv.unify env sigma Conversion.CUMUL typ prod)
     with PretypeError _ -> Inr t (* return the reduced type to avoid double reducing *)
 
 (* Invariant: if [proj] is [Some] then [args_len < npars]
@@ -555,7 +557,7 @@ let inh_app_fun ~program_mode ~resolve_tc ?use_coercions env sigma body typ =
 
 let type_judgment env sigma j =
   match EConstr.kind sigma (whd_all env sigma j.uj_type) with
-    | Sort s -> {utj_val = j.uj_val; utj_type = ESorts.kind sigma s }
+    | Sort s -> {utj_val = j.uj_val; utj_type = s }
     | _ -> error_not_a_type env sigma j
 
 let inh_tosort_force ?loc env sigma ({ uj_val; uj_type } as j) =
@@ -570,7 +572,7 @@ let inh_tosort_force ?loc env sigma ({ uj_val; uj_type } as j) =
 let inh_coerce_to_sort ?loc ?(use_coercions=true) env sigma j =
   let typ = whd_all env sigma j.uj_type in
     match EConstr.kind sigma typ with
-    | Sort s -> (sigma,{ utj_val = j.uj_val; utj_type = ESorts.kind sigma s })
+    | Sort s -> (sigma,{ utj_val = j.uj_val; utj_type = s })
     | Evar ev ->
         let (sigma,s) = Evardefine.define_evar_as_sort env sigma ev in
           (sigma,{ utj_val = j.uj_val; utj_type = s })
@@ -620,8 +622,45 @@ let lookup_reversible_path_to_common_point env sigma ~src_expected ~src_inferred
   in
     aux r
 
+type hook = env -> evar_map -> flags:Evarconv.unify_flags -> constr ->
+  inferred:types -> expected:types -> (evar_map * constr) option
+
+let all_hooks = ref (CString.Map.empty : hook CString.Map.t)
+
+let register_hook ~name ?(override=false) h =
+  if not override && CString.Map.mem name !all_hooks then
+    CErrors.anomaly ~label:"Coercion.register_hook"
+      Pp.(str "Hook already registered: \"" ++ str name ++ str "\".");
+  all_hooks := CString.Map.add name h !all_hooks
+
+let active_hooks = Summary.ref ~name:"coercion_hooks" ([] : string list)
+
+let deactivate_hook ~name =
+  active_hooks := List.filter (fun s -> not (String.equal s name)) !active_hooks
+
+let activate_hook ~name =
+  assert (CString.Map.mem name !all_hooks);
+  deactivate_hook ~name;
+  active_hooks := name :: !active_hooks
+
+let active_hooks () =
+  List.map (fun name -> CString.Map.get name !all_hooks) !active_hooks
+
+let add_reverse_coercion env sigma v'_ty v_ty v' v =
+  match Coqlib.lib_ref_opt "core.coercion.reverse_coercion" with
+  | None -> sigma, v'
+  | Some reverse_coercion ->
+     let sigma, v'_ty =
+       Evarsolve.refresh_universes ~onlyalg:true ~status:Evd.univ_flexible
+         (Some false) env sigma v'_ty in
+     let sigma, v_ty =
+       Evarsolve.refresh_universes ~onlyalg:true ~status:Evd.univ_flexible
+         (Some false) env sigma v_ty in
+     let sigma, reverse_coercion = Evd.fresh_global env sigma reverse_coercion in
+     Typing.checked_appvect env sigma reverse_coercion [| v'_ty; v_ty; v'; v |]
+
 let inh_coerce_to_fail ?(use_coercions=true) flags env sigma rigidonly v v_ty target_type =
-  if not use_coercions || (rigidonly && not (Heads.is_rigid env (EConstr.Unsafe.to_constr target_type) && Heads.is_rigid env (EConstr.Unsafe.to_constr v_ty)))
+  if not use_coercions || (rigidonly && not (Heads.is_rigid env sigma target_type && Heads.is_rigid env sigma v_ty))
   then
     raise NoCoercion
   else
@@ -642,12 +681,22 @@ let inh_coerce_to_fail ?(use_coercions=true) flags env sigma rigidonly v v_ty ta
           let sigma = unify_leq_delay ~flags env sigma direct_v rev_x in
           (try let _ = Evarutil.head_evar sigma (whd_evar sigma x) in raise Not_found
            with NoHeadEvar -> ());  (* fail if x is stil an unresolved evar *)
-          sigma, x, target_type, ReplaceCoe x
+          let sigma, rev_x = add_reverse_coercion env sigma target_type v_ty x v in
+          sigma, rev_x, target_type, ReplaceCoe rev_x
       in
       unify_leq_delay ~flags env sigma v'_ty target_type, v', trace
     with (Evarconv.UnableToUnify _ | Not_found) as exn ->
       let _, info = Exninfo.capture exn in
-      Exninfo.iraise (NoCoercion,info)
+      (* 3 if none of the above works, try hook *)
+      let hook_res =
+        List.fold_left
+          (fun r h ->
+            if r <> None then r else
+              h env sigma ~flags v ~inferred:v_ty ~expected:target_type)
+          None (active_hooks ()) in
+      match hook_res with
+      | Some (sigma, r) -> (sigma, r, ReplaceCoe r)
+      | None -> Exninfo.iraise (NoCoercion,info)
 
 let default_flags_of env =
   default_flags_of TransparentState.full
@@ -686,7 +735,7 @@ let rec inh_conv_coerce_to_fail ?loc ?use_coercions env sigma ?(flags=default_fl
         Exninfo.iraise (NoCoercionNoUnifier (best_failed_sigma,e), info)
 
 (* Look for cj' obtained from cj by inserting coercions, s.t. cj'.typ = t *)
-let inh_conv_coerce_to_gen ?loc ~program_mode ~resolve_tc ?use_coercions rigidonly flags env sigma cj t =
+let inh_conv_coerce_to_gen ?loc ~program_mode ~resolve_tc ?use_coercions rigidonly env sigma ?(flags=default_flags_of env) cj t =
   let (sigma, val', otrace) =
     try
       let (sigma, val', trace) = inh_conv_coerce_to_fail ?loc ?use_coercions env sigma ~flags rigidonly cj.uj_val cj.uj_type t in
@@ -718,7 +767,15 @@ let inh_conv_coerce_to_gen ?loc ~program_mode ~resolve_tc ?use_coercions rigidon
   in
   (sigma,{ uj_val = val'; uj_type = t },otrace)
 
-let inh_conv_coerce_to ?loc ~program_mode ~resolve_tc ?use_coercions env sigma ?(flags=default_flags_of env) =
-  inh_conv_coerce_to_gen ?loc ~program_mode ~resolve_tc ?use_coercions false flags env sigma
-let inh_conv_coerce_rigid_to ?loc ~program_mode ~resolve_tc ?use_coercions env sigma ?(flags=default_flags_of env) =
-  inh_conv_coerce_to_gen ?loc ~program_mode ~resolve_tc ?use_coercions true flags env sigma
+let inh_conv_coerce_to_gen ?loc ~program_mode ~resolve_tc ?use_coercions rigidonly env sigma ?flags cj t =
+  try inh_conv_coerce_to_gen ?loc ~program_mode ~resolve_tc ?use_coercions rigidonly env sigma ?flags cj t
+  with e when Option.has_some loc ->
+    let _, info as iexn = Exninfo.capture e in
+    match Loc.get_loc info with
+    | Some _ -> Exninfo.iraise iexn
+    | None -> Exninfo.iraise (e, Loc.add_loc info (Option.get loc))
+
+let inh_conv_coerce_to ?loc ~program_mode ~resolve_tc ?use_coercions env sigma ?flags =
+  inh_conv_coerce_to_gen ?loc ~program_mode ~resolve_tc ?use_coercions false ?flags env sigma
+let inh_conv_coerce_rigid_to ?loc ~program_mode ~resolve_tc ?use_coercions env sigma ?flags =
+  inh_conv_coerce_to_gen ?loc ~program_mode ~resolve_tc ?use_coercions true ?flags env sigma
