@@ -66,6 +66,7 @@ type coe_info_typ = {
   coe_value : GlobRef.t;
   coe_typ : Constr.t;
   coe_local : bool;
+  coe_reversible : bool;
   coe_is_identity : bool;
   coe_is_projection : Projection.Repr.t option;
   coe_source : cl_typ;
@@ -92,6 +93,10 @@ let class_tab =
 let coercion_tab =
   Summary.ref ~name:"coercion_tab" (CoeTypMap.empty : coe_info_typ CoeTypMap.t)
 
+let reachable_from cl =
+  try (ClTypMap.find cl !class_tab).cl_reachable_from
+  with Not_found -> ClTypSet.empty
+
 module ClGraph :
 sig
   type t
@@ -131,10 +136,10 @@ let inheritance_graph =
 let add_new_class cl s =
   class_tab := ClTypMap.add cl s !class_tab
 
-let add_new_coercion coe s =
+let add_coercion coe s =
   coercion_tab := CoeTypMap.add coe s !coercion_tab
 
-let add_new_path (x, y) p =
+let add_path (x, y) p =
   inheritance_graph := ClGraph.add x y p !inheritance_graph
 
 (* class_info : cl_typ -> int * cl_info_typ *)
@@ -244,21 +249,19 @@ let apply_on_class_of env sigma t cont =
     let (cl,u,args) = find_class_type env sigma t in
     let { cl_param = n1 } = class_info cl in
     if not (Int.equal (List.length args) n1) then raise Not_found;
-    t, cont cl
+    cont cl
   with Not_found ->
     (* Is it worth to be more incremental on the delta steps? *)
     let t = Tacred.hnf_constr env sigma t in
     let (cl, u, args) = find_class_type env sigma t in
     let { cl_param = n1 } = class_info cl in
     if not (Int.equal (List.length args) n1) then raise Not_found;
-    t, cont cl
+    cont cl
 
-let lookup_path_between env sigma (s,t) =
-  let (s,(t,p)) =
-    apply_on_class_of env sigma s (fun i ->
+let lookup_path_between env sigma ~src:s ~tgt:t =
+  apply_on_class_of env sigma s (fun i ->
       apply_on_class_of env sigma t (fun j ->
-        lookup_path_between_class (i,j))) in
-  (s,t,p)
+          lookup_path_between_class (i,j)))
 
 let lookup_path_to_fun_from env sigma s =
   apply_on_class_of env sigma s lookup_path_to_fun_from_class
@@ -266,24 +269,23 @@ let lookup_path_to_fun_from env sigma s =
 let lookup_path_to_sort_from env sigma s =
   apply_on_class_of env sigma s lookup_path_to_sort_from_class
 
-let mkNamed = let open GlobRef in function
-  | ConstRef c -> EConstr.mkConst c
-  | VarRef v -> EConstr.mkVar v
-  | ConstructRef c -> EConstr.mkConstruct c
-  | IndRef i -> EConstr.mkInd i
-
 let get_coercion_constructor env coe =
   let evd = Evd.from_env env in
-  let red x = fst (Reductionops.whd_all_stack env evd x) in
-  match EConstr.kind evd (red (mkNamed coe.coe_value)) with
+  let evd, c = Evd.fresh_global env evd coe.coe_value in
+  let c = fst (Reductionops.whd_all_stack env evd c) in
+  match EConstr.kind evd c with
   | Constr.Construct (c, _) ->
-      c, Inductiveops.constructor_nrealargs env c -1
+    (* we don't return the modified evd as we drop the universes *)
+    c, Inductiveops.constructor_nrealargs env c -1
   | _ -> raise Not_found
 
 let lookup_pattern_path_between env (s,t) =
   List.map (get_coercion_constructor env)
     (ClGraph.find (CL_IND s) (CL_IND t) !inheritance_graph)
 
+let path_is_reversible p =
+  List.for_all (fun c -> c.coe_reversible) p
+
 (* rajouter une coercion dans le graphe *)
 
 let path_printer : ((cl_typ * cl_typ) * inheritance_path -> Pp.t) ref =
@@ -302,7 +304,7 @@ let install_path_comparator f = path_comparator := f
 let compare_path env sigma cl p q = !path_comparator env sigma cl p q
 
 let warn_ambiguous_path =
-  CWarnings.create ~name:"ambiguous-paths" ~category:"typechecker"
+  CWarnings.create ~name:"ambiguous-paths" ~category:CWarnings.CoreCategories.coercions
     (fun l -> prlist_with_sep fnl (fun (c,p,q) ->
          str"New coercion path " ++ print_path (c,p) ++
          if List.is_empty q then
@@ -321,7 +323,7 @@ let different_class_params env ci =
     | CL_CONST c -> Environ.is_polymorphic env (GlobRef.ConstRef c)
     | _ -> false
 
-let add_coercion_in_graph env sigma ic =
+let add_coercion_in_graph env sigma ?(update=false) ic =
   let source = ic.coe_source in
   let target = ic.coe_target in
   let source_info = class_info source in
@@ -350,9 +352,10 @@ let add_coercion_in_graph env sigma ic =
   let try_add_new_path (i,j as ij) p =
     if cl_typ_eq i j then check_coherence ij p [];
     if not (cl_typ_eq i j) || different_class_params env i then
-      match lookup_path_between_class ij with
-      | q -> (if not (cl_typ_eq i j) then check_coherence ij p q); false
-      | exception Not_found -> add_new_path ij p; true
+      if update then let () = add_path ij p in true else
+        match lookup_path_between_class ij with
+        | q -> (if not (cl_typ_eq i j) then check_coherence ij p q); false
+        | exception Not_found -> add_path ij p; true
     else
       false
   in
@@ -438,11 +441,11 @@ let add_class env sigma cl =
       cl_reachable_to = cl_singleton;
       cl_repr = cl }
 
-let declare_coercion env sigma c =
+let declare_coercion env sigma ?update c =
   let () = add_class env sigma c.coe_source in
   let () = add_class env sigma c.coe_target in
-  let () = add_new_coercion c.coe_value c in
-  add_coercion_in_graph env sigma c
+  let () = add_coercion c.coe_value c in
+  add_coercion_in_graph env sigma ?update c
 
 (* For printing purpose *)
 let classes () =
