@@ -27,7 +27,7 @@ let case_info_pattern_eq i1 i2 =
   i1.cip_extensible == i2.cip_extensible
 
 let rec constr_pattern_eq p1 p2 = match p1, p2 with
-| PRef r1, PRef r2 -> GlobRef.equal r1 r2
+| PRef r1, PRef r2 -> GlobRef.CanOrd.equal r1 r2
 | PVar v1, PVar v2 -> Id.equal v1 v2
 | PEvar (ev1, ctx1), PEvar (ev2, ctx2) ->
   Evar.equal ev1 ev2 && List.equal constr_pattern_eq ctx1 ctx2
@@ -80,27 +80,6 @@ and rec_declaration_eq (n1, c1, r1) (n2, c2, r2) =
   Array.equal constr_pattern_eq c1 c2 &&
   Array.equal constr_pattern_eq r1 r2
 
-let rec occur_meta_pattern = function
-  | PApp (f,args) ->
-      (occur_meta_pattern f) || (Array.exists occur_meta_pattern args)
-  | PProj (_,arg) -> occur_meta_pattern arg
-  | PLambda (na,t,c)  -> (occur_meta_pattern t) || (occur_meta_pattern c)
-  | PProd (na,t,c)  -> (occur_meta_pattern t) || (occur_meta_pattern c)
-  | PLetIn (na,b,t,c)  ->
-     Option.fold_left (fun b t -> b || occur_meta_pattern t) (occur_meta_pattern b) t || (occur_meta_pattern c)
-  | PIf (c,c1,c2)  ->
-      (occur_meta_pattern c) ||
-      (occur_meta_pattern c1) || (occur_meta_pattern c2)
-  | PCase(_, p,c,br) ->
-      Option.cata (fun (_, p) -> occur_meta_pattern p) false p ||
-      (occur_meta_pattern c) ||
-      (List.exists (fun (_,_,p) -> occur_meta_pattern p) br)
-  | PArray (t,def,ty) ->
-      Array.exists occur_meta_pattern t || occur_meta_pattern def || occur_meta_pattern ty
-  | PMeta _ | PSoApp _ -> true
-  | PEvar _ | PVar _ | PRef _ | PRel _ | PSort _ | PFix _ | PCoFix _
-    | PInt _ | PFloat _ -> false
-
 let rec occurn_pattern n = function
   | PRel p -> Int.equal n p
   | PApp (f,args) ->
@@ -155,14 +134,16 @@ let head_of_constr_reference sigma c = match EConstr.kind sigma c with
   | Var id -> GlobRef.VarRef id
   | _ -> anomaly (Pp.str "Not a rigid reference.")
 
-let pattern_of_constr env sigma t =
+let pattern_of_constr ~broken env sigma t =
+  let t = EConstr.Unsafe.to_constr t in
+  let kind = if broken then Constr.kind else fun c -> EConstr.kind_upto sigma c in
   let rec pattern_of_constr env t =
   let open Context.Rel.Declaration in
   match kind t with
     | Rel n  -> PRel n
     | Meta n -> PMeta (Some (Id.of_string ("META" ^ string_of_int n)))
     | Var id -> PVar id
-    | Sort s -> PSort (Sorts.family s)
+    | Sort s -> PSort (EConstr.ESorts.family sigma (EConstr.ESorts.make s))
     | Cast (c,_,_)      -> pattern_of_constr env c
     | LetIn (na,c,t,b) -> PLetIn (na.binder_name,
                                   pattern_of_constr env c,Some (pattern_of_constr env t),
@@ -177,7 +158,8 @@ let pattern_of_constr env sigma t =
         (match
           match kind f with
           | Evar (evk,args) ->
-            (match snd (Evd.evar_source evk sigma) with
+            let EvarInfo evi = Evd.find sigma evk in
+            (match snd (Evd.evar_source evi) with
               Evar_kinds.MatchingVar (Evar_kinds.SecondOrderPatVar id) -> Some id
             | _ -> None)
           | _ -> None
@@ -190,21 +172,24 @@ let pattern_of_constr env sigma t =
     | Proj (p, c) ->
       pattern_of_constr env (EConstr.Unsafe.to_constr (Retyping.expand_projection env sigma p (EConstr.of_constr c) []))
     | Evar (evk,ctxt as ev) ->
-      (match snd (Evd.evar_source evk sigma) with
+      let EvarInfo evi = Evd.find sigma evk in
+      (match snd (Evd.evar_source evi) with
       | Evar_kinds.MatchingVar (Evar_kinds.FirstOrderPatVar id) ->
         PMeta (Some id)
       | Evar_kinds.GoalEvar | Evar_kinds.VarInstance _ ->
         (* These are the two evar kinds used for existing goals *)
         (* see Proofview.mark_in_evm *)
          if Evd.is_defined sigma evk then pattern_of_constr env (Evd.existential_value0 sigma ev)
-         else PEvar (evk,List.map (pattern_of_constr env) ctxt)
+         else
+          let ctxt = Evd.expand_existential sigma (EConstr.of_existential ev) in
+          PEvar (evk,List.map (fun c -> pattern_of_constr env (EConstr.Unsafe.to_constr c)) ctxt)
       | Evar_kinds.MatchingVar (Evar_kinds.SecondOrderPatVar ido) -> assert false
       | _ ->
         PMeta None)
     | Case (ci, u, pms, p0, iv, a, br0) ->
         let (ci, p, iv, a, br) = Inductive.expand_case env (ci, u, pms, p0, iv, a, br0) in
         let pattern_of_ctx c (nas, c0) =
-          let ctx, c = Term.decompose_lam_n_decls (Array.length nas) c in
+          let ctx, c = Term.decompose_lambda_n_decls (Array.length nas) c in
           let env = push_rel_context ctx env in
           let c = pattern_of_constr env c in
           (Array.map Context.binder_name nas, c)
@@ -238,6 +223,9 @@ let pattern_of_constr env sigma t =
     in
   pattern_of_constr env t
 
+let legacy_bad_pattern_of_constr env sigma c : constr_pattern = pattern_of_constr ~broken:true env sigma c
+let pattern_of_constr env sigma c : constr_pattern = pattern_of_constr ~broken:false env sigma c
+
 (* To process patterns, we need a translation without typing at all. *)
 
 let map_pattern_with_binders g f l = function
@@ -264,42 +252,6 @@ let map_pattern_with_binders g f l = function
   | (PVar _ | PEvar _ | PRel _ | PRef _  | PSort _  | PMeta _ | PInt _
      | PFloat _ as x) -> x
 
-let error_instantiate_pattern id l =
-  let is = match l with
-  | [_] -> "is"
-  | _ -> "are"
-  in
-  user_err  (str "Cannot substitute the term bound to " ++ Id.print id
-    ++ strbrk " in pattern because the term refers to " ++ pr_enum Id.print l
-    ++ strbrk " which " ++ str is ++ strbrk " not bound in the pattern.")
-
-let instantiate_pattern env sigma lvar c =
-  let open EConstr in
-  let open Vars in
-  let rec aux vars = function
-  | PVar id as x ->
-      (try
-        let ctx,c = Id.Map.find id lvar in
-        try
-          let inst =
-            List.map
-              (fun id -> mkRel (List.index Name.equal (Name id) vars))
-              ctx
-          in
-          let c = substl inst c in
-          (* FIXME: Stupid workaround to pattern_of_constr being evar sensitive *)
-          let c = Evarutil.nf_evar sigma c in
-          pattern_of_constr env sigma (EConstr.Unsafe.to_constr c)
-        with Not_found (* List.index failed *) ->
-          let vars =
-            List.map_filter (function Name id -> Some id | _ -> None) vars in
-          error_instantiate_pattern id (List.subtract Id.equal ctx vars)
-       with Not_found (* Map.find failed *) ->
-         x)
-  | c ->
-      map_pattern_with_binders (fun id vars -> id::vars) aux vars c in
-  aux [] c
-
 let rec liftn_pattern k n = function
   | PRel i as x -> if i >= n then PRel (i+k) else x
   | c -> map_pattern_with_binders (fun _ -> succ) (liftn_pattern k) n c
@@ -313,7 +265,7 @@ let rec subst_pattern env sigma subst pat =
     if ref' == ref then pat else (match t with
         | None -> PRef ref'
         | Some t ->
-          pattern_of_constr env sigma t.Univ.univ_abstracted_value)
+          pattern_of_constr env sigma (EConstr.of_constr t.Univ.univ_abstracted_value))
   | PVar _
   | PEvar _
   | PRel _
@@ -409,49 +361,79 @@ let mkPLambda_or_LetIn (na,_,bo,t) c =
 let it_mkPProd_or_LetIn = List.fold_left (fun c d -> mkPProd_or_LetIn d c)
 let it_mkPLambda_or_LetIn = List.fold_left (fun c d -> mkPLambda_or_LetIn d c)
 
-let err ?loc pp = user_err ?loc ~hdr:"pattern_of_glob_constr" pp
+let err ?loc pp = user_err ?loc pp
 
 let warn_cast_in_pattern =
-  CWarnings.create ~name:"cast-in-pattern" ~category:"automation"
-    (fun () -> Pp.strbrk "Casts are ignored in patterns")
+  CWarnings.create ~name:"cast-in-pattern" ~category:CWarnings.CoreCategories.automation
+    (fun () -> Pp.strbrk "Cast types are ignored in patterns")
+
+type meta_accu =
+  | Metas of Id.Set.t ref
+  | InCastType of Loc.t option
+
+let push_meta metas n = match metas with
+  | Metas metas -> metas := Id.Set.add n !metas
+  | InCastType loc ->
+    CErrors.user_err ?loc
+      Pp.(str "Cannot bind pattern variable in cast type:"
+          ++ spc() ++ str "cast types are ignored in patterns.")
+
+let in_cast_type loc = function
+  | Metas _ -> InCastType loc
+  | InCastType _ as v ->
+    (* We report the error with the loc of the outermost cast
+       Alternatively we could use the loc of the meta, or the loc of the innermost cast. *)
+    v
 
 let rec pat_of_raw metas vars = DAst.with_loc_val (fun ?loc -> function
   | GVar id ->
       (try PRel (List.index Name.equal (Name id) vars)
        with Not_found -> PVar id)
   | GPatVar (Evar_kinds.FirstOrderPatVar n) ->
-      metas := n::!metas; PMeta (Some n)
+      push_meta metas n; PMeta (Some n)
   | GRef (gr,_) ->
       PRef (canonical_gr gr)
   (* Hack to avoid rewriting a complete interpretation of patterns *)
   | GApp (c, cl) ->
     begin match DAst.get c with
     | GPatVar (Evar_kinds.SecondOrderPatVar n) ->
-      metas := n::!metas; PSoApp (n, List.map (pat_of_raw metas vars) cl)
+      push_meta metas n; PSoApp (n, List.map (pat_of_raw metas vars) cl)
     | _ ->
       PApp (pat_of_raw metas vars c,
             Array.of_list (List.map (pat_of_raw metas vars) cl))
     end
+  | GProj ((p,_), cl, c) ->
+    if Structures.PrimitiveProjections.mem p then
+      let p = Option.get @@ Structures.PrimitiveProjections.find_opt p in
+      let p = Projection.make p false in
+      PProj (p, pat_of_raw metas vars c)
+    else
+      PApp (PRef (GlobRef.ConstRef p), Array.map_of_list (pat_of_raw metas vars) (cl @ [c]))
   | GLambda (na,bk,c1,c2) ->
-      Name.iter (fun n -> metas := n::!metas) na;
+      Name.iter (fun n -> push_meta metas n) na;
       PLambda (na, pat_of_raw metas vars c1,
                pat_of_raw metas (na::vars) c2)
   | GProd (na,bk,c1,c2) ->
-      Name.iter (fun n -> metas := n::!metas) na;
+      Name.iter (fun n -> push_meta metas n) na;
       PProd (na, pat_of_raw metas vars c1,
                pat_of_raw metas (na::vars) c2)
   | GLetIn (na,c1,t,c2) ->
-      Name.iter (fun n -> metas := n::!metas) na;
+      Name.iter (fun n -> push_meta metas n) na;
       PLetIn (na, pat_of_raw metas vars c1,
                Option.map (pat_of_raw metas vars) t,
                pat_of_raw metas (na::vars) c2)
   | GSort gs ->
      (try PSort (Glob_ops.glob_sort_family gs)
       with Glob_ops.ComplexSort -> user_err ?loc (str "Unexpected universe in pattern."))
-  | GHole _ ->
-      PMeta None
-  | GCast (c,_) ->
-      warn_cast_in_pattern ();
+  | GHole _ -> PMeta None
+  | GGenarg _ -> CErrors.user_err ?loc Pp.(str "Quotation not supported in pattern.")
+  | GCast (c,_,t) ->
+      let () =
+        (* Checks that there are no pattern variables in the type *)
+        let _ : constr_pattern = pat_of_raw (in_cast_type t.loc metas) vars t in
+        ()
+      in
+      warn_cast_in_pattern ?loc ();
       pat_of_raw metas vars c
   | GIf (c,(_,None),b1,b2) ->
       PIf (pat_of_raw metas vars c,
@@ -542,7 +524,7 @@ and pat_of_glob_in_context metas vars decls c =
 and pats_of_glob_branches loc metas vars ind brs =
   let get_arg p = match DAst.get p with
     | PatVar na ->
-      Name.iter (fun n -> metas := n::!metas) na;
+      Name.iter (fun n -> push_meta metas n) na;
       na
     | PatCstr(_,_,_) -> err ?loc:p.CAst.loc (Pp.str "Non supported pattern.")
   in
@@ -575,6 +557,6 @@ and pats_of_glob_branches loc metas vars ind brs =
   get_pat Int.Set.empty brs
 
 let pattern_of_glob_constr c =
-  let metas = ref [] in
-  let p = pat_of_raw metas [] c in
-  (!metas,p)
+  let metas = ref Id.Set.empty in
+  let p = pat_of_raw (Metas metas) [] c in
+  (!metas, p)
