@@ -8,6 +8,8 @@
 (*         *     (see LICENSE file for the text of the license)         *)
 (************************************************************************)
 
+module CVars = Vars
+
 open Pp
 open CErrors
 open Util
@@ -34,6 +36,7 @@ type retype_error =
   | NotAType
   | BadRel
   | BadVariable of Id.t
+  | BadEvar of Evar.t
   | BadMeta of int
   | BadRecursiveType
   | NonFunctionalConstruction
@@ -41,10 +44,11 @@ type retype_error =
 let print_retype_error = function
   | NotASort -> str "Not a sort"
   | NotAnArity -> str "Not an arity"
-  | NotAType -> str "Not a type (1)"
-  | BadRel -> str "unbound local variable"
-  | BadVariable id -> str "variable " ++ Id.print id ++ str " unbound"
-  | BadMeta n -> str "unknown meta " ++ int n
+  | NotAType -> str "Not a type"
+  | BadRel -> str "Unbound local variable"
+  | BadVariable id -> str "Variable " ++ Id.print id ++ str " unbound"
+  | BadEvar e -> str "Unknown evar " ++ Evar.print e
+  | BadMeta n -> str "Unknown meta " ++ int n
   | BadRecursiveType -> str "Bad recursive type"
   | NonFunctionalConstruction -> str "Non-functional construction"
 
@@ -57,11 +61,11 @@ let anomaly_on_error f x =
  with RetypeError e -> anomaly ~label:"retyping" (print_retype_error e ++ str ".")
 
 let get_type_from_constraints env sigma t =
-  if isEvar sigma (fst (decompose_app_vect sigma t)) then
+  if isEvar sigma (fst (decompose_app sigma t)) then
     match
       List.map_filter (fun (pbty,env,t1,t2) ->
-        if is_fconv Reduction.CONV env sigma t t1 then Some t2
-        else if is_fconv Reduction.CONV env sigma t t2 then Some t1
+        if is_fconv Conversion.CONV env sigma t t1 then Some t2
+        else if is_fconv Conversion.CONV env sigma t t2 then Some t1
         else None)
         (snd (Evd.extract_all_conv_pbs sigma))
     with
@@ -99,7 +103,7 @@ let sort_of_atomic_type env sigma ft args =
     match EConstr.kind sigma (whd_all env sigma ar), args with
     | Prod (na, t, b), h::l ->
       concl_of_arity (push_rel (LocalDef (na, lift n h, t)) env) (n + 1) b l
-    | Sort s, [] -> ESorts.kind sigma s
+    | Sort s, [] -> s
     | _ -> retype_error NotASort
   in concl_of_arity env 0 ft (Array.to_list args)
 
@@ -109,11 +113,9 @@ let type_of_var env id =
 
 let decomp_sort env sigma t =
   match EConstr.kind sigma (whd_all env sigma t) with
-  | Sort s -> ESorts.kind sigma s
+  | Sort s -> s
   | _ -> retype_error NotASort
 
-let destSort sigma s = ESorts.kind sigma (destSort sigma s)
-
 let betazetaevar_applist sigma n c l =
   let rec stacklam n env t stack =
     if Int.equal n 0 then applist (substl env t, stack) else
@@ -124,6 +126,12 @@ let betazetaevar_applist sigma n c l =
     | _ -> anomaly (Pp.str "Not enough lambda/let's.") in
   stacklam n [] c l
 
+let type_of_constant env sigma (c,u) =
+  let cb = lookup_constant c env in
+  let () = check_hyps_inclusion env sigma (GlobRef.ConstRef c) cb.const_hyps in
+  let ty = CVars.subst_instance_constr (EConstr.Unsafe.to_instance u) cb.const_type in
+  EConstr.of_constr (rename_type ty (GlobRef.ConstRef c))
+
 let retype ?(polyprop=true) sigma =
   let rec type_of env cstr =
     match EConstr.kind sigma cstr with
@@ -136,10 +144,13 @@ let retype ?(polyprop=true) sigma =
       in
       lift n ty
     | Var id -> type_of_var env id
-    | Const (cst, u) -> EConstr.of_constr (rename_type_of_constant env (cst, EInstance.kind sigma u))
-    | Evar ev -> existential_type sigma ev
-    | Ind (ind, u) -> EConstr.of_constr (rename_type_of_inductive env (ind, EInstance.kind sigma u))
-    | Construct (cstr, u) -> EConstr.of_constr (rename_type_of_constructor env (cstr, EInstance.kind sigma u))
+    | Const c -> type_of_constant env sigma c
+    | Evar ev -> begin match Evd.existential_type_opt sigma ev with
+        | Some t -> t
+        | None -> retype_error (BadEvar (fst ev))
+      end
+    | Ind ind -> Inductiveops.e_type_of_inductive env sigma ind
+    | Construct c -> Inductiveops.e_type_of_constructor env sigma c
     | Case (ci,u,pms,p,iv,c,lf) ->
         let (_,p,iv,c,lf) = EConstr.expand_case env sigma (ci,u,pms,p,iv,c,lf) in
         let Inductiveops.IndType(indf,realargs) =
@@ -181,21 +192,21 @@ let retype ?(polyprop=true) sigma =
       let arr = EConstr.of_constr @@ Typeops.type_of_array env (EInstance.kind sigma u) in
       mkApp(arr, [|ty|])
 
-  and sort_of env t =
+  and sort_of env t : ESorts.t =
     match EConstr.kind sigma t with
     | Cast (c,_, s) when isSort sigma s -> destSort sigma s
     | Sort s ->
       begin match ESorts.kind sigma s with
-      | SProp | Prop | Set -> Sorts.type1
-      | Type u -> Sorts.sort_of_univ (Univ.Universe.super u)
+      | SProp | Prop | Set -> ESorts.type1
+      | Type u | QSort (_, u) -> ESorts.make (Sorts.sort_of_univ (Univ.Universe.super u))
       end
     | Prod (name,t,c2) ->
       let dom = sort_of env t in
       let rang = sort_of (push_rel (LocalAssum (name,t)) env) c2 in
-      Typeops.sort_of_product env dom rang
+      ESorts.make (Typeops.sort_of_product env (ESorts.kind sigma dom) (ESorts.kind sigma rang))
     | App(f,args) when Termops.is_template_polymorphic_ind env sigma f ->
       let t = type_of_global_reference_knowing_parameters env f args in
-        sort_of_atomic_type env sigma t args
+      sort_of_atomic_type env sigma t args
     | App(f,args) -> sort_of_atomic_type env sigma (type_of env f) args
     | Lambda _ | Fix _ | Construct _ -> retype_error NotAType
     | _ -> decomp_sort env sigma (type_of env t)
@@ -205,12 +216,7 @@ let retype ?(polyprop=true) sigma =
     | Ind (ind, u) ->
       let u = EInstance.kind sigma u in
       let mip = lookup_mind_specif env ind in
-      let paramtyps = Array.map_to_list (fun arg () ->
-          let t = type_of env arg in
-          let s = sort_of_arity_with_constraints env sigma t in
-          ESorts.kind sigma s)
-          args
-      in
+      let paramtyps = make_param_univs env sigma (ind,u) args in
       EConstr.of_constr
         (Inductive.type_of_inductive_knowing_parameters
            ~polyprop (mip, u) paramtyps)
@@ -219,13 +225,24 @@ let retype ?(polyprop=true) sigma =
       EConstr.of_constr (type_of_constructor env (cstr, u))
     | _ -> assert false
 
+  and make_param_univs env sigma indu args =
+    Array.map_to_list (fun arg ~expected ->
+          let t = type_of env arg in
+          let s = sort_of_arity_with_constraints env sigma t in
+          match ESorts.kind sigma s with
+          | Sorts.SProp -> TemplateUniv (Univ.Universe.make expected)
+          | Sorts.Prop -> TemplateProp
+          | Sorts.Set -> TemplateUniv Univ.Universe.type0
+          | Sorts.Type u | Sorts.QSort (_, u) -> TemplateUniv u)
+      args
+
   in type_of, sort_of, type_of_global_reference_knowing_parameters
 
 let get_sort_family_of ?(polyprop=true) env sigma t =
   let type_of,_,type_of_global_reference_knowing_parameters = retype ~polyprop sigma in
   let rec sort_family_of env t =
     match EConstr.kind sigma t with
-    | Cast (c,_, s) when isSort sigma s -> Sorts.family (destSort sigma s)
+    | Cast (c,_, s) when isSort sigma s -> ESorts.family sigma (destSort sigma s)
     | Sort _ -> InType
     | Prod (name,t,c2) ->
         let s2 = sort_family_of (push_rel (LocalAssum (name,t)) env) c2 in
@@ -233,12 +250,12 @@ let get_sort_family_of ?(polyprop=true) env sigma t =
            s2 == InSet && sort_family_of env t == InType then InType else s2
     | App(f,args) when Termops.is_template_polymorphic_ind env sigma f ->
         let t = type_of_global_reference_knowing_parameters env f args in
-        Sorts.family (sort_of_atomic_type env sigma t args)
+        ESorts.family sigma (sort_of_atomic_type env sigma t args)
     | App(f,args) ->
-        Sorts.family (sort_of_atomic_type env sigma (type_of env f) args)
+        ESorts.family sigma (sort_of_atomic_type env sigma (type_of env f) args)
     | Lambda _ | Fix _ | Construct _ -> retype_error NotAType
     | _ ->
-      Sorts.family (decomp_sort env sigma (type_of env t))
+      ESorts.family sigma (decomp_sort env sigma (type_of env t))
   in sort_family_of env t
 
 let get_sort_of ?(polyprop=true) env sigma t =
@@ -262,6 +279,28 @@ let get_type_of ?(polyprop=true) ?(lax=false) env sigma c =
   let f,_,_ = retype ~polyprop sigma in
     if lax then f env c else anomaly_on_error (f env) c
 
+let rec check_named env sigma c =
+  match EConstr.kind sigma c with
+  | Var id ->
+    (try ignore (lookup_named id env)
+     with Not_found -> retype_error (BadVariable id))
+  | Evar _ ->
+    (* This is cheating, but some devs exploit that a
+       dependency in the evar args may disappear *)
+    ()
+  | _ -> EConstr.iter sigma (check_named env sigma) c
+
+let reinterpret_get_type_of ~src env sigma c =
+  try
+    check_named env sigma c;
+    get_type_of ~lax:true env sigma c
+  with RetypeError e ->
+    user_err
+      (str "Cannot reinterpret " ++ Id.print src ++ str " bound to " ++
+       quote (Termops.Internal.print_constr_env env sigma c) ++
+       str " in the current environment" ++ spc() ++
+       surround (print_retype_error e))
+
 (* Makes an unsafe judgment from a constr *)
 let get_judgment_of env evc c = { uj_val = c; uj_type = get_type_of env evc c }
 
@@ -319,9 +358,9 @@ let relevance_of_term env sigma c =
   else Sorts.Relevant
 
 let relevance_of_type env sigma t =
-  let s = get_sort_family_of env sigma t in
-  Sorts.relevance_of_sort_family s
+  let s = get_sort_of env sigma t in
+  ESorts.relevance_of_sort sigma s
 
-let relevance_of_sort s = Sorts.relevance_of_sort (EConstr.Unsafe.to_sorts s)
+let relevance_of_sort = ESorts.relevance_of_sort
 
-let relevance_of_sort_family f =  Sorts.relevance_of_sort_family f
+let relevance_of_sort_family sigma f = Evarutil.nf_relevance sigma (Sorts.relevance_of_sort_family f)
