@@ -17,7 +17,6 @@ open Context
 open Vars
 open Environ
 open Inductive
-open Reduction
 open Vmvalues
 open Vm
 open Context.Rel.Declaration
@@ -33,8 +32,8 @@ let e_whd_all = Reductionops.clos_whd_flags CClosure.all
 
 let crazy_type =  mkSet
 
-let decompose_prod env t =
-  let (name,dom,codom) = destProd (whd_all env t) in
+let decompose_prod env sigma t =
+  let (name,dom,codom) = destProd (EConstr.Unsafe.to_constr (e_whd_all env sigma (EConstr.of_constr t))) in
   let name = map_annot (function
   | Anonymous -> Name (Id.of_string "x")
   | Name _ as na -> na) name
@@ -60,7 +59,7 @@ let invert_tag cst tag reloc_tbl =
 
 let app_type env sigma c =
   let t = e_whd_all env sigma c in
-  decompose_appvect (EConstr.Unsafe.to_constr t)
+  decompose_app (EConstr.Unsafe.to_constr t)
 
 let find_rectype_a env sigma c =
   let (t, l) = app_type env sigma c in
@@ -76,7 +75,7 @@ let type_constructor mind mib u (ctx, typ) params =
   let ndecls = Context.Rel.length mib.mind_params_ctxt in
   if Int.equal ndecls 0 then ctyp
   else
-    let _,ctyp = decompose_prod_n_assum ndecls ctyp in
+    let _,ctyp = decompose_prod_n_decls ndecls ctyp in
     substl (subst_of_rel_context_instance mib.mind_params_ctxt params)
       ctyp
 
@@ -104,15 +103,18 @@ let construct_of_constr_block = construct_of_constr false
 let type_of_ind env (ind, u) =
   type_of_inductive (Inductive.lookup_mind_specif env ind, u)
 
-let build_branches_type env sigma (mind,_ as _ind) mib mip u params p =
+let get_case_annot decls =
+  Array.map_of_list (fun decl -> get_annot decl) (List.rev decls)
+
+let build_branches_type env sigma (mind,_ as _ind) mib mip u params (pctx, p) =
   let rtbl = mip.mind_reloc_tbl in
   (* [build_one_branch i cty] construit le type de la ieme branche (commence
      a 0) et les lambda correspondant aux realargs *)
+  let p = it_mkLambda_or_LetIn p pctx in (* TODO: prevent useless cut? *)
   let build_one_branch i cty =
     let typi = type_constructor mind mib u cty params in
-    let decl,indapp = Reductionops.splay_prod env sigma (EConstr.of_constr typi) in
+    let decl,indapp = Reductionops.hnf_decompose_prod env sigma (EConstr.of_constr typi) in
     let decl = List.map (on_snd EConstr.Unsafe.to_constr) decl in
-    let decl_with_letin,_ = decompose_prod_assum typi in
     let ((ind,u),cargs) = find_rectype_a env sigma indapp in
     let nparams = Array.length params in
     let carity = snd (rtbl.(i)) in
@@ -126,10 +128,18 @@ let build_branches_type env sigma (mind,_ as _ind) mib mip u params p =
       let dep_cstr = mkApp(mkApp(mkConstructU (cstr,u),params),relargs) in
       mkApp(papp,[|dep_cstr|])
     in
-    decl, decl_with_letin, codom
+    let decl_with_letin = List.firstn mip.mind_consnrealdecls.(i) (fst cty) in
+    let nas = get_case_annot decl_with_letin in
+    let rec get_lift decls = match decls with
+    | [] -> Esubst.el_id
+    | LocalDef _ :: decls -> Esubst.el_shft 1 (get_lift decls)
+    | LocalAssum _ :: decls -> Esubst.el_lift (get_lift decls)
+    in
+    decl, nas, get_lift decl_with_letin, codom
   in Array.mapi build_one_branch mip.mind_nf_lc
 
-let build_case_type p realargs c =
+let build_case_type (pctx, p) realargs c =
+  let p = it_mkLambda_or_LetIn p pctx in (* TODO: prevent useless cut? *)
   mkApp(mkApp(p, realargs), [|c|])
 
 (* La fonction de normalisation *)
@@ -158,14 +168,14 @@ and nf_whd env sigma whd typ =
       let t = ta.(i) in
       let _, args = nf_args env sigma vargs t in
       mkApp(cfd,args)
-  | Vconstr_const n ->
+  | Vconst n ->
     construct_of_constr_const env sigma n typ
-  | Vconstr_block b ->
+  | Vblock b ->
       let tag = btag b in
       let (tag,ofs) =
         if tag = Obj.last_non_constant_constructor_tag then
           match whd_val (bfield b 0) with
-          | Vconstr_const tag -> (tag+Obj.last_non_constant_constructor_tag, 1)
+          | Vconst tag -> (tag+Obj.last_non_constant_constructor_tag, 1)
           | _ -> assert false
         else (tag, 0) in
       let capp,ctyp = construct_of_constr_block env sigma tag typ in
@@ -174,9 +184,9 @@ and nf_whd env sigma whd typ =
   | Vint64 i -> i |> Uint63.of_int64 |> mkInt
   | Vfloat64 f -> f |> Float64.of_float |> mkFloat
   | Varray t -> nf_array env sigma t typ
-  | Vatom_stk(Aid idkey, stk) ->
+  | Vaccu (Aid idkey, stk) ->
       constr_type_of_idkey env sigma idkey stk
-  | Vatom_stk(Aind ((mi,i) as ind), stk) ->
+  | Vaccu (Aind ((mi, i) as ind), stk) ->
      let mib = Environ.lookup_mind mi env in
      let nb_univs =
        Univ.AbstractContext.size (Declareops.inductive_polymorphic_context mib)
@@ -185,10 +195,8 @@ and nf_whd env sigma whd typ =
        let pind = (ind, u) in (mkIndU pind, type_of_ind env pind)
      in
      nf_univ_args ~nb_univs mk env sigma stk
-  | Vatom_stk(Asort s, stk) ->
+  | Vaccu (Asort s, stk) ->
     assert (List.is_empty stk); mkSort s
-  | Vuniv_level lvl ->
-    assert false
 
 and nf_univ_args ~nb_univs mk env sigma stk =
   let u =
@@ -205,25 +213,28 @@ and nf_univ_args ~nb_univs mk env sigma stk =
   nf_stk ~from:nb_univs env sigma t ty stk
 
 and nf_evar env sigma evk stk =
-  let evi = try Evd.find sigma evk with Not_found -> assert false in
-  let hyps = Environ.named_context_of_val (Evd.evar_filtered_hyps evi) in
-  let concl = EConstr.to_constr ~abort_on_undefined_evars:false sigma @@ Evd.evar_concl evi in
+  let evi = try Evd.find_undefined sigma evk with Not_found -> assert false in
+  let hyps = EConstr.named_context_of_val (Evd.evar_filtered_hyps evi) in
   if List.is_empty hyps then
-    nf_stk env sigma (mkEvar (evk, [])) concl stk
+    let concl = EConstr.to_constr ~abort_on_undefined_evars:false sigma @@ Evd.evar_concl evi in
+    nf_stk env sigma (mkEvar (evk, SList.empty)) concl stk
   else match stk with
   | Zapp args :: stk ->
     (* We assume that there is no consecutive Zapp nodes in a VM stack. Is that
        really an invariant? *)
     (* Let-bound arguments are present in the evar arguments but not in the
        type, so we turn the let into a product. *)
+    let concl = Evd.evar_concl evi in
     let hyps = Context.Named.drop_bodies hyps in
-    let fold accu d = Term.mkNamedProd_or_LetIn d accu in
+    let fold accu d = EConstr.mkNamedProd_or_LetIn sigma d accu in
     let t = List.fold_left fold concl hyps in
+    let t = EConstr.to_constr ~abort_on_undefined_evars:false sigma t in
     let t, args = nf_args env sigma args t in
     let inst, args = Array.chop (List.length hyps) args in
     (* Evar instances are reversed w.r.t. argument order *)
-    let inst = Array.rev_to_list inst in
-    let c = mkApp (mkEvar (evk, inst), args) in
+    let inst = Array.to_list inst in
+    let ev = EConstr.(Unsafe.to_constr @@ mkLEvar sigma (evk, List.rev_map of_constr inst)) in
+    let c = mkApp (ev, args) in
     nf_stk env sigma c t stk
   | _ ->
     CErrors.anomaly (Pp.str "Argument size mismatch when decompiling an evar")
@@ -262,7 +273,7 @@ and nf_stk ?from:(from=0) env sigma c t stk  =
   | Zfix (f,vargs) :: stk ->
       assert (from = 0) ;
       let fa, typ = nf_fix_app env sigma f vargs in
-      let _,_,codom = decompose_prod env typ in
+      let _,_,codom = decompose_prod env sigma typ in
       nf_stk env sigma (mkApp(fa,[|c|])) (subst1 c codom) stk
   | Zswitch sw :: stk ->
       assert (from = 0) ;
@@ -270,27 +281,32 @@ and nf_stk ?from:(from=0) env sigma c t stk  =
       let (mib,mip) = Inductive.lookup_mind_specif env ind in
       let nparams = mib.mind_nparams in
       let params,realargs = Util.Array.chop nparams allargs in
-      let nparamdecls = Context.Rel.length (Inductive.inductive_paramdecls (mib,u)) in
-      let pT =
-        hnf_prod_applist_assum env nparamdecls (type_of_ind env (ind,u)) (Array.to_list params) in
-      let p, relevance = nf_predicate env sigma (ind,u) mip params (type_of_switch sw) pT in
+      let pctx =
+        let realdecls, _ = List.chop mip.mind_nrealdecls mip.mind_arity_ctxt in
+        let nas = List.rev_map RelDecl.get_annot realdecls @ [nameR (Id.of_string "c")] in
+        expand_arity (mib, mip) (ind, u) params (Array.of_list nas)
+      in
+      let p, relevance = nf_predicate env sigma (ind,u) mip params (type_of_switch sw) pctx in
       (* Calcul du type des branches *)
-      let btypes = build_branches_type env sigma ind mib mip u params p in
+      let btypes = build_branches_type env sigma ind mib mip u params (pctx, p) in
       (* calcul des branches *)
       let bsw = branch_of_switch (nb_rel env) sw in
       let mkbranch i (n,v) =
-        let decl,decl_with_letin,codom = btypes.(i) in
+        let decl, nas, lft, codom = btypes.(i) in
         let b = nf_val (Termops.push_rels_assum decl env) sigma v codom in
-        Termops.it_mkLambda_or_LetIn_from_no_LetIn b decl_with_letin
+        (* No let-ins in [codom], so we have to ignore the ones from the branch *)
+        let b = exliftn lft b in
+        (nas, b)
       in
       let branchs = Array.mapi mkbranch bsw in
-      let tcase = build_case_type p realargs c in
+      let tcase = build_case_type (pctx, p) realargs c in
+      let p = (get_case_annot pctx, p) in
       let ci = Inductiveops.make_case_info env ind relevance RegularStyle in
       let iv = if Typeops.should_invert_case env ci then
           CaseInvert {indices=realargs}
         else NoInvert
       in
-      nf_stk env sigma (mkCase (Inductive.contract_case env (ci, p, iv, c, branchs))) tcase stk
+      nf_stk env sigma (mkCase (ci, u, params, p, iv, c, branchs)) tcase stk
   | Zproj p :: stk ->
     let () = assert (from = 0) in
     let ((ind, u), args) = Inductiveops.find_mrectype env sigma (EConstr.of_constr t) in
@@ -307,39 +323,20 @@ and nf_stk ?from:(from=0) env sigma c t stk  =
     let p = Projection.make p true in
     nf_stk env sigma (mkProj (p, c)) ty stk
 
-and nf_predicate env sigma ind mip params v pT =
-  match kind (whd_allnolet env pT) with
-  | LetIn (name,b,t,pT) ->
-      let body, rel =
-        nf_predicate (push_rel (LocalDef (name,b,t)) env) sigma ind mip params v pT in
-      mkLetIn (name,b,t,body), rel
-  | Prod (name,dom,codom) -> begin
+and nf_predicate env sigma ind mip params v pctx =
+  (* TODO: we should expose some variant of Vm.mkrel_vstack *)
+  let fold decl (k, v) = match decl with
+  | LocalDef _ -> (k + 1, v)
+  | LocalAssum _ ->
     match whd_val v with
-    | Vfun f ->
-      let k = nb_rel env in
-      let vb = reduce_fun k f in
-      let body, rel =
-        nf_predicate (push_rel (LocalAssum (name,dom)) env) sigma ind mip params vb codom in
-      mkLambda(name,dom,body), rel
-    | _ -> assert false
-    end
-  | _ ->
-    match whd_val v with
-    | Vfun f ->
-      let k = nb_rel env in
-      let vb = reduce_fun k f in
-      let name = Name (Id.of_string "c") in
-      let n = mip.mind_nrealargs in
-      let rargs = Array.init n (fun i -> mkRel (n-i)) in
-      let params = if Int.equal n 0 then params else Array.map (lift n) params in
-      let dom = mkApp(mkIndU ind,Array.append params rargs) in
-      let r = Inductive.relevance_of_inductive env (fst ind) in
-      let name = make_annot name r in
-      let env = push_rel (LocalAssum (name,dom)) env in
-      let body = nf_vtype env sigma vb in
-      let rel = Retyping.relevance_of_type env sigma (EConstr.of_constr body) in
-      mkLambda(name,dom,body), rel
+    | Vfun f -> (k + 1, reduce_fun k f)
     | _ -> assert false
+  in
+  let (_, v) = List.fold_right fold pctx (nb_rel env, v) in
+  let env = push_rel_context pctx env in
+  let body = nf_vtype env sigma v in
+  let rel = Retyping.relevance_of_type env sigma (EConstr.of_constr body) in
+  body, rel
 
 and nf_args env sigma vargs ?from:(f=0) t =
   let t = ref t in
@@ -347,7 +344,7 @@ and nf_args env sigma vargs ?from:(f=0) t =
   let args =
     Array.init len
       (fun i ->
-        let _,dom,codom = decompose_prod env !t in
+        let _,dom,codom = decompose_prod env sigma !t in
         let c = nf_val env sigma (arg vargs (f+i)) dom in
         t := subst1 c codom; c) in
   !t,args
@@ -358,7 +355,7 @@ and nf_bargs env sigma b ofs t =
   let args =
     Array.init len
       (fun i ->
-        let _,dom,codom = decompose_prod env !t in
+        let _,dom,codom = decompose_prod env sigma !t in
         let c = nf_val env sigma (bfield b (i+ofs)) dom in
         t := subst1 c codom; c) in
   args
@@ -367,7 +364,7 @@ and nf_fun env sigma f typ =
   let k = nb_rel env in
   let vb = reduce_fun k f in
   let name,dom,codom =
-    try decompose_prod env typ
+    try decompose_prod env sigma typ
     with DestKO ->
       CErrors.anomaly
         Pp.(strbrk "Returned a functional value in type " ++
@@ -425,11 +422,16 @@ and nf_array env sigma t typ =
   let u = snd (destConst ty) in
   mkArray(u, t, nf_val env sigma vdef typ_elem, typ_elem)
 
+let evars_of_evar_map sigma =
+  { Genlambda.evars_val = Evd.evar_handler sigma }
+
 let cbv_vm env sigma c t  =
+  if not (Environ.typing_flags env).enable_VM then
+    CErrors.user_err Pp.(str "vm_compute reduction has been disabled.");
   if Termops.occur_meta sigma c then
     CErrors.user_err Pp.(str "vm_compute does not support metas.");
   (* This evar-normalizes terms beforehand *)
   let c = EConstr.Unsafe.to_constr c in
   let t = EConstr.Unsafe.to_constr t in
-  let v = Vmsymtable.val_of_constr env (Evd.existential_opt_value0 sigma) c in
+  let v = Vmsymtable.val_of_constr env (evars_of_evar_map sigma) c in
   EConstr.of_constr (nf_val env sigma v t)
