@@ -40,9 +40,11 @@ type option_command =
   | OptionUnset
   | OptionAppend of string
 
+type require_injection = { lib: string; prefix: string option; export: Lib.export_flag option; }
+
 type injection_command =
   | OptionInjection of (Goptions.option_name * option_command)
-  | RequireInjection of (string * string option * bool option)
+  | RequireInjection of require_injection
   | WarnNoNative of string
   | WarnNativeDeprecated
 
@@ -53,6 +55,8 @@ type coqargs_logic_config = {
   toplevel_name     : top;
 }
 
+type time_config = ToFeedback | ToFile of string
+
 type coqargs_config = {
   logic       : coqargs_logic_config;
   rcfile      : string option;
@@ -61,7 +65,7 @@ type coqargs_config = {
   native_compiler : native_compiler;
   native_output_dir : CUnix.physical_path;
   native_include_dirs : CUnix.physical_path list;
-  time        : bool;
+  time        : time_config option;
   print_emacs : bool;
 }
 
@@ -80,7 +84,7 @@ type coqargs_pre = {
 type coqargs_query =
   | PrintWhere | PrintConfig
   | PrintVersion | PrintMachineReadableVersion
-  | PrintHelp of Usage.specific_usage
+  | PrintHelp of Boot.Usage.specific_usage
 
 type coqargs_main =
   | Queries of coqargs_query list
@@ -116,7 +120,7 @@ let default_config = {
   native_compiler = default_native;
   native_output_dir = ".coq-native";
   native_include_dirs = [];
-  time         = false;
+  time         = None;
   print_emacs  = false;
 
   (* Quiet / verbosity options should be here *)
@@ -158,7 +162,7 @@ let add_vo_include opts unix_path coq_path implicit =
         unix_path; coq_path; has_ml = false; implicit; recursive = true } :: opts.pre.vo_includes }}
 
 let add_vo_require opts d p export =
-  { opts with pre = { opts.pre with injections = RequireInjection (d, p, export) :: opts.pre.injections }}
+  { opts with pre = { opts.pre with injections = RequireInjection {lib=d; prefix=p; export} :: opts.pre.injections }}
 
 let add_load_vernacular opts verb s =
     { opts with pre = { opts.pre with load_vernacular_list = (CUnix.make_suffix s ".v",verb) :: opts.pre.load_vernacular_list }}
@@ -221,10 +225,10 @@ let set_option = let open Goptions in function
   | opt, OptionAppend v -> set_string_option_append_value_gen ~locality:OptLocal opt v
 
 let get_compat_file = function
-  | "8.15" -> "Coq.Compat.Coq815"
-  | "8.14" -> "Coq.Compat.Coq814"
-  | "8.13" -> "Coq.Compat.Coq813"
-  | ("8.12" | "8.11" | "8.10" | "8.9" | "8.8" | "8.7" | "8.6" | "8.5" | "8.4" | "8.3" | "8.2" | "8.1" | "8.0") as s ->
+  | "8.18" -> "Coq.Compat.Coq818"
+  | "8.17" -> "Coq.Compat.Coq817"
+  | "8.16" -> "Coq.Compat.Coq816"
+  | ("8.15" | "8.14" | "8.13" | "8.12" | "8.11" | "8.10" | "8.9" | "8.8" | "8.7" | "8.6" | "8.5" | "8.4" | "8.3" | "8.2" | "8.1" | "8.0") as s ->
     CErrors.user_err
       Pp.(str "Compatibility with version " ++ str s ++ str " not supported.")
   | s ->
@@ -243,17 +247,19 @@ let parse_option_set opt =
 
 let get_native_compiler s =
   (* We use two boolean flags because the four states make sense, even if
-  only three are accessible to the user at the moment. The selection of the
-  produced artifact(s) (`.vo`, `.vio`, `.coq-native`, ...) should be done by
-  a separate flag, and the "ondemand" value removed. Once this is done, use
-  [get_bool] here. *)
+     only three are accessible to the user at the moment. The selection of the
+     produced artifact(s) (`.vo`, `.vio`, `.coq-native`, ...) should be done by
+     a separate flag, and the "ondemand" value removed. Once this is done, use
+     [get_bool] here. *)
   let n = match s with
     | ("yes" | "on") -> NativeOn {ondemand=false}
     | "ondemand" -> NativeOn {ondemand=true}
     | ("no" | "off") -> NativeOff
     | _ ->
-       error_wrong_arg ("Error: (yes|no|ondemand) expected after option -native-compiler.") in
-  if Coq_config.native_compiler = NativeOff && n <> NativeOff then
+      error_wrong_arg ("Error: (yes|no|ondemand) expected after option -native-compiler.")
+  in
+  if n = NativeOff then n, []
+  else if Coq_config.native_compiler = NativeOff then
     NativeOff, [WarnNativeDeprecated; WarnNoNative s]
   else n, [WarnNativeDeprecated]
 
@@ -291,7 +297,7 @@ let parse_args ~usage ~init arglist : t * string list =
       }}
 
     |"-compat" ->
-      add_vo_require oval (get_compat_file (next ())) None (Some false)
+      add_vo_require oval (get_compat_file (next ())) None (Some Lib.Import)
 
     |"-exclude-dir" ->
       System.exclude_directory (next ()); oval
@@ -299,9 +305,6 @@ let parse_args ~usage ~init arglist : t * string list =
     |"-init-file" ->
       { oval with config = { oval.config with rcfile = Some (next ()); }}
 
-    |"-load-vernac-object" ->
-      add_vo_require oval (next ()) None None
-
     |"-load-vernac-source"|"-l" ->
       add_load_vernacular oval false (next ())
 
@@ -317,18 +320,21 @@ let parse_args ~usage ~init arglist : t * string list =
       Flags.profile_ltac_cutoff := get_float ~opt (next ());
       oval
 
-    |"-rfrom" ->
+    |"-load-vernac-object"|"-require" ->
+      add_vo_require oval (next ()) None None
+
+    |"-require-import" | "-ri" -> add_vo_require oval (next ()) None (Some Lib.Import)
+
+    |"-require-export" | "-re" -> add_vo_require oval (next ()) None (Some Lib.Export)
+
+    |"-require-from"|"-rfrom" ->
       let from = next () in add_vo_require oval (next ()) (Some from) None
 
-    |"-require-import" | "-ri" -> add_vo_require oval (next ()) None (Some false)
-
-    |"-require-export" | "-re" -> add_vo_require oval (next ()) None (Some true)
-
     |"-require-import-from" | "-rifrom" ->
-      let from = next () in add_vo_require oval (next ()) (Some from) (Some false)
+      let from = next () in add_vo_require oval (next ()) (Some from) (Some Lib.Import)
 
     |"-require-export-from" | "-refrom" ->
-      let from = next () in add_vo_require oval (next ()) (Some from) (Some true)
+      let from = next () in add_vo_require oval (next ()) (Some from) (Some Lib.Export)
 
     |"-top" ->
       let topname = Libnames.dirpath_of_string (next ()) in
@@ -368,7 +374,7 @@ let parse_args ~usage ~init arglist : t * string list =
       { oval with config = {oval.config with native_include_dirs = include_dir :: oval.config.native_include_dirs } }
 
     (* Options with zero arg *)
-    |"-test-mode" -> Vernacinterp.test_mode := true; oval
+    |"-test-mode" -> Synterp.test_mode := true; oval
     |"-beautify" -> Flags.beautify := true; oval
     |"-config"|"--config" -> set_query oval PrintConfig
 
@@ -399,9 +405,10 @@ let parse_args ~usage ~init arglist : t * string list =
       Flags.quiet := true;
       Flags.make_warn false;
       oval
-    |"-time" -> { oval with config = { oval.config with time = true }}
+    |"-time" -> { oval with config = { oval.config with time = Some ToFeedback }}
+    |"-time-file" -> { oval with config = { oval.config with time = Some (ToFile (next())) }}
     |"-type-in-type" -> set_logic (fun o -> { o with type_in_type = true }) oval
-    |"-unicode" -> add_vo_require oval "Utf8_core" None (Some false)
+    |"-unicode" -> add_vo_require oval "Utf8_core" None (Some Lib.Import)
     |"-where" -> set_query oval PrintWhere
     |"-h"|"-H"|"-?"|"-help"|"--help" -> set_query oval (PrintHelp usage)
     |"-v"|"--version" -> set_query oval PrintVersion
@@ -435,20 +442,12 @@ let parse_args ~usage ~init args =
 (******************************************************************************)
 (* Startup LoadPath and Modules                                               *)
 (******************************************************************************)
+
 (* prelude_data == From Coq Require Import Prelude. *)
-let prelude_data = "Prelude", Some "Coq", Some false
+let prelude_data = RequireInjection { lib = "Prelude"; prefix = Some "Coq"; export = Some Lib.Import; }
 
 let injection_commands opts =
-  if opts.pre.load_init then RequireInjection prelude_data :: opts.pre.injections else opts.pre.injections
-
-let build_load_path opts =
-  let ml_path, vo_path =
-    if opts.pre.boot then [],[]
-    else
-      let coqenv = Boot.Env.init () in
-      Coqloadpath.init_load_path ~coqenv in
-  ml_path @ opts.pre.ml_includes ,
-  vo_path @ opts.pre.vo_includes
+  if opts.pre.load_init then prelude_data :: opts.pre.injections else opts.pre.injections
 
 let dirpath_of_file f =
   let ldir0 =
