@@ -67,7 +67,7 @@ module RelDecl = Context.Rel.Declaration
 
 let hid = Id.of_string "H"
 let xid = Id.of_string "X"
-let default_id_of_sort = function InSProp | InProp | InSet -> hid | InType -> xid
+let default_id_of_sort = function InSProp | InProp | InSet -> hid | InType | InQSort -> xid
 let fresh env id avoid =
   let freshid = next_global_ident_away id avoid in
   freshid, Id.Set.add freshid avoid
@@ -93,7 +93,7 @@ let name_context env hyps =
           let d' = name_assumption env d in (push_rel d' env, d' :: hyps))
        (env,[]) (List.rev hyps))
 
-let my_it_mkLambda_or_LetIn s c = it_mkLambda_or_LetIn c s
+let my_it_mkLambda_or_LetIn s c = Term.it_mkLambda_or_LetIn c s
 let my_it_mkProd_or_LetIn s c = Term.it_mkProd_or_LetIn c s
 let my_it_mkLambda_or_LetIn_name env s c =
   let mkLambda_or_LetIn_name d b = mkLambda_or_LetIn (name_assumption env d) b in
@@ -141,7 +141,7 @@ let get_sym_eq_data env (ind,u) =
   if List.exists is_local_def realsign then
     error "Inductive equalities with local definitions in arity not supported.";
   let constrsign,ccl = mip.mind_nf_lc.(0) in
-  let _,constrargs = decompose_app ccl in
+  let _,constrargs = decompose_app_list ccl in
   if not (Int.equal (Context.Rel.length constrsign) (Context.Rel.length mib.mind_params_ctxt)) then
     error "Constructor must have no arguments"; (* This can be relaxed... *)
   let params,constrargs = List.chop mib.mind_nparams constrargs in
@@ -176,7 +176,7 @@ let get_non_sym_eq_data env (ind,u) =
   if List.exists is_local_def realsign then
     error "Inductive equalities with local definitions in arity not supported";
   let constrsign,ccl = mip.mind_nf_lc.(0) in
-  let _,constrargs = decompose_app ccl in
+  let _,constrargs = decompose_app_list ccl in
   if not (Int.equal (Context.Rel.length constrsign) (Context.Rel.length mib.mind_params_ctxt)) then
     error "Constructor must have no arguments";
   let _,constrargs = List.chop mib.mind_nparams constrargs in
@@ -649,7 +649,7 @@ let fix_r2l_forward_rew_scheme env (c, ctx') =
   let sigma = Evd.from_env env in
   let t = Retyping.get_type_of env sigma (EConstr.of_constr c) in
   let t = EConstr.Unsafe.to_constr t in
-  let ctx,_ = decompose_prod_assum t in
+  let ctx,_ = decompose_prod_decls t in
   match ctx with
   | hp :: p :: ind :: indargs ->
      let c' =
@@ -687,8 +687,9 @@ let fix_r2l_forward_rew_scheme env (c, ctx') =
 let build_r2l_rew_scheme dep env ind k =
   let sigma = Evd.from_env env in
   let (sigma, indu) = Evd.fresh_inductive_instance env sigma ind in
-  let (sigma, c, _) = build_case_analysis_scheme env sigma indu dep k in
-    c, Evd.evar_universe_context sigma
+  let (sigma, c) = build_case_analysis_scheme env sigma indu dep k in
+  let (c, _) = Indrec.eval_case_analysis c in
+  EConstr.Unsafe.to_constr c, Evd.evar_universe_context sigma
 
 (**********************************************************************)
 (* Register the rewriting schemes                                     *)
@@ -799,7 +800,7 @@ let build_congr env (eq,refl,ctx) ind =
     RelDecl.get_type decl, RelDecl.get_relevance decl
   in
   let constrsign,ccl = mip.mind_nf_lc.(0) in
-  let _,constrargs = decompose_app ccl in
+  let _,constrargs = decompose_app_list ccl in
   if not (Int.equal (Context.Rel.length constrsign) (Context.Rel.length mib.mind_params_ctxt)) then
     error "Constructor must have no arguments";
   let b = List.nth constrargs (i + mib.mind_nparams - 1) in
