@@ -44,7 +44,7 @@ let meta2 = mkmeta 2
 let match_with_non_recursive_type env sigma t =
   match EConstr.kind sigma t with
     | App _ ->
-        let (hdapp,args) = decompose_app sigma t in
+        let (hdapp,args) = decompose_app_list sigma t in
         (match EConstr.kind sigma hdapp with
            | Ind (ind,u) ->
                if (Environ.lookup_mind (fst ind) env).mind_finite == CoFinite then
@@ -85,8 +85,6 @@ let is_lax_conjunction = function
 | Some false -> true
 | _ -> false
 
-let prod_assum sigma t = fst (decompose_prod_assum sigma t)
-
 (* whd_beta normalize the types of arguments in a product *)
 let rec whd_beta_prod env sigma c = match EConstr.kind sigma c with
   | Prod (n,t,c) -> mkProd (n,Reductionops.whd_beta env sigma t,whd_beta_prod env sigma c)
@@ -94,7 +92,7 @@ let rec whd_beta_prod env sigma c = match EConstr.kind sigma c with
   | _ -> c
 
 let match_with_one_constructor env sigma style onlybinary allow_rec t =
-  let (hdapp,args) = decompose_app sigma t in
+  let (hdapp,args) = decompose_app_list sigma t in
   let res = match EConstr.kind sigma hdapp with
   | Ind ind ->
       let (mib,mip) = Inductive.lookup_mind_specif env (fst ind) in
@@ -120,8 +118,8 @@ let match_with_one_constructor env sigma style onlybinary allow_rec t =
           let ctx, cty = mip.mind_nf_lc.(0) in
           let cty = EConstr.of_constr (Term.it_mkProd_or_LetIn cty ctx) in
           let ctyp = whd_beta_prod env sigma
-            (Termops.prod_applist_assum sigma (Context.Rel.length mib.mind_params_ctxt) cty args) in
-          let cargs = List.map RelDecl.get_type (prod_assum sigma ctyp) in
+            (Termops.prod_applist_decls sigma (Context.Rel.length mib.mind_params_ctxt) cty args) in
+          let cargs = List.map RelDecl.get_type (EConstr.prod_decls sigma ctyp) in
           if not (is_lax_conjunction style) || has_nodep_prod env sigma ctyp then
             (* Record or non strict conjunction *)
             Some (hdapp,List.rev cargs)
@@ -172,7 +170,7 @@ let test_strict_disjunction (mib, mip) =
   Array.for_all_i check 0 mip.mind_nf_lc
 
 let match_with_disjunction ?(strict=false) ?(onlybinary=false) env sigma t =
-  let (hdapp,args) = decompose_app sigma t in
+  let (hdapp,args) = decompose_app_list sigma t in
   let res = match EConstr.kind sigma hdapp with
   | Ind (ind,u)  ->
       let car = constructors_nrealargs env ind in
@@ -263,7 +261,7 @@ let mkGApp f args = DAst.make @@ GApp (f, args)
 let mkGHole = DAst.make @@
   GHole (QuestionMark {
         Evar_kinds.default_question_mark with Evar_kinds.qm_obligation=Define false;
-  }, Namegen.IntroAnonymous, None)
+  }, Namegen.IntroAnonymous)
 let mkGProd id c1 c2 = DAst.make @@
   GProd (Name (Id.of_string id), Explicit, c1, c2)
 let mkGArrow c1 c2 = DAst.make @@
@@ -332,7 +330,7 @@ let is_inductive_equality env ind =
   Int.equal nconstr 1 && Int.equal (constructor_nrealargs env (ind,1)) 0
 
 let match_with_equality_type env sigma t =
-  let (hdapp,args) = decompose_app sigma t in
+  let (hdapp,args) = decompose_app_list sigma t in
   match EConstr.kind sigma hdapp with
   | Ind (ind,_) when is_inductive_equality env ind -> Some (hdapp,args)
   | _ -> None
@@ -376,7 +374,7 @@ let match_with_forall_term env sigma c =
 let is_forall_term env sigma c = Option.has_some (match_with_forall_term env sigma c)
 
 let match_with_nodep_ind env sigma t =
-  let (hdapp,args) = decompose_app sigma t in
+  let (hdapp,args) = decompose_app_list sigma t in
   match EConstr.kind sigma hdapp with
   | Ind (ind, _)  ->
      let (mib,mip) = Inductive.lookup_mind_specif env ind in
@@ -396,7 +394,7 @@ let match_with_nodep_ind env sigma t =
 let is_nodep_ind env sigma t = Option.has_some (match_with_nodep_ind env sigma t)
 
 let match_with_sigma_type env sigma t =
-  let (hdapp,args) = decompose_app sigma t in
+  let (hdapp,args) = decompose_app_list sigma t in
   match EConstr.kind sigma hdapp with
   | Ind (ind, _) ->
      let (mib,mip) = Inductive.lookup_mind_specif env ind in
@@ -425,17 +423,17 @@ let rec first_match matcher = function
 
 (*** Equality *)
 
-let match_eq sigma eqn (ref, hetero) =
+let match_eq env sigma eqn (ref, hetero) =
   let ref =
     try Lazy.force ref
     with e when CErrors.noncritical e -> raise PatternMatchingFailure
   in
   match EConstr.kind sigma eqn with
   | App (c, [|t; x; y|]) ->
-    if not hetero && isRefX sigma ref c then PolymorphicLeibnizEq (t, x, y)
+    if not hetero && isRefX env sigma ref c then PolymorphicLeibnizEq (t, x, y)
     else raise PatternMatchingFailure
   | App (c, [|t; x; t'; x'|]) ->
-    if hetero && isRefX sigma ref c then HeterogenousEq (t, x, t', x')
+    if hetero && isRefX env sigma ref c then HeterogenousEq (t, x, t', x')
     else raise PatternMatchingFailure
   | _ -> raise PatternMatchingFailure
 
@@ -447,8 +445,8 @@ let equalities =
    (lazy(lib_ref "core.JMeq.type"), true), check_jmeq_loaded, build_coq_jmeq_data;
    (lazy(lib_ref "core.identity.type"), false), no_check, build_coq_identity_data]
 
-let find_eq_data sigma eqn = (* fails with PatternMatchingFailure *)
-  let d,k = first_match (match_eq sigma eqn) equalities in
+let find_eq_data env sigma eqn = (* fails with PatternMatchingFailure *)
+  let d,k = first_match (match_eq env sigma eqn) equalities in
   let hd,u = destInd sigma (fst (destApp sigma eqn)) in
     d,u,k
 
@@ -461,13 +459,13 @@ let extract_eq_args env sigma = function
       else raise PatternMatchingFailure
 
 let find_eq_data_decompose env sigma eqn =
-  let (lbeq,u,eq_args) = find_eq_data sigma eqn in
+  let (lbeq,u,eq_args) = find_eq_data env sigma eqn in
   (lbeq,u,extract_eq_args env sigma eq_args)
 
 let find_this_eq_data_decompose env sigma eqn =
   let (lbeq,u,eq_args) =
     try (*first_match (match_eq eqn) inversible_equalities*)
-      find_eq_data sigma eqn
+      find_eq_data env sigma eqn
     with PatternMatchingFailure ->
       user_err  (str "No primitive equality found.") in
   let eq_args =
@@ -480,9 +478,9 @@ let find_this_eq_data_decompose env sigma eqn =
 
 let match_sigma env sigma ex =
   match EConstr.kind sigma ex with
-  | App (f, [| a; p; car; cdr |]) when isRefX sigma (lib_ref "core.sig.intro") f ->
+  | App (f, [| a; p; car; cdr |]) when is_lib_ref env sigma "core.sig.intro" f ->
       build_sigma (), (snd (destConstruct sigma f), a, p, car, cdr)
-  | App (f, [| a; p; car; cdr |]) when isRefX sigma (lib_ref "core.sigT.intro") f ->
+  | App (f, [| a; p; car; cdr |]) when is_lib_ref env sigma "core.sigT.intro" f ->
     build_sigma_type (), (snd (destConstruct sigma f), a, p, car, cdr)
   | _ -> raise PatternMatchingFailure
 
