@@ -16,7 +16,6 @@ open Constr
 open Context
 open EConstr
 open Vars
-open Reduction
 open Tacticals
 open Tactics
 open Pretype_errors
@@ -32,9 +31,6 @@ open Libnames
 open Proofview.Notations
 open Context.Named.Declaration
 
-module NamedDecl = Context.Named.Declaration
-(* module RelDecl = Context.Rel.Declaration *)
-
 module TC = Typeclasses
 
 (** Typeclass-based generalized rewriting. *)
@@ -98,7 +94,7 @@ let extends_undefined evars evars' =
 
 let app_poly_check env evars f args =
   let (evars, cstrs), fc = f env evars in
-  let evars, t = Typing.solve_evars env evars (mkApp (fc, args)) in
+  let evars, t = Typing.checked_appvect env evars fc args in
   (evars, cstrs), t
 
 let app_poly_nocheck env evars f args =
@@ -143,7 +139,7 @@ type hypinfo = {
   sort : bool; (* true = Prop; false = Type *)
   c1 : constr;
   c2 : constr;
-  holes : Clenv.hole list;
+  holes : EClause.hole list;
 }
 
 let error_no_relation () = user_err Pp.(str "Cannot find a relation to rewrite.")
@@ -192,16 +188,16 @@ let decompose_applied_relation env sigma (c,l) =
   let open Context.Rel.Declaration in
   let ctype = Retyping.get_type_of env sigma c in
   let find_rel ty =
-    let sigma, cl = Clenv.make_evar_clause env sigma ty in
-    let sigma = Clenv.solve_evar_clause env sigma true cl l in
-    let { Clenv.cl_holes = holes; Clenv.cl_concl = t } = cl in
+    let sigma, cl = EClause.make_evar_clause env sigma ty in
+    let sigma = EClause.solve_evar_clause env sigma true cl l in
+    let { EClause.cl_holes = holes; EClause.cl_concl = t } = cl in
     match decompose_app_rel env sigma t with
     | None -> None
     | Some (equiv, ty1, ty2, concl, c1, c2) ->
       match evd_convertible env sigma ty1 ty2 with
       | None -> None
       | Some sigma ->
-        let args = Array.map_of_list (fun h -> h.Clenv.hole_evar) holes in
+        let args = Array.map_of_list (fun h -> h.EClause.hole_evar) holes in
         let value = mkApp (c, args) in
           Some (sigma, { prf=value;
                   car=ty1; rel = equiv; sort = Sorts.is_prop (ESorts.kind sigma concl);
@@ -210,7 +206,7 @@ let decompose_applied_relation env sigma (c,l) =
     match find_rel ctype with
     | Some c -> c
     | None ->
-      let ctx,t' = Reductionops.splay_prod env sigma ctype in (* Search for underlying eq *)
+      let ctx,t' = Reductionops.hnf_decompose_prod env sigma ctype in (* Search for underlying eq *)
       let t' = it_mkProd_or_LetIn t' (List.map (fun (n,t) -> LocalAssum (n, t)) ctx) in
       match find_rel t' with
       | Some c -> c
@@ -258,22 +254,22 @@ end) = struct
 
   let proper_class =
     let r = lazy (find_reference morphisms "Proper") in
-    fun env sigma -> TC.class_info env sigma (Lazy.force r)
+    fun () -> Option.get (TC.class_info (Lazy.force r))
 
   let proper_proxy_class =
     let r = lazy (find_reference morphisms "ProperProxy") in
-    fun env sigma -> TC.class_info env sigma (Lazy.force r)
+    fun () -> Option.get (TC.class_info (Lazy.force r))
 
-  let proper_proj env sigma =
-    mkConst (Option.get (List.hd (proper_class env sigma).TC.cl_projs).TC.meth_const)
+  let proper_proj () =
+    UnsafeMonomorphic.mkConst (Option.get (List.hd (proper_class ()).TC.cl_projs).TC.meth_const)
 
   let proper_type env (sigma,cstrs) =
-    let l = (proper_class env sigma).TC.cl_impl in
+    let l = (proper_class ()).TC.cl_impl in
     let (sigma, c) = Evd.fresh_global env sigma l in
     (sigma, cstrs), c
 
   let proper_proxy_type env (sigma,cstrs) =
-    let l = (proper_proxy_class env sigma).TC.cl_impl in
+    let l = (proper_proxy_class ()).TC.cl_impl in
     let (sigma, c) = Evd.fresh_global env sigma l in
     (sigma, cstrs), c
 
@@ -373,23 +369,23 @@ end) = struct
       else (* None in Prop, use arrow *)
         (app_poly env evd arrow [| a; b |]), unfold_impl n
 
-  let rec decomp_pointwise sigma n c =
+  let rec decomp_pointwise env sigma n c =
     if Int.equal n 0 then c
     else
       match EConstr.kind sigma c with
-      | App (f, [| a; b; relb |]) when isRefX sigma (pointwise_relation_ref ()) f ->
-        decomp_pointwise sigma (pred n) relb
-      | App (f, [| a; b; arelb |]) when isRefX sigma (forall_relation_ref ()) f ->
-        decomp_pointwise sigma (pred n) (Reductionops.beta_applist sigma (arelb, [mkRel 1]))
+      | App (f, [| a; b; relb |]) when isRefX env sigma (pointwise_relation_ref ()) f ->
+        decomp_pointwise env sigma (pred n) relb
+      | App (f, [| a; b; arelb |]) when isRefX env sigma (forall_relation_ref ()) f ->
+        decomp_pointwise env sigma (pred n) (Reductionops.beta_applist sigma (arelb, [mkRel 1]))
       | _ -> invalid_arg "decomp_pointwise"
 
-  let rec apply_pointwise sigma rel = function
+  let rec apply_pointwise env sigma rel = function
     | arg :: args ->
       (match EConstr.kind sigma rel with
-      | App (f, [| a; b; relb |]) when isRefX sigma (pointwise_relation_ref ()) f ->
-        apply_pointwise sigma relb args
-      | App (f, [| a; b; arelb |]) when isRefX sigma (forall_relation_ref ()) f ->
-        apply_pointwise sigma (Reductionops.beta_applist sigma (arelb, [arg])) args
+      | App (f, [| a; b; relb |]) when isRefX env sigma (pointwise_relation_ref ()) f ->
+        apply_pointwise env sigma relb args
+      | App (f, [| a; b; arelb |]) when isRefX env sigma (forall_relation_ref ()) f ->
+        apply_pointwise env sigma (Reductionops.beta_applist sigma (arelb, [arg])) args
       | _ -> invalid_arg "apply_pointwise")
     | [] -> rel
 
@@ -444,14 +440,14 @@ end) = struct
 
   let unlift_cstr env sigma = function
     | None -> None
-    | Some codom -> Some (decomp_pointwise (goalevars sigma) 1 codom)
+    | Some codom -> Some (decomp_pointwise env (goalevars sigma) 1 codom)
 
   (** Looking up declared rewrite relations (instances of [RewriteRelation]) *)
   let is_applied_rewrite_relation env sigma rels t =
     match EConstr.kind sigma t with
     | App (c, args) when Array.length args >= 2 ->
       let head = if isApp sigma c then fst (destApp sigma c) else c in
-        if isRefX sigma (coq_eq_ref ()) head then None
+        if isRefX env sigma (coq_eq_ref ()) head then None
         else
           (try
             let env' = push_rel_context rels env in
@@ -622,8 +618,8 @@ let solve_remaining_by env sigma holes by =
   | None -> sigma
   | Some tac ->
     let map h =
-      if h.Clenv.hole_deps then None
-      else match EConstr.kind sigma h.Clenv.hole_evar with
+      if h.EClause.hole_deps then None
+      else match EConstr.kind sigma h.EClause.hole_evar with
       | Evar (evk, _) ->
         Some evk
       | _ -> None
@@ -647,8 +643,8 @@ let solve_remaining_by env sigma holes by =
       | None -> sigma
         (* Evar should not be defined, but just in case *)
       | Some evi ->
-        let env = Environ.reset_with_named_context evi.evar_hyps env in
-        let ty = evi.evar_concl in
+        let env = Evd.evar_env env evi in
+        let ty = Evd.evar_concl evi in
         let name, poly = Id.of_string "rewrite", false in
         let c, sigma = Proof.refine_by_tactic ~name ~poly env sigma ty solve_tac in
         Evd.define evk (EConstr.of_constr c) sigma
@@ -722,7 +718,7 @@ let unify_eqn (car, rel, prf, c1, c2, holes, sort) l2r flags env (sigma, cstrs)
     let sigma = TC.resolve_typeclasses ~filter:(no_constraints cstrs)
       ~fail:true env sigma in
     let sigma = solve_remaining_by env sigma holes by in
-    let nf c = Reductionops.nf_evar sigma (Reductionops.nf_meta env sigma c) in
+    let nf c = Reductionops.nf_evar sigma c in
     let c1 = nf c1 and c2 = nf c2
     and rew_car = nf car and rel = nf rel
     and prf = nf prf in
@@ -784,7 +780,7 @@ let poly_subrelation sort =
   if sort then PropGlobal.subrelation else TypeGlobal.subrelation
 
 let resolve_subrelation env car rel sort prf rel' res =
-  if Termops.eq_constr (fst res.rew_evars) rel rel' then res
+  if Termops.eq_constr env (fst res.rew_evars) rel rel' then res
   else
     let evars, app = app_poly_check env res.rew_evars (poly_subrelation sort) [|car; rel; rel'|] in
     let evars, subrel = new_cstr_evar evars env app in
@@ -828,7 +824,7 @@ let resolve_morphism env m args args' (b,cstr) evars =
     let env' = EConstr.push_named (LocalDef (make_annot dosub_id Sorts.Relevant, dosub, appsub)) env in
     let evars, morph = new_cstr_evar evars env' app in
     (* Replace the free [dosub_id] in the evar by the global reference *)
-    let morph = Vars.replace_vars [dosub_id , dosub] morph in
+    let morph = Vars.replace_vars (fst evars) [dosub_id , dosub] morph in
     evars, morph, sigargs, appm, morphobjs, morphobjs'
   in
   let projargs, subst, evars, respars, typeargs =
@@ -880,7 +876,7 @@ let apply_rule unify : occurrences_count pure_strategy =
         | Some rew ->
           let b, occs = update_occurrence_counter occs in
             if not b then (occs, Fail)
-            else if Termops.eq_constr (fst rew.rew_evars) t rew.rew_to then (occs, Identity)
+            else if Termops.eq_constr env (fst rew.rew_evars) t rew.rew_to then (occs, Identity)
             else
               let res = { rew with rew_car = ty } in
               let res = Success (coerce env cstr res) in
@@ -935,7 +931,7 @@ let fold_match ?(force=false) env sigma c =
   let cty = Retyping.get_type_of env sigma c in
   let dep, pred, sk =
     let env', ctx, body =
-      let ctx, pred = decompose_lam_assum sigma p in
+      let ctx, pred = decompose_lambda_decls sigma p in
       let env' = push_rel_context ctx env in
         env', ctx, pred
     in
@@ -966,12 +962,16 @@ let fold_match ?(force=false) env sigma c =
       raise Not_found
   in
   let app =
+    let sk = if Global.is_polymorphic (ConstRef sk)
+      then CErrors.anomaly Pp.(str "Unexpected univ poly in Rewrite.fold_match")
+      else UnsafeMonomorphic.mkConst sk
+    in
     let ind, args = Inductiveops.find_mrectype env sigma cty in
     let pars, args = List.chop ci.ci_npar args in
     let meths = Array.to_list brs in
-      applist (mkConst sk, pars @ [pred] @ meths @ args @ [c])
+      applist (sk, pars @ [pred] @ meths @ args @ [c])
   in
-    sk, env, app
+    sk, app
 
 let unfold_match env sigma sk app =
   match EConstr.kind sigma app with
@@ -1067,7 +1067,7 @@ let subterm all flags (s : 'a pure_strategy) : 'a pure_strategy =
                         let app = if prop then PropGlobal.apply_pointwise
                           else TypeGlobal.apply_pointwise
                         in
-                          RewPrf (app (goalevars evars) rel argsl, mkApp (prf, args))
+                          RewPrf (app env (goalevars evars) rel argsl, mkApp (prf, args))
                       | x -> x
                     in
                     let res =
@@ -1205,7 +1205,7 @@ let subterm all flags (s : 'a pure_strategy) : 'a pure_strategy =
             else
               match try Some (fold_match env (goalevars evars) t) with Not_found -> None with
               | None -> state, c'
-              | Some (cst, _, t') ->
+              | Some (cst, t') ->
                  let state, res = aux { state ; env ; unfresh ;
                                         term1 = t' ; ty1 = ty ;
                                         cstr = (prop,cstr) ; evars } in
@@ -1369,22 +1369,23 @@ module Strategies =
         fail cs
 
     let inj_open hint = (); fun sigma ->
-      let ctx = UState.of_context_set hint.Autorewrite.rew_ctx in
-      let sigma = Evd.merge_universe_context sigma ctx in
-      (sigma, (EConstr.of_constr hint.Autorewrite.rew_lemma, NoBindings))
+      let (ctx, lemma) = Autorewrite.RewRule.rew_lemma hint in
+      (* not sure sideff:true is really needed here *)
+      let sigma = Evd.merge_context_set ~sideff:true UnivRigid sigma ctx in
+      (sigma, (EConstr.of_constr lemma, NoBindings))
 
     let old_hints (db : string) : 'a pure_strategy =
       let rules = Autorewrite.find_rewrites db in
         lemmas
-          (List.map (fun hint -> (inj_open hint, hint.Autorewrite.rew_l2r,
-                                  hint.Autorewrite.rew_tac)) rules)
+          (List.map (fun hint -> (inj_open hint, Autorewrite.RewRule.rew_l2r hint,
+                                  Autorewrite.RewRule.rew_tac hint)) rules)
 
     let hints (db : string) : 'a pure_strategy = { strategy =
-      fun ({ term1 = t } as input) ->
+      fun ({ term1 = t; env } as input) ->
       let t = EConstr.Unsafe.to_constr t in
-      let rules = Autorewrite.find_matches db t in
-      let lemma hint = (inj_open hint, hint.Autorewrite.rew_l2r,
-                        hint.Autorewrite.rew_tac) in
+      let rules = Autorewrite.find_matches env db t in
+      let lemma hint = (inj_open hint, Autorewrite.RewRule.rew_l2r hint,
+                        Autorewrite.RewRule.rew_tac hint) in
       let lems = List.map lemma rules in
       (lemmas lems).strategy input
                                                  }
@@ -1394,7 +1395,7 @@ module Strategies =
           let rfn, ckind = Redexpr.reduction_of_red_expr env r in
           let sigma = goalevars evars in
           let (sigma, t') = rfn env sigma t in
-            if Termops.eq_constr sigma t' t then
+            if Termops.eq_constr env sigma t' t then
               state, Identity
             else
               state, Success { rew_car = ty; rew_from = t; rew_to = t';
@@ -1447,7 +1448,7 @@ let rewrite_with l2r flags c occs : strategy = { strategy =
                                                }
 
 let apply_strategy (s : strategy) env unfresh concl (prop, cstr) evars =
-  let ty = Retyping.get_type_of env (goalevars evars) concl in
+  let evars, ty = get_type_of_refresh env evars concl in
   let _, res = s.strategy { state = () ; env ; unfresh ;
                             term1 = concl ; ty1 = ty ;
                             cstr = (prop, Some cstr) ; evars } in
@@ -1456,7 +1457,7 @@ let apply_strategy (s : strategy) env unfresh concl (prop, cstr) evars =
 let solve_constraints env (evars,cstrs) =
   let oldtcs = Evd.get_typeclass_evars evars in
   let evars' = Evd.set_typeclass_evars evars cstrs in
-  let evars' = TC.resolve_typeclasses env ~filter:TC.all_evars ~split:false ~fail:true evars' in
+  let evars' = TC.resolve_typeclasses env ~filter:TC.all_evars ~fail:true evars' in
   Evd.set_typeclass_evars evars' oldtcs
 
 let nf_zeta =
@@ -1466,13 +1467,22 @@ exception RewriteFailure of Environ.env * Evd.evar_map * pretype_error
 
 type result = (evar_map * constr option * types) option option
 
+exception UnsolvedConstraints of Environ.env * Evd.evar_map * Evar.t
+
+let () = CErrors.register_handler begin function
+| UnsolvedConstraints (env, evars, ev) ->
+  Some (str "Unsolved constraint remaining: " ++ spc () ++
+    Termops.pr_evar_info env evars (Evd.find_undefined evars ev) ++ str ".")
+| _ -> None
+end
+
 let cl_rewrite_clause_aux ?(abs=None) strat env avoid sigma concl is_hyp : result =
   let sigma, sort = Typing.sort_of env sigma concl in
   let evdref = ref sigma in
   let evars = (!evdref, Evar.Set.empty) in
   let evars, cstr =
     let prop, (evars, arrow) =
-      if Sorts.is_prop sort then true, app_poly_sort true env evars impl [||]
+      if ESorts.is_prop sigma sort then true, app_poly_sort true env evars impl [||]
       else false, app_poly_sort false env evars TypeGlobal.arrow [||]
     in
     match is_hyp with
@@ -1488,15 +1498,8 @@ let cl_rewrite_clause_aux ?(abs=None) strat env avoid sigma concl is_hyp : resul
   | Success res ->
     let (_, cstrs) = res.rew_evars in
     let evars = solve_constraints env res.rew_evars in
-    let () =
-      Evar.Set.iter
-        (fun ev ->
-           if not (Evd.is_defined evars ev) then
-             user_err
-               (str "Unsolved constraint remaining: " ++ spc () ++
-                Termops.pr_evar_info env evars (Evd.find evars ev) ++ str "."))
-        cstrs
-    in
+    let iter ev = if not (Evd.is_defined evars ev) then raise (UnsolvedConstraints (env, evars, ev)) in
+    let () = Evar.Set.iter iter cstrs in
     let newt = res.rew_to in
     let res = match res.rew_prf with
       | RewCast c -> None
@@ -1516,39 +1519,8 @@ let cl_rewrite_clause_aux ?(abs=None) strat env avoid sigma concl is_hyp : resul
     in
     Some (Some (evars, res, newt))
 
-(** Insert a declaration after the last declaration it depends on *)
-let rec insert_dependent env sigma decl accu hyps = match hyps with
-| [] -> List.rev_append accu [decl]
-| ndecl :: rem ->
-  if occur_var_in_decl env sigma (NamedDecl.get_id ndecl) decl then
-    List.rev_append accu (decl :: hyps)
-  else
-    insert_dependent env sigma decl (ndecl :: accu) rem
-
 let assert_replacing id newt tac =
-  let prf = Proofview.Goal.enter begin fun gl ->
-    let concl = Proofview.Goal.concl gl in
-    let env = Proofview.Goal.env gl in
-    let sigma = Tacmach.project gl in
-    let ctx = named_context env in
-    let after, before = List.split_when (NamedDecl.get_id %> Id.equal id) ctx in
-    let nc = match before with
-    | [] -> assert false
-    | d :: rem -> insert_dependent env sigma
-                    (LocalAssum (make_annot (NamedDecl.get_id d) Sorts.Relevant, newt)) [] after @ rem
-    in
-    let env' = Environ.reset_with_named_context (val_of_named_context nc) env in
-    Refine.refine ~typecheck:true begin fun sigma ->
-      let (sigma, ev) = Evarutil.new_evar env' sigma concl in
-      let (sigma, ev') = Evarutil.new_evar env sigma newt in
-      let map d =
-        let n = NamedDecl.get_id d in
-        if Id.equal n id then ev' else mkVar n
-      in
-      let (e, _) = destEvar sigma ev in
-      (sigma, mkEvar (e, List.map map nc))
-    end
-  end in
+  let prf = Tactics.assert_after_replacing id newt in
   Proofview.tclTHEN prf (Proofview.tclFOCUS 2 2 tac)
 
 let newfail n s =
@@ -1783,12 +1755,12 @@ let rec strategy_of_ast = function
                                              evars = (sigma,cstrevars evars) }) }
   | StratFold c -> Strategies.fold_glob (fst c)
 
-let proper_projection env sigma r ty =
+let proper_projection sigma r ty =
   let rel_vect n m = Array.init m (fun i -> mkRel(n+m-i)) in
-  let ctx, inst = decompose_prod_assum sigma ty in
+  let ctx, inst = decompose_prod_decls sigma ty in
   let mor, args = destApp sigma inst in
   let instarg = mkApp (r, rel_vect 0 (List.length ctx)) in
-  let app = mkApp (PropGlobal.proper_proj env sigma,
+  let app = mkApp (PropGlobal.proper_proj (),
                   Array.append args [| instarg |]) in
     it_mkLambda_or_LetIn app ctx
 
@@ -1827,35 +1799,10 @@ let default_morphism env sigma sign m =
   in
   let evars, morph = app_poly_check env evars PropGlobal.proper_type [| t; sign; m |] in
   let evars, mor = TC.resolve_one_typeclass env (goalevars evars) morph in
-    mor, proper_projection env sigma mor morph
+    mor, proper_projection sigma mor morph
 
 (** Bind to "rewrite" too *)
 
-(** Taken from original setoid_replace, to emulate the old rewrite semantics where
-    lemmas are first instantiated and then rewrite proceeds. *)
-
-let check_evar_map_of_evars_defs env evd =
- let metas = Evd.meta_list evd in
- let check_freemetas_is_empty rebus =
-  Evd.Metaset.iter
-   (fun m ->
-     if Evd.meta_defined evd m then ()
-     else begin
-       raise
-         (Logic.RefinerError (env, evd, Logic.UnresolvedBindings [Evd.meta_name evd m]))
-     end)
- in
-  List.iter
-   (fun (_,binding) ->
-     match binding with
-        Evd.Cltyp (_,{Evd.rebus=rebus; Evd.freemetas=freemetas}) ->
-         check_freemetas_is_empty rebus freemetas
-      | Evd.Clval (_,({Evd.rebus=rebus1; Evd.freemetas=freemetas1},_),
-                 {Evd.rebus=rebus2; Evd.freemetas=freemetas2}) ->
-         check_freemetas_is_empty rebus1 freemetas1 ;
-         check_freemetas_is_empty rebus2 freemetas2
-   ) metas
-
 (* Find a subterm which matches the pattern to rewrite for "rewrite" *)
 let unification_rewrite l2r c1 c2 sigma prf car rel but env =
   let (sigma,c') =
@@ -1878,7 +1825,6 @@ let unification_rewrite l2r c1 c2 sigma prf car rel but env =
   let c1 = if l2r then nf c' else nf c1
   and c2 = if l2r then nf c2 else nf c'
   and car = nf car and rel = nf rel in
-  check_evar_map_of_evars_defs env sigma;
   let prf = nf prf in
   let prfty = nf (Retyping.get_type_of env sigma prf) in
   let sort = sort_of_rel env sigma but in
@@ -1928,10 +1874,18 @@ let _ = Hook.set Equality.general_setoid_rewrite_clause general_s_rewrite
 
 (** [setoid_]{reflexivity,symmetry,transitivity} tactics *)
 
-let not_declared ~info env sigma ty rel =
-  tclFAIL ~info
-    (str" The relation " ++ Printer.pr_econstr_env env sigma rel ++ str" is not a declared " ++
-     str ty ++ str" relation. Maybe you need to require the Coq.Classes.RelationClasses library")
+exception RelationNotDeclared of Environ.env * Evd.evar_map * string * EConstr.types
+
+let () = CErrors.register_handler begin function
+| RelationNotDeclared (env, sigma, ty, concl) ->
+  let rel, _, _, _, _, _ = decompose_app_rel_error env sigma concl in
+  Some (str" The relation " ++ Printer.pr_econstr_env env sigma rel ++ str" is not a declared " ++
+    str ty ++ str" relation. Maybe you need to require the Coq.Classes.RelationClasses library")
+| _ -> None
+end
+
+let not_declared ~info env sigma ty concl =
+  Proofview.tclZERO ~info (RelationNotDeclared (env, sigma, ty, concl))
 
 let setoid_proof ty fn fallback =
   Proofview.Goal.enter begin fun gl ->
@@ -1943,7 +1897,7 @@ let setoid_proof ty fn fallback =
         try
           let rel, ty1, ty2, concl, _, _ = decompose_app_rel_error env sigma concl in
           let (sigma, t) = Typing.type_of env sigma rel in
-          let car = snd (List.hd (fst (Reductionops.splay_prod env sigma t))) in
+          let car = snd (List.hd (fst (Reductionops.hnf_decompose_prod env sigma t))) in
             (try init_relation_classes () with _ -> raise Not_found);
             fn env sigma car rel
         with e when CErrors.noncritical e ->
@@ -1958,9 +1912,7 @@ let setoid_proof ty fn fallback =
               begin function (e', info) -> match e' with
                 | Hipattern.NoEquationFound ->
                     begin match e with
-                    | (Not_found, _) ->
-                      let rel, _, _, _, _, _ = decompose_app_rel_error env sigma concl in
-                      not_declared ~info env sigma ty rel
+                    | (Not_found, _) -> not_declared ~info env sigma ty concl
                     | (e, info) ->
                       Proofview.tclZERO ~info e
                     end
@@ -2010,8 +1962,8 @@ let setoid_symmetry_in id =
   let env = Proofview.Goal.env gl in
   let sigma = Proofview.Goal.sigma gl in
   let ctype = Retyping.get_type_of env sigma (mkVar id) in
-  let binders,concl = decompose_prod_assum sigma ctype in
-  let (equiv, args) = decompose_app sigma concl in
+  let binders,concl = decompose_prod_decls sigma ctype in
+  let (equiv, args) = decompose_app_list sigma concl in
   let rec split_last_two = function
     | [c1;c2] -> [],(c1, c2)
     | x::y::z -> let l,res = split_last_two (y::z) in x::l, res
