@@ -18,14 +18,22 @@ sort > desired <<EOT
 ./test/.coq-native/Ntest_test.cmx
 ./test/.coq-native/Ntest_test.cmxs
 ./test/test.glob
-./test/test.cmi
-./test/test.cmx
-./test/test_aux.cmi
-./test/test_aux.cmx
-./test/test_plugin.cmxa
-./test/test_plugin.cmxs
 ./test/test.v
 ./test/test.vo
+./test/test_plugin.cmxs
 EOT
 (coqc -config | grep -q "NATIVE_COMPILER_DEFAULT=yes") || sed -i.bak '/\.coq-native/d' desired
-exec diff -u desired actual
+diff -u desired actual
+
+(cd "$(find tmp -name coq-test-suite)" && find .) | sort > actual
+sort > desired <<EOT
+.
+./META
+./test.cmi
+./test.cmx
+./test_aux.cmi
+./test_aux.cmx
+./test_plugin.cmxa
+./test_plugin.cmxs
+EOT
+diff -u desired actual
