@@ -15,22 +15,6 @@ open Printf
 
 let (>) f g = fun x -> g (f x)
 
-let output_channel = ref stdout
-let makefile_name = ref "Makefile"
-let make_name = ref ""
-
-let print x = output_string !output_channel x
-let printf x = Printf.fprintf !output_channel x
-
-let rec print_list sep = function
-  | [ x ] -> print x
-  | x :: l -> print x; print sep; print_list sep l
-  | [] -> ()
-
-let rec print_prefix_list sep = function
-  | x :: l -> print sep; print x; print_prefix_list sep l
-  | [] -> ()
-
 let usage_coq_makefile () =
   output_string stderr "Usage summary:\
 \n\
@@ -67,10 +51,6 @@ let usage_coq_makefile () =
 \n[--help]: equivalent to [-h]\n";
   exit 1
 
-let is_genrule r = (* generic rule (like bar%foo: ...) *)
-    let genrule = Str.regexp("%") in
-      Str.string_match genrule r 0
-
 let is_prefix dir1 dir2 =
   let l1 = String.length dir1 in
   let l2 = String.length dir2 in
@@ -158,7 +138,8 @@ let generate_meta_file p =
     match p.meta_file with
     | Absent -> p
     | Generate proj ->
-        let cmname = List.map (fun { thing } -> thing) (p.mllib_files @ p.mlpack_files) in
+        let cmname = List.map (fun { thing } -> thing)
+            (files_by_suffix p.files [".mllib"; ".mlpack"]) in
         let dir, cmname =
           match cmname with
           | [] -> Printf.eprintf "In order to generate a META file one needs an .mlpack or .mllib file\n"; exit 1
@@ -246,28 +227,43 @@ let generate_conf_coq_config oc =
   fprintf oc "COQMF_WINDRIVE=%s\n" (windrive coqlib)
 ;;
 
-let check_metafile { mllib_files; mlpack_files; meta_file; _ } =
-  if mlpack_files @ mllib_files <> [] && meta_file = Absent then begin
+let check_metafile p =
+  if files_by_suffix p.files [".mlpack"; ".mllib"] <> [] && p.meta_file = Absent then begin
     eprintf "Warning: it is recommended you provide a META.package-name file\n";
     eprintf "Warning: since you build plugins. See also -generate-meta-for-package.\n";
   end
 
-let generate_conf_files oc
-  { v_files; mli_files; mlg_files; ml_files; mllib_files; mlpack_files; meta_file }
+let write_coqbin oc =
+  fprintf oc "%s\n"
+    "COQBIN?=\n\
+    ifneq (,$(COQBIN))\n\
+    # add an ending /\n\
+    COQBIN:=$(COQBIN)/\n\
+    endif\n\
+    COQMKFILE ?= \"$(COQBIN)coq_makefile\""
+
+let generate_conf_files oc p
 =
   let module S = String in
-  let map = map_sourced_list in
+  let fout varname suffix =
+    fprintf oc "COQMF_%s := $(filter %%%s, $(COQMF_SOURCES))\n" varname suffix;
+  in
   section oc "Project files.";
-  fprintf oc "COQMF_VFILES = %s\n"      (S.concat " " (map quote v_files));
-  fprintf oc "COQMF_MLIFILES = %s\n"    (S.concat " " (map quote mli_files));
-  fprintf oc "COQMF_MLFILES = %s\n"     (S.concat " " (map quote ml_files));
-  fprintf oc "COQMF_MLGFILES = %s\n"    (S.concat " " (map quote mlg_files));
-  fprintf oc "COQMF_MLPACKFILES = %s\n" (S.concat " " (map quote mlpack_files));
-  fprintf oc "COQMF_MLLIBFILES = %s\n"  (S.concat " " (map quote mllib_files));
-  fprintf oc "COQMF_METAFILE = %s\n"  (match meta_file with Present x -> x | _ -> "");
-  let cmdline_vfiles = filter_cmdline v_files in
-  fprintf oc "COQMF_CMDLINE_VFILES = %s\n" (S.concat " " (List.map quote cmdline_vfiles));
-;;
+
+  let cmdline_vfiles = p.cmd_line_files in
+  fprintf oc "COQMF_CMDLINE_VFILES := %s\n" (S.concat " " (map_sourced_list quote cmdline_vfiles));
+  let proj_arg = match p.project_file with
+    | Some pfile -> Printf.sprintf "-f %s" pfile
+    | None -> ""
+  in
+  fprintf oc "COQMF_SOURCES := $(shell $(COQMKFILE) -sources-of %s $(COQMF_CMDLINE_VFILES))\n" proj_arg;
+  fout "VFILES" ".v";
+  fout "MLIFILES" ".mli";
+  fout "MLFILES" ".ml";
+  fout "MLGFILES" ".mlg";
+  fout "MLPACKFILES" ".mlpack";
+  fout "MLLIBFILES" ".mllib";
+  fprintf oc "COQMF_METAFILE = %s\n"  (match p.meta_file with Present x -> x | _ -> "")
 
 let rec all_start_with prefix = function
   | [] -> true
@@ -320,6 +316,7 @@ let generate_conf_defs oc { defs; extra_args } =
 let generate_conf oc project args  =
   fprintf oc "# This configuration file was generated by running:\n";
   fprintf oc "# %s\n\n" (String.concat " " (List.map quote args));
+  write_coqbin oc;
   generate_conf_files oc project;
   generate_conf_includes oc project;
   generate_conf_coq_config oc;
@@ -329,9 +326,7 @@ let generate_conf oc project args  =
 ;;
 
 let ensure_root_dir
-  ({ ml_includes; r_includes; q_includes;
-     v_files; ml_files; mli_files; mlg_files;
-     mllib_files; mlpack_files } as project)
+  ({ ml_includes; r_includes; q_includes; files } as project)
   =
   let exists f = List.exists (forget_source > f) in
   let here = Sys.getcwd () in
@@ -339,9 +334,7 @@ let ensure_root_dir
   if exists (fun { canonical_path = x } -> x = here) ml_includes
   || exists (fun ({ canonical_path = x },_) -> is_prefix x here) r_includes
   || exists (fun ({ canonical_path = x },_) -> is_prefix x here) q_includes
-  || (not_tops v_files &&
-      not_tops mli_files && not_tops mlg_files &&
-      not_tops ml_files && not_tops mllib_files && not_tops mlpack_files)
+  || not_tops files
   then
     project
   else
@@ -413,14 +406,7 @@ let destination_of { ml_includes; q_includes; r_includes; } file =
   | [s] -> Printf.printf "%s" (quote s)
   | _ -> assert false
 
-let share_prefix s1 s2 =
-  let s1 = String.split_on_char '.' s1 in
-  let s2 = String.split_on_char '.' s2 in
-  match s1, s2 with
-  | x :: _ , y :: _ -> x = y
-  | _ -> false
-
-let _ =
+let () =
   let _fhandle = Feedback.(add_feeder (console_feedback_listener Format.err_formatter)) in
   let prog, args =
     let args = Array.to_list Sys.argv in
@@ -431,6 +417,11 @@ let _ =
     | "-destination-of" :: tgt :: rest -> Some tgt, rest
     | _ -> None, args in
 
+  (* -sources-of and -destination-of must be the first parameter *)
+  let only_sources, args = match args with
+    | "-sources-of" :: rest -> true, rest
+    | _ -> false, args in
+
   let project =
     let warning_fn x = Format.eprintf "%s@\n%!" x in
     try cmdline_args_to_project ~warning_fn ~curdir:Filename.current_dir_name args
@@ -441,6 +432,12 @@ let _ =
     exit 0
   end;
 
+  if only_sources then begin
+    let paths = String.concat " " (List.map (fun i -> i.thing) project.files) in
+    Printf.printf "%s" paths;
+    exit 0
+  end;
+
   if project.makefile = None then
     eprintf "Warning: Omitting -o is deprecated\n\n";
     (* We want to know the name of the Makefile (say m) in order to
