@@ -20,6 +20,11 @@ let create_empty_file filename =
   let f = open_out filename in
   close_out f
 
+let source ldir file = Loc.InFile {
+    dirpath=Some (Names.DirPath.to_string ldir);
+    file = file;
+  }
+
 (* Compile a vernac file *)
 let compile opts stm_options injections copts ~echo ~f_in ~f_out =
   let open Vernac.State in
@@ -48,7 +53,7 @@ let compile opts stm_options injections copts ~echo ~f_in ~f_out =
       let doc, sid = Topfmt.(in_phase ~phase:LoadingPrelude)
           Stm.new_doc
           Stm.{ doc_type = VoDoc long_f_dot_out; injections; } in
-      let state = { doc; sid; proof = None; time = opts.config.time } in
+      let state = { doc; sid; proof = None; time = Option.map Vernac.make_time_output opts.config.time } in
       let state = Load.load_init_vernaculars opts ~state in
       let ldir = Stm.get_ldir ~doc:state.doc in
       Aux_file.(start_aux_file
@@ -61,10 +66,12 @@ let compile opts stm_options injections copts ~echo ~f_in ~f_out =
 
       let wall_clock1 = Unix.gettimeofday () in
       let check = Stm.AsyncOpts.(stm_options.async_proofs_mode = APoff) in
-      let state = Vernac.load_vernac ~echo ~check ~interactive:false ~state ~ldir long_f_dot_in in
-      let doc, state = Stm.join ~doc:state.doc in
+      let source = source ldir long_f_dot_in in
+      let state = Vernac.load_vernac ~echo ~check ~state ~source long_f_dot_in in
+      let fullstate = Stm.finish ~doc:state.doc in
+      ensure_no_pending_proofs ~filename:long_f_dot_in fullstate;
+      let () = Stm.join ~doc:state.doc in
       let wall_clock2 = Unix.gettimeofday () in
-      ensure_no_pending_proofs ~filename:long_f_dot_in state;
       (* In .vo production, dump a complete .vo file. *)
       if mode = BuildVo
         then Library.save_library_to ~output_native_objects Library.ProofsTodoNone ldir long_f_dot_out;
@@ -81,35 +88,22 @@ let compile opts stm_options injections copts ~echo ~f_in ~f_out =
       Dumpglob.end_dump_glob ()
 
   | BuildVio | BuildVos ->
-      (* We need to disable error resiliency, otherwise some errors
-         will be ignored in batch mode. c.f. #6707
-
-         This is not necessary in the vo case as it fully checks the
-         document anyways. *)
-      let stm_options = let open Stm.AsyncOpts in
-        { stm_options with
-          async_proofs_mode = APon;
-          async_proofs_n_workers = 0;
-          async_proofs_cmd_error_resilience = false;
-          async_proofs_tac_error_resilience = FNone;
-        } in
-      Stm.init_process stm_options;
-
       let doc, sid = Topfmt.(in_phase ~phase:LoadingPrelude)
           Stm.new_doc
           Stm.{ doc_type = VioDoc long_f_dot_out; injections;
               } in
 
-      let state = { doc; sid; proof = None; time = opts.config.time } in
+      let state = { doc; sid; proof = None; time = Option.map Vernac.make_time_output opts.config.time } in
       let state = Load.load_init_vernaculars opts ~state in
       let ldir = Stm.get_ldir ~doc:state.doc in
-      let state = Vernac.load_vernac ~echo ~check:false ~interactive:false ~state long_f_dot_in in
-      let doc, state = Stm.finish ~doc:state.doc in
+      let source = source ldir long_f_dot_in in
+      let state = Vernac.load_vernac ~echo ~check:false ~source ~state long_f_dot_in in
+      let state = Stm.finish ~doc:state.doc in
       ensure_no_pending_proofs state ~filename:long_f_dot_in;
       let create_vos = (mode = BuildVos) in
       (* In .vos production, the output .vos file contains compiled statements.
          In .vio production, the output .vio file contains compiled statements and suspended proofs. *)
-      let () = ignore (Stm.snapshot_vio ~create_vos ~doc ~output_native_objects ldir long_f_dot_out) in
+      let () = Stm.snapshot_vio ~create_vos ~doc ~output_native_objects ldir long_f_dot_out in
       Stm.reset_task_queue ();
       (* In .vio production, dump an empty .vos file to indicate that the .vio should be loaded. *)
       (* EJGA: This is problematic in a vio + vio2vo run, as there is
