@@ -76,7 +76,7 @@ let reset_input_buffer doc ic ibuf =
   ibuf.str <- Bytes.empty;
   ibuf.len <- 0;
   ibuf.bols <- [];
-  ibuf.tokens <- Pcoq.Parsable.make (Stream.from (prompt_char doc ic ibuf));
+  ibuf.tokens <- Pcoq.Parsable.make (Gramlib.Stream.from (prompt_char doc ic ibuf));
   ibuf.start <- 0
 
 (* Functions to print underlined locations from an input buffer. *)
@@ -230,7 +230,7 @@ let top_buffer =
     str = Bytes.empty;
     len = 0;
     bols = [];
-    tokens = Pcoq.Parsable.make (Stream.of_list []);
+    tokens = Pcoq.Parsable.make (Gramlib.Stream.empty ());
     start = 0 }
 
 let set_prompt prompt =
@@ -242,10 +242,10 @@ let set_prompt prompt =
 
 (* Read the input stream until a dot is encountered *)
 let parse_to_dot =
-  let rec dot st = match LStream.next st with
+  let rec dot kwstate st = match Gramlib.LStream.next kwstate st with
     | Tok.KEYWORD ("."|"...") -> ()
     | Tok.EOI -> ()
-    | _ -> dot st
+    | _ -> dot kwstate st
   in
   Pcoq.Entry.(of_parser "Coqtoplevel.dot" { parser_fun = dot })
 
@@ -356,7 +356,7 @@ let print_anyway_opts = [
 let print_anyway c =
   let open Vernacexpr in
   match c.expr with
-  | VernacSetOption (_, opt, _) -> List.mem opt print_anyway_opts
+  | VernacSynterp (VernacSetOption (_, opt, _)) -> List.mem opt print_anyway_opts
   | _ -> false
 
 (* print the proof step, possibly with diffs highlighted, *)
@@ -397,10 +397,12 @@ let top_goal_print ~doc c oldp newp =
     TopErr.print_error_for_buffer ?loc Feedback.Error msg top_buffer
   [@@ocaml.warning "-3"]
 
-let exit_on_error =
+let { Goptions.get = exit_on_error } =
   let open Goptions in
-  declare_bool_option_and_ref ~depr:false ~key:["Coqtop";"Exit";"On";"Error"]
+  declare_bool_option_and_ref
+    ~key:["Coqtop";"Exit";"On";"Error"]
     ~value:false
+    ()
 
 let show_proof_diff_cmd ~state diff_opt =
   let open Vernac.State in
@@ -438,16 +440,7 @@ let process_toplevel_command ~state stm =
   | VernacShowGoal { gid; sid } ->
     let proof = Stm.get_proof ~doc:state.doc (Stateid.of_int sid) in
     let goal = Printer.pr_goal_emacs ~proof gid sid in
-    let evars =
-      match proof with
-      | None -> mt()
-      | Some p ->
-        let gl = (Evar.unsafe_of_int gid) in
-        let { Proof.sigma } = Proof.data p in
-        try Printer.print_dependent_evars (Some gl) sigma [ gl ]
-        with Not_found -> mt()
-    in
-    Feedback.msg_notice (v 0 (goal ++ evars));
+    let () = Feedback.msg_notice goal in
     state
 
   | VernacShowProofDiffs diff_opt ->
