@@ -16,7 +16,8 @@ open Coqargs
 
 let () = at_exit flush_all
 
-let get_version () =
+let get_version ~boot =
+  if boot then Coq_config.version else
   try
     let env = Boot.Env.init () in
     let revision = Boot.Env.revision env |> Boot.Path.to_string in
@@ -27,8 +28,8 @@ let get_version () =
     Printf.sprintf "%s (%s)" ver rev
   with _ -> Coq_config.version
 
-let print_header () =
-  Feedback.msg_info (str "Welcome to Coq " ++ str (get_version ()));
+let print_header ~boot () =
+  Feedback.msg_info (str "Welcome to Coq " ++ str (get_version ~boot));
   flush_all ()
 
 
@@ -48,7 +49,7 @@ let fatal_error_exn exn =
 
 type ('a,'b) custom_toplevel =
   { parse_extra : string list -> 'a * string list
-  ; usage : Usage.specific_usage
+  ; usage : Boot.Usage.specific_usage
   ; init_extra : 'a -> Coqargs.injection_command list -> opts:Coqargs.t -> 'b
   ; initial_args : Coqargs.t
   ; run : 'a -> opts:Coqargs.t -> 'b -> unit
@@ -80,57 +81,6 @@ let start_coq custom =
 (** ****************************************)
 (** Specific support for coqtop executable *)
 
-type color = [`ON | `AUTO | `EMACS | `OFF]
-
-let init_color opts =
-  let has_color = match opts with
-  | `OFF -> false
-  | `EMACS -> false
-  | `ON -> true
-  | `AUTO ->
-    Terminal.has_style Unix.stdout &&
-    Terminal.has_style Unix.stderr &&
-    (* emacs compilation buffer does not support colors by default,
-       its TERM variable is set to "dumb". *)
-    try Sys.getenv "TERM" <> "dumb" with Not_found -> false
-  in
-  let term_color =
-    if has_color then begin
-      let colors = try Some (Sys.getenv "COQ_COLORS") with Not_found -> None in
-      match colors with
-      | None -> Topfmt.default_styles (); true        (* Default colors *)
-      | Some "" -> false                              (* No color output *)
-      | Some s -> Topfmt.parse_color_config s; true   (* Overwrite all colors *)
-    end
-    else begin
-      Topfmt.default_styles (); false                 (* textual markers, no color *)
-    end
-  in
-  if opts = `EMACS then
-    Topfmt.set_emacs_print_strings ()
-  else if not term_color then begin
-      Proof_diffs.write_color_enabled term_color;
-    if Proof_diffs.show_diffs () then
-      (prerr_endline "Error: -diffs requires enabling -color"; exit 1)
-  end;
-  Topfmt.init_terminal_output ~color:term_color
-
-let print_style_tags opts =
-  let () = init_color opts in
-  let tags = Topfmt.dump_tags () in
-  let iter (t, st) =
-    let opt = Terminal.eval st ^ t ^ Terminal.reset ^ "\n" in
-    print_string opt
-  in
-  let make (t, st) =
-    let tags = List.map string_of_int (Terminal.repr st) in
-    (t ^ "=" ^ String.concat ";" tags)
-  in
-  let repr = List.map make tags in
-  let () = Printf.printf "COQ_COLORS=\"%s\"\n" (String.concat ":" repr) in
-  let () = List.iter iter tags in
-  flush_all ()
-
 let ltac_debug_answer = let open DebugHook.Answer in function
     | Prompt prompt ->
       (* No newline *)
@@ -159,7 +109,7 @@ type run_mode = Interactive | Batch | Query of query
 
 type toplevel_options = {
   run_mode : run_mode;
-  color_mode : color;
+  color_mode : Colors.color;
 }
 
 let init_document opts stm_options injections =
@@ -175,19 +125,19 @@ let init_document opts stm_options injections =
   let doc, sid =
     Stm.(new_doc
            { doc_type = Interactive opts.config.logic.toplevel_name;
-             injections; stm_options;
+             injections;
            }) in
-  { doc; sid; proof = None; time = opts.config.time }
+  { doc; sid; proof = None; time = Option.map Vernac.make_time_output opts.config.time }
 
 let init_toploop opts stm_opts injections =
   let state = init_document opts stm_opts injections in
-  let state = Ccompile.load_init_vernaculars opts ~state in
+  let state = Load.load_init_vernaculars opts ~state in
   state
 
 let coqtop_init ({ run_mode; color_mode }, async_opts) injections ~opts =
   if run_mode != Interactive then Flags.quiet := true;
-  init_color (if opts.config.print_emacs then `EMACS else color_mode);
-  Flags.if_verbose print_header ();
+  Colors.init_color (if opts.config.print_emacs then `EMACS else color_mode);
+  Flags.if_verbose (print_header ~boot:opts.pre.boot) ();
   DebugHook.Intf.(set
     { read_cmd = ltac_debug_parse
     ; submit_answer = ltac_debug_answer
@@ -195,22 +145,6 @@ let coqtop_init ({ run_mode; color_mode }, async_opts) injections ~opts =
     });
   init_toploop opts async_opts injections
 
-let set_color = function
-  | "yes" | "on" -> `ON
-  | "no" | "off" -> `OFF
-  | "auto" ->`AUTO
-  | _ ->
-    error_wrong_arg ("Error: on/off/auto expected after option color")
-
-let parse_extra_colors extras =
-  let rec parse_extra color_mode = function
-  | "-color" :: next :: rest -> parse_extra (set_color next) rest
-  | "-list-tags" :: rest -> parse_extra color_mode rest
-  | x :: rest ->
-    let color_mode, rest = parse_extra color_mode rest in color_mode, x :: rest
-  | [] -> color_mode, [] in
-  parse_extra `AUTO extras
-
 let coqtop_parse_extra extras =
   let rec parse_extra run_mode  = function
   | "-batch" :: rest -> parse_extra Batch  rest
@@ -219,13 +153,18 @@ let coqtop_parse_extra extras =
     let run_mode, rest = parse_extra run_mode rest in run_mode, x :: rest
   | [] -> run_mode, [] in
   let run_mode, extras = parse_extra Interactive extras in
-  let color_mode, extras = parse_extra_colors extras in
+  let color_mode, extras = Colors.parse_extra_colors extras in
   let async_opts, extras = Stmargs.parse_args ~init:Stm.AsyncOpts.default_opts extras in
   ({ run_mode; color_mode}, async_opts), extras
 
+let fix_windows_dirsep s =
+  if Sys.win32 then Str.(global_replace (regexp "\\(.\\)\\") "\\1/" s)
+  else s
+
 let get_native_name s =
   (* We ignore even critical errors because this mode has to be super silent *)
   try
+    fix_windows_dirsep @@
     Filename.(List.fold_left concat (dirname s)
                 [ !Nativelib.output_dir
                 ; Library.native_name_from_filename s
@@ -235,14 +174,14 @@ let get_native_name s =
 let coqtop_run ({ run_mode; color_mode },_) ~opts state =
   match run_mode with
   | Interactive -> Coqloop.loop ~opts ~state;
-  | Query PrintTags -> print_style_tags color_mode; exit 0
+  | Query PrintTags -> Colors.print_style_tags color_mode; exit 0
   | Query (PrintModUid sl) ->
       let s = String.concat " " (List.map get_native_name sl) in
       print_endline s;
       exit 0
   | Batch -> exit 0
 
-let coqtop_specific_usage = Usage.{
+let coqtop_specific_usage = Boot.Usage.{
   executable_name = "coqtop";
   extra_args = "";
   extra_options = "\n\
