@@ -25,10 +25,10 @@ let project_hint ~poly pri l2r r =
   let t =
     Tacred.reduce_to_quantified_ref env sigma (lib_ref "core.iff.type") t
   in
-  let sign, ccl = decompose_prod_assum sigma t in
+  let sign, ccl = decompose_prod_decls sigma t in
   let a, b =
     match snd (decompose_app sigma ccl) with
-    | [a; b] -> (a, b)
+    | [|a; b|] -> (a, b)
     | _ -> assert false
   in
   let p = if l2r then lib_ref "core.iff.proj1" else lib_ref "core.iff.proj2" in
@@ -65,7 +65,7 @@ let project_hint ~poly pri l2r r =
   (info, true, Hints.PathAny, Hints.hint_globref (GlobRef.ConstRef c))
 
 let warn_deprecated_hint_constr =
-  CWarnings.create ~name:"fragile-hint-constr" ~category:"automation"
+  CWarnings.create ~name:"fragile-hint-constr" ~category:CWarnings.CoreCategories.automation
     (fun () ->
       Pp.strbrk
         "Declaring arbitrary terms as hints is fragile; it is recommended to \
@@ -157,11 +157,8 @@ let interp_hints ~poly h =
     HintsResolveEntry (List.flatten (List.map constr_hints_of_ind lqid))
   | HintsExtern (pri, patcom, tacexp) ->
     let pat = Option.map (fp sigma) patcom in
-    let l = match pat with None -> [] | Some (l, _) -> l in
-    let ltacvars =
-      List.fold_left (fun accu x -> Id.Set.add x accu) Id.Set.empty l
-    in
-    let env = Genintern.{(empty_glob_sign env) with ltacvars} in
+    let ltacvars = match pat with None -> Id.Set.empty | Some (l, _) -> l in
+    let env = Genintern.{(empty_glob_sign ~strict:true env) with ltacvars} in
     let _, tacexp = Genintern.generic_intern env tacexp in
     HintsExternEntry
       ({Typeclasses.hint_priority = Some pri; hint_pattern = pat}, tacexp)
