@@ -14,8 +14,10 @@ open Vernacexpr
 val do_fixpoint :
      pm:Declare.OblState.t
   -> scope:Locality.definition_scope
+  -> ?clearbody:bool
   -> poly:bool
   -> ?typing_flags:Declarations.typing_flags
+  -> ?deprecation:Deprecation.t
   -> ?using:Vernacexpr.section_subset_expr
   -> fixpoint_expr list
   -> Declare.OblState.t
@@ -24,6 +26,7 @@ val do_cofixpoint :
      pm:Declare.OblState.t
   -> scope:Locality.definition_scope
   -> poly:bool
+  -> ?deprecation:Deprecation.t
   -> ?using:Vernacexpr.section_subset_expr
   -> cofixpoint_expr list
   -> Declare.OblState.t
