@@ -34,7 +34,7 @@ module Delayed :
 sig
 
 type 'a delayed
-val in_delayed : string -> ObjFile.in_handle -> segment:string -> 'a delayed * Digest.t
+val in_delayed : string -> ObjFile.in_handle -> segment:'a ObjFile.id -> 'a delayed * Digest.t
 val fetch_delayed : 'a delayed -> 'a
 
 end =
@@ -82,6 +82,7 @@ type compilation_unit_name = DirPath.t
 
 type library_disk = {
   md_compiled : Safe_typing.compiled_library;
+  md_syntax_objects : Declaremods.library_objects;
   md_objects : Declaremods.library_objects;
 }
 
@@ -109,14 +110,16 @@ type library_summary = {
 
 (* This is a map from names to loaded libraries *)
 let libraries_table : library_summary DPmap.t ref =
-  Summary.ref DPmap.empty ~name:"LIBRARY"
+  Summary.ref DPmap.empty ~stage:Summary.Stage.Synterp ~name:"LIBRARY"
 
 (* This is the map of loaded libraries filename *)
 (* (not synchronized so as not to be caught in the states on disk) *)
 let libraries_filename_table = ref DPmap.empty
 
 (* These are the _ordered_ sets of loaded, imported and exported libraries *)
-let libraries_loaded_list = Summary.ref [] ~name:"LIBRARY-LOAD"
+let libraries_loaded_list = Summary.ref [] ~stage:Summary.Stage.Synterp ~name:"LIBRARY-LOAD"
+
+let loaded_native_libraries = Summary.ref DPset.empty ~stage:Summary.Stage.Interp ~name:"NATIVE-LIBRARY-LOAD"
 
 (* Opaque proof tables *)
 
@@ -141,12 +144,6 @@ let library_full_filename dir =
   try DPmap.find dir !libraries_filename_table
   with Not_found -> "<unavailable filename>"
 
-let overwrite_library_filenames f =
-  let f =
-    if Filename.is_relative f then Filename.concat (Sys.getcwd ()) f else f in
-  DPmap.iter (fun dir _ -> register_library_filename dir f)
-    !libraries_table
-
 let library_is_loaded dir =
   try let _ = find_library dir in true
   with Not_found -> false
@@ -157,17 +154,20 @@ let library_is_loaded dir =
 let register_loaded_library m =
   let libname = m.libsum_name in
   let rec aux = function
-    | [] ->
-        if Flags.get_native_compiler () then begin
-            let dirname = Filename.dirname (library_full_filename libname) in
-            Nativelib.enable_library dirname libname
-          end;
-        [libname]
+    | [] -> [libname]
     | m'::_ as l when DirPath.equal m' libname -> l
     | m'::l' -> m' :: aux l' in
   libraries_loaded_list := aux !libraries_loaded_list;
   libraries_table := DPmap.add libname m !libraries_table
 
+let register_native_library libname =
+  if (Global.typing_flags ()).enable_native_compiler
+    && not (DPset.mem libname !loaded_native_libraries) then begin
+      let dirname = Filename.dirname (library_full_filename libname) in
+      loaded_native_libraries := DPset.add libname !loaded_native_libraries;
+      Nativelib.enable_library dirname libname
+  end
+
   let loaded_libraries () = !libraries_loaded_list
 
 (** Delayed / available tables of opaque terms *)
@@ -244,64 +244,81 @@ let mk_summary m = {
   libsum_digests = m.library_digests;
 }
 
-let intern_from_file f =
+let mk_intern_library sum lib digest_lib univs digest_univs proofs =
+  add_opaque_table sum.md_name (ToFetch proofs);
+  let open Safe_typing in
+  match univs with
+  | None -> mk_library sum lib (Dvo_or_vi digest_lib) Univ.ContextSet.empty
+  | Some (uall,true) ->
+    mk_library sum lib (Dvivo (digest_lib, digest_univs)) uall
+  | Some (_,false) ->
+    mk_library sum lib (Dvo_or_vi digest_lib) Univ.ContextSet.empty
+
+type ('uid, 'doc) tasks = (('uid, 'doc) Stateid.request * bool) list
+
+let summary_seg : seg_sum ObjFile.id = ObjFile.make_id "summary"
+let library_seg : seg_lib ObjFile.id = ObjFile.make_id "library"
+let universes_seg : seg_univ option ObjFile.id = ObjFile.make_id "universes"
+let tasks_seg () : (Opaqueproof.opaque_handle option, 'doc) tasks option ObjFile.id = ObjFile.make_id "tasks"
+let opaques_seg : seg_proofs ObjFile.id = ObjFile.make_id "opaques"
+
+let intern_from_file lib_resolver dir =
+  let f = lib_resolver dir in
+  Feedback.feedback(Feedback.FileDependency (Some f, DirPath.to_string dir));
   let ch = raw_intern_library f in
-  let (lsd : seg_sum), digest_lsd = ObjFile.marshal_in_segment ch ~segment:"summary" in
-  let ((lmd : seg_lib), digest_lmd) = ObjFile.marshal_in_segment ch ~segment:"library" in
-  let (univs : seg_univ option), digest_u = ObjFile.marshal_in_segment ch ~segment:"universes" in
-  let ((del_opaque : seg_proofs delayed),_) = in_delayed f ch ~segment:"opaques" in
+  let lsd, digest_lsd = ObjFile.marshal_in_segment ch ~segment:summary_seg in
+  let lmd, digest_lmd = ObjFile.marshal_in_segment ch ~segment:library_seg in
+  let univs, digest_u = ObjFile.marshal_in_segment ch ~segment:universes_seg in
+  let del_opaque, _ = in_delayed f ch ~segment:opaques_seg in
   ObjFile.close_in ch;
   System.check_caml_version ~caml:lsd.md_ocaml ~file:f;
   register_library_filename lsd.md_name f;
-  add_opaque_table lsd.md_name (ToFetch del_opaque);
-  let open Safe_typing in
-  match univs with
-  | None -> mk_library lsd lmd (Dvo_or_vi digest_lmd) Univ.ContextSet.empty
-  | Some (uall,true) ->
-      mk_library lsd lmd (Dvivo (digest_lmd,digest_u)) uall
-  | Some (_,false) ->
-      mk_library lsd lmd (Dvo_or_vi digest_lmd) Univ.ContextSet.empty
-
-let rec intern_library ~lib_resolver (needed, contents) (dir, f) from =
-  (* Look if in the current logical environment *)
-  try (find_library dir).libsum_digests, (needed, contents)
-  with Not_found ->
-  (* Look if already listed and consequently its dependencies too *)
-  try (DPmap.find dir contents).library_digests, (needed, contents)
-  with Not_found ->
-  Feedback.feedback(Feedback.FileDependency (from, DirPath.to_string dir));
   (* [dir] is an absolute name which matches [f] which must be in loadpath *)
-  let f = match f with Some f -> f | None -> lib_resolver dir in
-  let m = intern_from_file f in
-  if not (DirPath.equal dir m.library_name) then
+  if not (DirPath.equal dir lsd.md_name) then
     user_err
       (str "The file " ++ str f ++ str " contains library" ++ spc () ++
-       DirPath.print m.library_name ++ spc () ++ str "and not library" ++
+       DirPath.print lsd.md_name ++ spc () ++ str "and not library" ++
        spc() ++ DirPath.print dir ++ str ".");
   Feedback.feedback (Feedback.FileLoaded(DirPath.to_string dir, f));
-  m.library_digests, intern_library_deps ~lib_resolver (needed, contents) dir m f
+  lsd, lmd, digest_lmd, univs, digest_u, del_opaque
 
-and intern_library_deps ~lib_resolver libs dir m from =
+let rec intern_library ~intern (needed, contents as acc) dir =
+  (* Look if in the current logical environment *)
+  try (find_library dir).libsum_digests, acc
+  with Not_found ->
+  (* Look if already listed and consequently its dependencies too *)
+  try (DPmap.find dir contents).library_digests, acc
+  with Not_found ->
+  let lsd, lmd, digest_lmd, univs, digest_u, del_opaque = intern dir in
+  let m = mk_intern_library lsd lmd digest_lmd univs digest_u del_opaque in
+  m.library_digests, intern_library_deps ~intern acc dir m
+
+and intern_library_deps ~intern libs dir m =
   let needed, contents =
-    Array.fold_left (intern_mandatory_library ~lib_resolver dir from)
+    Array.fold_left (intern_mandatory_library ~intern dir)
       libs m.library_deps in
   (dir :: needed, DPmap.add dir m contents )
 
-and intern_mandatory_library ~lib_resolver caller from libs (dir,d) =
-  let digest, libs = intern_library ~lib_resolver libs (dir, None) (Some from) in
-  if not (Safe_typing.digest_match ~actual:digest ~required:d) then
-    user_err (str "Compiled library " ++ DirPath.print caller ++
-    str " (in file " ++ str from ++ str ") makes inconsistent assumptions \
-    over library " ++ DirPath.print dir);
+and intern_mandatory_library ~intern caller libs (dir,d) =
+  let digest, libs = intern_library ~intern libs dir in
+  let () = if not (Safe_typing.digest_match ~actual:digest ~required:d) then
+    let from = library_full_filename caller in
+    user_err
+      (str "Compiled library " ++ DirPath.print caller ++
+       str " (in file " ++ str from ++
+       str ") makes inconsistent assumptions over library " ++
+       DirPath.print dir)
+  in
   libs
 
-let rec_intern_library ~lib_resolver libs (dir, f) =
-  let _, libs = intern_library ~lib_resolver libs (dir, Some f) None in
+let rec_intern_library ~lib_resolver libs dir =
+  let intern dir = intern_from_file lib_resolver dir in
+  let _, libs = intern_library ~intern libs dir in
   libs
 
 let native_name_from_filename f =
   let ch = raw_intern_library f in
-  let (lmd : seg_sum), digest_lmd = ObjFile.marshal_in_segment ch ~segment:"summary" in
+  let lmd, digest_lmd = ObjFile.marshal_in_segment ch ~segment:summary_seg in
   Nativecode.mod_uid_of_dirpath lmd.md_name
 
 (**********************************************************************)
@@ -323,12 +340,19 @@ let native_name_from_filename f =
 
 let register_library m =
   let l = m.library_data in
-  Declaremods.register_library
+  Declaremods.Interp.register_library
     m.library_name
     l.md_compiled
     l.md_objects
     m.library_digests
     m.library_extra_univs;
+  register_native_library m.library_name
+
+let register_library_syntax m =
+  let l = m.library_data in
+  Declaremods.Synterp.register_library
+    m.library_name
+    l.md_syntax_objects;
   register_loaded_library (mk_summary m)
 
 (* Follow the semantics of Anticipate object:
@@ -357,32 +381,56 @@ let in_require : require_obj -> obj =
      discharge_function = discharge_require;
      classify_function = (fun o -> Anticipate) }
 
+let load_require_syntax _ needed =
+  List.iter register_library_syntax needed
+
+let cache_require_syntax o =
+  load_require_syntax 1 o
+
+let discharge_require_syntax o = Some o
+
+(* open_function is never called from here because an Anticipate object *)
+
+type require_obj_syntax = library_t list
+
+let in_require_syntax : require_obj_syntax -> obj =
+  declare_object
+    {(default_object "REQUIRE-SYNTAX") with
+     object_stage = Summary.Stage.Synterp;
+     cache_function = cache_require_syntax;
+     load_function = load_require_syntax;
+     open_function = (fun _ _ -> assert false);
+     discharge_function = discharge_require_syntax;
+     classify_function = (fun o -> Anticipate) }
+
 (* Require libraries, import them if [export <> None], mark them for export
    if [export = Some true] *)
 
 let warn_require_in_module =
-  CWarnings.create ~name:"require-in-module" ~category:"fragile"
+  CWarnings.create ~name:"require-in-module" ~category:CWarnings.CoreCategories.fragile
     (fun () -> strbrk "Use of “Require” inside a module is fragile." ++ spc() ++
                strbrk "It is not recommended to use this functionality in finished proof scripts.")
 
-let require_library_from_dirpath ~lib_resolver modrefl =
-  let needed, contents = List.fold_left (rec_intern_library ~lib_resolver) ([], DPmap.empty) modrefl in
-  let needed = List.rev_map (fun dir -> DPmap.find dir contents) needed in
+let require_library_from_dirpath needed =
   if Lib.is_module_or_modtype () then warn_require_in_module ();
   Lib.add_leaf (in_require needed)
 
+let require_library_syntax_from_dirpath ~lib_resolver modrefl =
+  let needed, contents = List.fold_left (rec_intern_library ~lib_resolver) ([], DPmap.empty) modrefl in
+  let needed = List.rev_map (fun dir -> DPmap.find dir contents) needed in
+  Lib.add_leaf (in_require_syntax needed);
+  needed
+
 (************************************************************************)
 (*s Initializing the compilation of a library. *)
 
-type ('uid, 'doc) tasks = (('uid, 'doc) Stateid.request * bool) list
-
 let load_library_todo f =
   let ch = raw_intern_library f in
-  let (s0 : seg_sum), _ = ObjFile.marshal_in_segment ch ~segment:"summary" in
-  let (s1 : seg_lib), _ = ObjFile.marshal_in_segment ch ~segment:"library" in
-  let (s2 : seg_univ option), _ = ObjFile.marshal_in_segment ch ~segment:"universes" in
-  let (tasks : (Opaqueproof.opaque_handle option, 'doc) tasks option), _ = ObjFile.marshal_in_segment ch ~segment:"tasks" in
-  let (s4 : seg_proofs), _ = ObjFile.marshal_in_segment ch ~segment:"opaques" in
+  let s0, _ = ObjFile.marshal_in_segment ch ~segment:summary_seg in
+  let s1, _ = ObjFile.marshal_in_segment ch ~segment:library_seg in
+  let s2, _ = ObjFile.marshal_in_segment ch ~segment:universes_seg in
+  let tasks, _ = ObjFile.marshal_in_segment ch ~segment:(tasks_seg ()) in
+  let s4, _ = ObjFile.marshal_in_segment ch ~segment:opaques_seg in
   ObjFile.close_in ch;
   System.check_caml_version ~caml:s0.md_ocaml ~file:f;
   if tasks = None then user_err (str "Not a .vio file.");
@@ -425,11 +473,11 @@ type 'doc todo_proofs =
 let save_library_base f sum lib univs tasks proofs =
   let ch = raw_extern_library f in
   try
-    ObjFile.marshal_out_segment ch ~segment:"summary" (sum    : seg_sum);
-    ObjFile.marshal_out_segment ch ~segment:"library" (lib    : seg_lib);
-    ObjFile.marshal_out_segment ch ~segment:"universes" (univs  : seg_univ option);
-    ObjFile.marshal_out_segment ch ~segment:"tasks" (tasks : (Opaqueproof.opaque_handle option, 'doc) tasks option);
-    ObjFile.marshal_out_segment ch ~segment:"opaques" (proofs : seg_proofs);
+    ObjFile.marshal_out_segment ch ~segment:summary_seg sum;
+    ObjFile.marshal_out_segment ch ~segment:library_seg lib;
+    ObjFile.marshal_out_segment ch ~segment:universes_seg univs;
+    ObjFile.marshal_out_segment ch ~segment:(tasks_seg ()) tasks;
+    ObjFile.marshal_out_segment ch ~segment:opaques_seg proofs;
     ObjFile.close_out ch
   with reraise ->
     let reraise = Exninfo.capture reraise in
@@ -438,6 +486,24 @@ let save_library_base f sum lib univs tasks proofs =
     Sys.remove f;
     Exninfo.iraise reraise
 
+(* This is the basic vo save structure *)
+let save_library_struct ~output_native_objects dir =
+  let md_compiled, md_objects, md_syntax_objects, ast =
+    Declaremods.end_library ~output_native_objects dir in
+  let sd =
+    { md_name = dir
+    ; md_deps = Array.of_list (current_deps ())
+    ; md_ocaml = Coq_config.caml_version
+    } in
+  let md =
+    { md_compiled
+    ; md_syntax_objects
+    ; md_objects
+    } in
+  if Array.exists (fun (d,_) -> DirPath.equal d dir) sd.md_deps then
+    error_recursively_dependent_library dir;
+  sd, md, ast
+
 let save_library_to todo_proofs ~output_native_objects dir f =
   assert(
     let expected_extension = match todo_proofs with
@@ -457,7 +523,6 @@ let save_library_to todo_proofs ~output_native_objects dir f =
   let () = assert (not (Future.UUIDSet.is_empty except) ||
     Safe_typing.is_joined_environment (Global.safe_env ()))
   in
-  let cenv, seg, ast = Declaremods.end_library ~output_native_objects dir in
   let tasks, utab =
     match todo_proofs with
     | ProofsTodoNone -> None, None
@@ -471,18 +536,8 @@ let save_library_to todo_proofs ~output_native_objects dir f =
           tasks in
       Some tasks,
       Some (Univ.ContextSet.empty,false)
-    in
-    let sd = {
-    md_name = dir;
-    md_deps = Array.of_list (current_deps ());
-    md_ocaml = Coq_config.caml_version;
-  } in
-  let md = {
-    md_compiled = cenv;
-    md_objects = seg;
-  } in
-  if Array.exists (fun (d,_) -> DirPath.equal d dir) sd.md_deps then
-    error_recursively_dependent_library dir;
+  in
+  let sd, md, ast = save_library_struct ~output_native_objects dir in
   (* Writing vo payload *)
   save_library_base f sd md utab tasks opaque_table;
   (* Writing native code files *)
