@@ -16,11 +16,11 @@ module DP = Names.DirPath
 type t = {
   path_physical : CUnix.physical_path;
   path_logical : DP.t;
-  path_implicit : bool;
+  path_implicit : bool;  (* true for -R, false for -Q in command line *)
   path_root : (CUnix.physical_path * DP.t);
 }
 
-let load_paths = Summary.ref ([] : t list) ~name:"LOADPATHS"
+let load_paths = Summary.ref ([] : t list) ~stage:Summary.Stage.Synterp ~name:"LOADPATHS"
 
 let logical p = p.path_logical
 let physical p = p.path_physical
@@ -49,7 +49,7 @@ let find_with_logical_path dirpath =
   List.filter (fun p -> Names.DirPath.equal p.path_logical dirpath) !load_paths
 
 let warn_file_found_multiple_times =
-  CWarnings.create ~name:"ambiguous-extra-dep" ~category:"filesystem"
+  CWarnings.create ~name:"ambiguous-extra-dep" ~category:CWarnings.CoreCategories.filesystem
     (fun (file,from,other,extra) ->
       Pp.(str "File " ++ str file ++ str " found twice in " ++
       Names.DirPath.print from ++ str":" ++ spc () ++ str other ++ str " (selected)," ++
@@ -88,7 +88,7 @@ let remove_load_path dir =
   load_paths := List.filter filter !load_paths
 
 let warn_overriding_logical_loadpath =
-  CWarnings.create ~name:"overriding-logical-loadpath" ~category:"loadpath"
+  CWarnings.create ~name:"overriding-logical-loadpath" ~category:CWarnings.CoreCategories.filesystem
     (fun (phys_path, old_path, coq_path) ->
        Pp.(seq [str phys_path; strbrk " was previously bound to "
                ; DP.print old_path; strbrk "; it is remapped to "
@@ -176,12 +176,11 @@ let locate_file fname =
 (************************************************************************)
 (*s Locate absolute or partially qualified library names in the path *)
 
-type library_location = LibLoaded | LibInPath
 type locate_error = LibUnmappedDir | LibNotFound
 type 'a locate_result = ('a, locate_error) result
 
 let warn_several_object_files =
-  CWarnings.create ~name:"several-object-files" ~category:"require"
+  CWarnings.create ~name:"several-object-files" ~category:CWarnings.CoreCategories.filesystem
     Pp.(fun (vi, vo) ->
         seq [ str "Loading"; spc (); str vi
             ; strbrk " instead of "; str vo
@@ -246,7 +245,7 @@ let locate_absolute_library dir : CUnix.physical_path locate_result =
     | Error fail -> Error fail
 
 let locate_qualified_library ?root qid :
-  (library_location * DP.t * CUnix.physical_path) locate_result =
+  (DP.t * CUnix.physical_path) locate_result =
   (* Search library in loadpath *)
   let dir, base = Libnames.repr_qualid qid in
   match expand_path ?root dir with
@@ -270,11 +269,7 @@ let locate_qualified_library ?root qid :
     match result with
     | Ok (dir,file) ->
       let library = Libnames.add_dirpath_suffix dir base in
-      (* Look if loaded *)
-      if Library.library_is_loaded library
-      then Ok (LibLoaded, library, Library.library_full_filename library)
-      (* Otherwise, look for it in the file system *)
-      else Ok (LibInPath, library, file)
+      Ok (library, file)
     | Error _ as e -> e
 
 let error_unmapped_dir qid =
@@ -307,7 +302,8 @@ type vo_path =
   ; coq_path  : DP.t
   (** Coq prefix for the path *)
   ; implicit  : bool
-  (** [implicit = true] avoids having to qualify with [coq_path] *)
+  (** [implicit = true] avoids having to qualify with [coq_path]
+      true for -R, false for -Q in command line *)
   ; has_ml    : bool
   (** If [has_ml] is true, the directory will also be added to the ml include path *)
   ; recursive : bool
@@ -315,11 +311,11 @@ type vo_path =
   }
 
 let warn_cannot_open_path =
-  CWarnings.create ~name:"cannot-open-path" ~category:"filesystem"
+  CWarnings.create ~name:"cannot-open-path" ~category:CWarnings.CoreCategories.filesystem
     (fun unix_path -> Pp.(str "Cannot open " ++ str unix_path))
 
 let warn_cannot_use_directory =
-  CWarnings.create ~name:"cannot-use-directory" ~category:"filesystem"
+  CWarnings.create ~name:"cannot-use-directory" ~category:CWarnings.CoreCategories.filesystem
     (fun d ->
        Pp.(str "Directory " ++ str d ++
            strbrk " cannot be used as a Coq identifier (skipped)"))
@@ -338,6 +334,7 @@ let add_vo_path lp =
   let recursive = lp.recursive in
   if System.exists_dir unix_path then
     let dirs = if recursive then System.all_subdirs ~unix_path else [] in
+    let dirs = List.sort (fun a b -> String.compare (fst a) (fst b)) dirs in
     let prefix = DP.repr lp.coq_path in
     let convert_dirs (lp, cp) =
       try
