@@ -25,28 +25,13 @@ open Libobject
 open Libnames
 open Globnames
 open Printer
-open Printmod
 open Context.Rel.Declaration
 
 module RelDecl = Context.Rel.Declaration
 module NamedDecl = Context.Named.Declaration
 
-type object_pr = {
-  print_inductive           : MutInd.t -> UnivNames.univ_name_list option -> Pp.t;
-  print_constant_with_infos : Constant.t -> UnivNames.univ_name_list option -> Pp.t;
-  print_section_variable    : env -> Evd.evar_map -> variable -> Pp.t;
-  print_syntactic_def       : env -> KerName.t -> Pp.t;
-  print_module              : bool -> ModPath.t -> Pp.t;
-  print_modtype             : ModPath.t -> Pp.t;
-  print_named_decl          : env -> Evd.evar_map -> Constr.named_declaration -> Pp.t;
-  print_library_entry       : env -> Evd.evar_map -> bool -> (object_name * Lib.node) -> Pp.t option;
-  print_context             : env -> Evd.evar_map -> bool -> int option -> Lib.library_segment -> Pp.t;
-  print_typed_value_in_env  : Environ.env -> Evd.evar_map -> EConstr.constr * EConstr.types -> Pp.t;
-  print_eval                : Reductionops.reduction_function -> env -> Evd.evar_map -> Constrexpr.constr_expr -> EConstr.unsafe_judgment -> Pp.t;
-}
-
-let gallina_print_module  = print_module
-let gallina_print_modtype = print_modtype
+let print_module mp = Printmod.print_module ~with_body:true mp
+let print_modtype = Printmod.print_modtype
 
 (**************)
 (** Utilities *)
@@ -73,10 +58,10 @@ let print_ref reduce ref udecl =
   let typ, univs = Typeops.type_of_global_in_context env ref in
   let inst = Univ.make_abstract_instance univs in
   let bl = Printer.universe_binders_with_opt_names (Environ.universes_of_global env ref) udecl in
-  let sigma = Evd.from_ctx (UState.of_binders bl) in
+  let sigma = Evd.from_ctx (UState.of_names bl) in
   let typ =
     if reduce then
-      let ctx,ccl = Reductionops.splay_prod_assum env sigma (EConstr.of_constr typ)
+      let ctx,ccl = Reductionops.hnf_decompose_prod_decls env sigma (EConstr.of_constr typ)
       in EConstr.to_constr sigma (EConstr.it_mkProd_or_LetIn ccl ctx)
     else typ in
   let typ = Arguments_renaming.rename_type typ ref in
@@ -140,14 +125,13 @@ let print_impargs_list prefix l =
 
 let need_expansion impl ref =
   let typ, _ = Typeops.type_of_global_in_context (Global.env ()) ref in
-  let ctx = Term.prod_assum typ in
+  let ctx = Term.prod_decls typ in
   let nprods = List.count is_local_assum ctx in
   not (List.is_empty impl) && List.length impl >= nprods &&
     let _,lastimpl = List.chop nprods impl in
       List.exists is_status_implicit lastimpl
 
 let print_impargs ref =
-  let ref = Smartlocate.smart_global ref in
   let impl = implicits_of_global ref in
   let has_impl = not (List.is_empty impl) in
   (* Need to reduce since implicits are computed with products flattened *)
@@ -198,14 +182,18 @@ let print_opacity ref =
 (*******************)
 
 let print_if_is_coercion ref =
-  if Coercionops.coercion_exists ref then [pr_global ref ++ str " is a coercion"] else []
+  if Coercionops.coercion_exists ref then
+    let i = Coercionops.coercion_info ref in
+    let r = if i.Coercionops.coe_reversible then " reversible" else "" in
+    [pr_global ref ++ str " is a" ++ str r ++ str " coercion"]
+  else []
 
 (*******************)
 (* *)
 
 let pr_template_variables = function
   | [] -> mt ()
-  | vars -> str "on " ++ prlist_with_sep spc UnivNames.(pr_with_global_universes empty_binders) vars
+  | vars -> str " on " ++ prlist_with_sep spc UnivNames.pr_with_global_universes vars
 
 let print_polymorphism ref =
   let poly = Global.is_polymorphic ref in
@@ -214,8 +202,8 @@ let print_polymorphism ref =
   [ pr_global ref ++ str " is " ++
       (if poly then str "universe polymorphic"
        else if template_poly then
-         str "template universe polymorphic "
-         ++ h (pr_template_variables template_variables)
+         str "template universe polymorphic"
+         ++ if !Detyping.print_universes then h (pr_template_variables template_variables) else mt()
        else str "not universe polymorphic") ]
 
 let print_type_in_type ref =
@@ -252,13 +240,9 @@ let needs_extra_scopes ref scopes =
   let ty, _ctx = Typeops.type_of_global_in_context env ref in
   aux env ty scopes
 
-let implicit_name_of_pos = function
-  | Constrexpr.ExplByName id -> Name id
-  | Constrexpr.ExplByPos (n,k) -> Anonymous
-
 let implicit_kind_of_status = function
   | None -> Anonymous, Glob_term.Explicit
-  | Some (pos,_,(maximal,_)) -> implicit_name_of_pos pos, if maximal then Glob_term.MaxImplicit else Glob_term.NonMaxImplicit
+  | Some ((na,_,_),_,(maximal,_)) -> na, if maximal then Glob_term.MaxImplicit else Glob_term.NonMaxImplicit
 
 let extra_implicit_kind_of_status imp =
   let _,imp = implicit_kind_of_status imp in
@@ -266,14 +250,14 @@ let extra_implicit_kind_of_status imp =
 
 let dummy = {
   Vernacexpr.implicit_status = Glob_term.Explicit;
-  name = Anonymous;
-  recarg_like = false;
-  notation_scope = None;
-}
+   name = Anonymous;
+   recarg_like = false;
+   notation_scope = [];
+ }
 
 let is_dummy = function
   | Vernacexpr.(RealArg {implicit_status; name; recarg_like; notation_scope}) ->
-    name = Anonymous && not recarg_like && notation_scope = None && implicit_status = Glob_term.Explicit
+    name = Anonymous && not recarg_like && notation_scope = [] && implicit_status = Glob_term.Explicit
   | _ -> false
 
 let rec main_implicits i renames recargs scopes impls =
@@ -290,8 +274,8 @@ let rec main_implicits i renames recargs scopes impls =
       | [], (None::_ | []) -> (Anonymous, Glob_term.Explicit)
     in
     let notation_scope = match scopes with
-      | scope :: _ -> Option.map CAst.make scope
-      | [] -> None
+      | scope :: _ -> List.map CAst.make scope
+      | [] -> []
     in
     let status = {Vernacexpr.implicit_status; name; recarg_like; notation_scope} in
     let tl = function [] -> [] | _::tl -> tl in
@@ -323,7 +307,12 @@ let print_arguments ref =
     | Some (UnfoldWhen { nargs; recargs }) -> [], recargs, nargs
     | Some (UnfoldWhenNoMatch { nargs; recargs }) -> [`ReductionDontExposeCase], recargs, nargs
   in
-  let renames = try Arguments_renaming.arguments_names ref with Not_found -> [] in
+  let names, not_renamed =
+    try Arguments_renaming.arguments_names ref, false
+    with Not_found ->
+      let env = Global.env () in
+      let ty, _ = Typeops.type_of_global_in_context env ref in
+      List.map pi1 (Impargs.compute_implicits_names env (Evd.from_env env) (EConstr.of_constr ty)), true in
   let scopes = Notation.find_arguments_scope ref in
   let flags = if needs_extra_scopes ref scopes then `ExtraScopes::flags else flags in
   let impls = Impargs.extract_impargs_data (Impargs.implicits_of_global ref) in
@@ -331,7 +320,7 @@ let print_arguments ref =
     | (_, impls) :: rest -> impls, rest
     | [] -> assert false
   in
-  let impls = main_implicits 0 renames recargs scopes impls in
+  let impls = main_implicits 0 names recargs scopes impls in
   let moreimpls = List.map (fun (_,i) -> List.map extra_implicit_kind_of_status i) moreimpls in
   let bidi = Pretyping.get_bidirectionality_hint ref in
   let impls = insert_fake_args nargs_for_red bidi impls in
@@ -340,10 +329,27 @@ let print_arguments ref =
     let open Constrexpr in
     let open Vernacexpr in
     [Ppvernac.pr_vernac_expr
-       (VernacArguments (CAst.make (AN qid), impls, moreimpls, flags)) ++
-     (if renames = [] then mt () else
+       (VernacSynPure (VernacArguments (CAst.make (AN qid), impls, moreimpls, flags))) ++
+     (if not_renamed then mt () else
       fnl () ++ str "  (where some original arguments have been renamed)")]
 
+let print_section_deps ref =
+  let hyps = let open GlobRef in match ref with
+  | VarRef _ -> None
+  | ConstRef c ->
+    let bd = Global.lookup_constant c in
+    Some bd.const_hyps
+  | IndRef (mind,_) | ConstructRef ((mind,_),_) ->
+    let mb = Global.lookup_mind mind in
+    Some mb.mind_hyps
+  in
+  let hyps = Option.map (List.filter NamedDecl.is_local_assum) hyps in
+  match hyps with
+  | None | Some [] -> []
+  | Some hyps ->
+    [hov 0 (pr_global ref ++ str (String.plural (List.length hyps) " uses section variable") ++ spc () ++
+            hv 1 (prlist_with_sep spc (fun d -> Id.print (NamedDecl.get_id d)) (List.rev hyps)) ++ str ".")]
+
 let print_name_infos ref =
   let type_info_for_implicit =
     if need_expansion (select_impargs_size 0 (implicits_of_global ref)) ref then
@@ -356,6 +362,7 @@ let print_name_infos ref =
   print_primitive ref @
   type_info_for_implicit @
   print_arguments ref @
+  print_section_deps ref @
   print_if_is_coercion ref
 
 let print_inductive_args sp mipv =
@@ -388,7 +395,8 @@ type locatable = Locatable : 'a locatable_info -> locatable
 type logical_name =
   | Term of GlobRef.t
   | Dir of Nametab.GlobDirRef.t
-  | Syntactic of KerName.t
+  | Abbreviation of abbreviation
+  | Module of ModPath.t
   | ModuleType of ModPath.t
   | Other : 'a * 'a locatable_info -> logical_name
   | Undefined of qualid
@@ -404,10 +412,12 @@ exception ObjFound of logical_name
 let locate_any_name qid =
   try Term (Nametab.locate qid)
   with Not_found ->
-  try Syntactic (Nametab.locate_syndef qid)
+  try Abbreviation (Nametab.locate_abbreviation qid)
   with Not_found ->
   try Dir (Nametab.locate_dir qid)
   with Not_found ->
+  try Module (Nametab.locate_module qid)
+  with Not_found ->
   try ModuleType (Nametab.locate_modtype qid)
   with Not_found ->
     let iter _ (Locatable info) = match info.locate qid with
@@ -417,26 +427,36 @@ let locate_any_name qid =
     try String.Map.iter iter !locatable_map; Undefined qid
     with ObjFound obj -> obj
 
+let canonical_info ref =
+  let cref = canonical_gr ref in
+  if GlobRef.UserOrd.equal ref cref then mt ()
+  else match Nametab.path_of_global cref with
+    | path -> spc() ++ str "(syntactically equal to" ++ spc() ++ pr_path path ++ str ")"
+    | exception Not_found -> spc() ++ str "(missing canonical, bug?)"
+
 let pr_located_qualid = function
   | Term ref ->
       let ref_str = let open GlobRef in match ref with
           ConstRef _ -> "Constant"
         | IndRef _ -> "Inductive"
         | ConstructRef _ -> "Constructor"
-        | VarRef _ -> "Variable" in
-      str ref_str ++ spc () ++ pr_path (Nametab.path_of_global ref)
-  | Syntactic kn ->
-      str "Notation" ++ spc () ++ pr_path (Nametab.path_of_syndef kn)
+        | VarRef _ -> "Variable"
+      in
+      let extra = canonical_info ref in
+      str ref_str ++ spc () ++ pr_path (Nametab.path_of_global ref) ++ extra
+  | Abbreviation kn ->
+      str "Notation" ++ spc () ++ pr_path (Nametab.path_of_abbreviation kn)
   | Dir dir ->
-      let s,dir =
+      let s,mp =
         let open Nametab in
         let open GlobDirRef in match dir with
-        | DirOpenModule { obj_dir ; _ } -> "Open Module", obj_dir
-        | DirOpenModtype { obj_dir ; _ } -> "Open Module Type", obj_dir
-        | DirOpenSection { obj_dir ; _ } -> "Open Section", obj_dir
-        | DirModule { obj_dir ; _ } -> "Module", obj_dir
+        | DirOpenModule mp -> "Open Module", mp
+        | DirOpenModtype mp -> "Open Module Type", mp
+        | DirOpenSection _ -> anomaly (Pp.str"Sections are not locatable")
       in
-      str s ++ spc () ++ DirPath.print dir
+      str s ++ spc () ++ str (ModPath.to_string mp)
+  | Module mp ->
+    str "Module" ++ spc () ++ DirPath.print (Nametab.dirpath_of_module mp)
   | ModuleType mp ->
       str "Module Type" ++ spc () ++ pr_path (Nametab.path_of_modtype mp)
   | Other (obj, info) -> info.name obj
@@ -472,19 +492,22 @@ let locate_term qid =
   let expand = function
     | TrueGlobal ref ->
         Term ref, Nametab.shortest_qualid_of_global Id.Set.empty ref
-    | SynDef kn ->
-        Syntactic kn, Nametab.shortest_qualid_of_syndef Id.Set.empty kn
+    | Abbrev kn ->
+        Abbreviation kn, Nametab.shortest_qualid_of_abbreviation Id.Set.empty kn
   in
   List.map expand (Nametab.locate_extended_all qid)
 
 let locate_module qid =
+  let all = Nametab.locate_extended_all_module qid in
+  let map mp = Module mp, Nametab.shortest_qualid_of_module mp in
+  let mods = List.map map all in
+  (* Don't forget the opened modules: they are not part of the same name tab. *)
   let all = Nametab.locate_extended_all_dir qid in
   let map dir = let open Nametab.GlobDirRef in match dir with
-  | DirModule { Nametab.obj_mp ; _ } -> Some (Dir dir, Nametab.shortest_qualid_of_module obj_mp)
   | DirOpenModule _ -> Some (Dir dir, qid)
   | _ -> None
   in
-  List.map_filter map all
+  mods @ List.map_filter map all
 
 let locate_modtype qid =
   let all = Nametab.locate_extended_all_modtype qid in
@@ -547,7 +570,7 @@ let print_located_qualid ref = print_located_qualid "object" LocAny ref
 (**** Printing declarations and judgments *)
 (****  Gallina layer                  *****)
 
-let gallina_print_typed_value_in_env env sigma (trm,typ) =
+let print_typed_value_in_env env sigma (trm,typ) =
   (pr_leconstr_env ~inctx:true env sigma trm ++ fnl () ++
      str "     : " ++ pr_letype_env env sigma typ)
 
@@ -568,7 +591,7 @@ let print_named_def env sigma name body typ =
 let print_named_assum env sigma name typ =
   str "*** [" ++ str name ++ str " : " ++ pr_ltype_env env sigma typ ++ str "]"
 
-let gallina_print_named_decl env sigma =
+let print_named_decl_gen env sigma =
   let open Context.Named.Declaration in
   function
   | LocalAssum (id, typ) ->
@@ -582,19 +605,19 @@ let assumptions_for_print lna =
 (*********************)
 (* *)
 
-let gallina_print_inductive sp udecl =
+let print_inductive sp udecl =
   let env = Global.env() in
   let mib = Environ.lookup_mind sp env in
   let mipv = mib.mind_packets in
-  pr_mutual_inductive_body env sp mib udecl ++
+  Printmod.pr_mutual_inductive_body env sp mib udecl ++
   with_line_skip
     (print_primitive_record mib.mind_finite mipv mib.mind_record @
      print_inductive_args sp mipv)
 
 let print_named_decl env sigma id =
-  gallina_print_named_decl env sigma (Global.lookup_named id) ++ fnl ()
+  print_named_decl_gen env sigma (Global.lookup_named id) ++ fnl ()
 
-let gallina_print_section_variable env sigma id =
+let print_section_variable env sigma id =
   print_named_decl env sigma id ++
   with_line_skip (print_name_infos (GlobRef.VarRef id))
 
@@ -612,26 +635,13 @@ let print_instance sigma cb =
     pr_universe_instance sigma inst
   else mt()
 
-let print_constant with_values sep sp udecl =
+let print_constant with_values sp udecl =
   let cb = Global.lookup_constant sp in
   let val_0 = Global.body_of_constant_body Library.indirect_accessor cb in
   let typ = cb.const_type in
-  let univs =
-    let open Univ in
-    let otab = Global.opaque_tables () in
-    match cb.const_body with
-    | Undef _ | Def _ | Primitive _ -> cb.const_universes
-    | OpaqueDef o ->
-      let body_uctxs = Opaqueproof.force_constraints Library.indirect_accessor otab o in
-      match cb.const_universes with
-      | Monomorphic ctx ->
-        Monomorphic (ContextSet.union body_uctxs ctx)
-      | Polymorphic ctx ->
-        assert(ContextSet.is_empty body_uctxs);
-        Polymorphic ctx
-  in
+  let univs = cb.const_universes in
   let uctx =
-    UState.of_binders
+    UState.of_names
       (Printer.universe_binders_with_opt_names (Declareops.constant_polymorphic_context cb) udecl)
   in
   let env = Global.env () and sigma = Evd.from_ctx uctx in
@@ -640,25 +650,26 @@ let print_constant with_values sep sp udecl =
     match val_0 with
     | None ->
         str"*** [ " ++
-        print_basename sp ++ print_instance sigma cb ++ str " : " ++ cut () ++ pr_ltype typ ++
+        print_basename sp ++ print_instance sigma cb ++ str " :" ++ spc () ++ pr_ltype typ ++
         str" ]" ++
         Printer.pr_universes sigma univs
     | Some (c, priv, ctx) ->
       let priv = match priv with
       | Opaqueproof.PrivateMonomorphic () -> None
-      | Opaqueproof.PrivatePolymorphic (_, ctx) -> Some ctx
+      | Opaqueproof.PrivatePolymorphic ctx -> Some ctx
       in
-        print_basename sp ++ print_instance sigma cb ++ str sep ++ cut () ++
-        (if with_values then print_typed_body env sigma (Some c,typ) else pr_ltype typ)++
-        Printer.pr_universes sigma univs ?priv)
+      print_basename sp ++ print_instance sigma cb ++
+      str (if with_values then " =" else " :") ++ spc() ++
+      (if with_values then print_typed_body env sigma (Some c,typ) else pr_ltype typ)++
+      Printer.pr_universes sigma univs ?priv)
 
-let gallina_print_constant_with_infos sp udecl =
-  print_constant true " = " sp udecl ++
+let print_constant_with_infos sp udecl =
+  print_constant true sp udecl ++
   with_line_skip (print_name_infos (GlobRef.ConstRef sp))
 
-let gallina_print_syntactic_def env kn =
-  let qid = Nametab.shortest_qualid_of_syndef Id.Set.empty kn
-  and (vars,a) = Syntax_def.search_syntactic_definition kn in
+let print_abbreviation env kn =
+  let qid = Nametab.shortest_qualid_of_abbreviation Id.Set.empty kn
+  and (vars,a) = Abbreviation.search_abbreviation kn in
   let c = Notation_ops.glob_constr_of_notation_constr a in
   hov 2
     (hov 4
@@ -666,8 +677,9 @@ let gallina_print_syntactic_def env kn =
         prlist (fun id -> spc () ++ Id.print id) (List.map fst vars) ++
         spc () ++ str ":=") ++
      spc () ++
-     Constrextern.without_specific_symbols
-       [Notation.SynDefRule kn] (pr_glob_constr_env env (Evd.from_env env)) c)
+     Vernacstate.System.protect (fun () ->
+         Abbreviation.toggle_abbreviation ~on:false ~use:ParsingAndPrinting kn;
+         pr_glob_constr_env env (Evd.from_env env) c) ())
 
 module DynHandle = Libobject.Dyn.Map(struct type 'a t = 'a -> Pp.t option end)
 
@@ -679,90 +691,72 @@ let handle h (Libobject.Dyn.Dyn (tag, o)) = match DynHandle.find tag h with
    no reason that an object in the stack corresponds to a user-facing
    declaration. It may have been so at the time this was written, but this
    needs to be done in a more principled way. *)
-let gallina_print_leaf_entry env sigma with_values ((sp, kn),lobj) =
-  let sep = if with_values then " = " else " : " in
+let print_library_leaf env sigma with_values mp lobj =
   match lobj with
   | AtomicObject o ->
     let handler =
-      DynHandle.add Declare.Internal.objVariable begin fun _ ->
+      DynHandle.add Declare.Internal.objVariable begin fun id ->
           (* Outside sections, VARIABLES still exist but only with universes
              constraints *)
-          (try Some(print_named_decl env sigma (basename sp)) with Not_found -> None)
+          (try Some(print_named_decl env sigma id) with Not_found -> None)
       end @@
-      DynHandle.add Declare.Internal.Constant.tag begin fun _ ->
-          Some (print_constant with_values sep (Constant.make1 kn) None)
+      DynHandle.add Declare.Internal.Constant.tag begin fun (id,_) ->
+        let kn = Constant.make2 mp (Label.of_id id) in
+        Some (print_constant with_values kn None)
       end @@
-      DynHandle.add DeclareInd.Internal.objInductive begin fun _ ->
-          Some (gallina_print_inductive (MutInd.make1 kn) None)
+      DynHandle.add DeclareInd.Internal.objInductive begin fun (id,_) ->
+        let kn = MutInd.make2 mp (Label.of_id id) in
+        Some (print_inductive kn None)
       end @@
       DynHandle.empty
     in
     handle handler o
-  | ModuleObject _ ->
-    let (mp,l) = KerName.repr kn in
-    Some (print_module with_values (MPdot (mp,l)))
-  | ModuleTypeObject _ ->
-    let (mp,l) = KerName.repr kn in
-          Some (print_modtype (MPdot (mp,l)))
-  | _ -> None
+  | ModuleObject (id,_) ->
+    Some (Printmod.print_module ~with_body:with_values (MPdot (mp,Label.of_id id)))
+  | ModuleTypeObject (id,_) ->
+    Some (print_modtype (MPdot (mp, Label.of_id id)))
+  | IncludeObject _ | KeepObject _ | ExportObject _ -> None
 
-let gallina_print_library_entry env sigma with_values ent =
-  let pr_name (sp,_) = Id.print (basename sp) in
-  match ent with
-    | (oname,Lib.Leaf lobj) ->
-      gallina_print_leaf_entry env sigma with_values (oname,lobj)
-    | (oname,Lib.OpenedSection (dir,_)) ->
-      Some (str " >>>>>>> Section " ++ pr_name oname)
-    | (_,Lib.CompilingLibrary { Nametab.obj_dir; _ }) ->
-      Some (str " >>>>>>> Library " ++ DirPath.print obj_dir)
-    | (oname,Lib.OpenedModule _) ->
-      Some (str " >>>>>>> Module " ++ pr_name oname)
+let decr = Option.map ((+) (-1))
 
-let gallina_print_context env sigma with_values =
+let is_done = Option.equal Int.equal (Some 0)
+
+let print_leaves env sigma with_values mp =
   let rec prec n = function
-    | h::rest when Option.is_empty n || Option.get n > 0 ->
-      (match gallina_print_library_entry env sigma with_values h with
-       | None -> prec n rest
-       | Some pp -> prec (Option.map ((+) (-1)) n) rest ++ pp ++ fnl ())
-    | _ -> mt ()
+    | [] -> n, mt()
+    | o :: rest ->
+      if is_done n then n, mt()
+      else begin match print_library_leaf env sigma with_values mp o with
+        | Some pp ->
+          let n, prest = prec (decr n) rest in
+          n, prest ++ pp
+        | None -> prec n rest
+      end
   in
   prec
 
-let gallina_print_eval red_fun env sigma _ {uj_val=trm;uj_type=typ} =
-  let ntrm = red_fun env sigma trm in
-  (str "     = " ++ gallina_print_typed_value_in_env env sigma (ntrm,typ))
+let print_context env sigma with_values =
+  let rec prec n = function
+    | [] -> mt()
+    | (node, leaves) :: rest ->
+      if is_done n then mt()
+      else
+        let mp = (Lib.node_prefix node).Nametab.obj_mp in
+        let n, pleaves = print_leaves env sigma with_values mp n leaves in
+        if is_done n then pleaves
+        else prec n rest ++ pleaves
+  in
+  prec
 
-(******************************************)
-(**** Printing abstraction layer          *)
+let pr_prefix_name prefix = Id.print (snd (split_dirpath prefix.Nametab.obj_dir))
 
-let default_object_pr = {
-  print_inductive           = gallina_print_inductive;
-  print_constant_with_infos = gallina_print_constant_with_infos;
-  print_section_variable    = gallina_print_section_variable;
-  print_syntactic_def       = gallina_print_syntactic_def;
-  print_module              = gallina_print_module;
-  print_modtype             = gallina_print_modtype;
-  print_named_decl          = gallina_print_named_decl;
-  print_library_entry       = gallina_print_library_entry;
-  print_context             = gallina_print_context;
-  print_typed_value_in_env  = gallina_print_typed_value_in_env;
-  print_eval                = gallina_print_eval;
-}
-
-let object_pr = ref default_object_pr
-let set_object_pr = (:=) object_pr
-
-let print_inductive x = !object_pr.print_inductive x
-let print_constant_with_infos c = !object_pr.print_constant_with_infos c
-let print_section_variable c = !object_pr.print_section_variable c
-let print_syntactic_def x = !object_pr.print_syntactic_def x
-let print_module x  = !object_pr.print_module  x
-let print_modtype x = !object_pr.print_modtype x
-let print_named_decl x = !object_pr.print_named_decl x
-let print_library_entry x = !object_pr.print_library_entry x
-let print_context x = !object_pr.print_context x
-let print_typed_value_in_env x = !object_pr.print_typed_value_in_env x
-let print_eval x = !object_pr.print_eval x
+let print_library_node = function
+  | Lib.OpenedSection (prefix, _) ->
+    str " >>>>>>> Section " ++ pr_prefix_name prefix
+  | Lib.OpenedModule (_,_,prefix,_) ->
+    str " >>>>>>> Module " ++ pr_prefix_name prefix
+  | Lib.CompilingLibrary { Nametab.obj_dir; _ } ->
+    str " >>>>>>> Library " ++ DirPath.print obj_dir
 
 (******************************************)
 (**** Printing declarations and judgments *)
@@ -771,9 +765,9 @@ let print_eval x = !object_pr.print_eval x
 let print_judgment env sigma {uj_val=trm;uj_type=typ} =
   print_typed_value_in_env env sigma (trm, typ)
 
-let print_safe_judgment env sigma j =
-  let trm = Safe_typing.j_val j in
-  let typ = Safe_typing.j_type j in
+let print_safe_judgment {Safe_typing.jdg_env=senv; jdg_val=trm; jdg_type=typ} =
+  let env = Safe_typing.env_of_safe_env senv in
+  let sigma = Evd.from_env env in
   let trm = EConstr.of_constr trm in
   let typ = EConstr.of_constr typ in
   print_typed_value_in_env env sigma (trm, typ)
@@ -792,55 +786,62 @@ let handleF h (Libobject.Dyn.Dyn (tag, o)) = match DynHandleF.find tag h with
 | f -> f o
 | exception Not_found -> mt ()
 
-(* TODO: see the comment for {!gallina_print_leaf_entry} *)
+(* TODO: see the comment for {!print_leaf_entry} *)
+let print_full_pure_atomic env sigma mp lobj =
+  let handler =
+    DynHandleF.add Declare.Internal.Constant.tag begin fun (id,_) ->
+      let kn = KerName.make mp (Label.of_id id) in
+      let con = Global.constant_of_delta_kn kn in
+      let cb = Global.lookup_constant con in
+      let typ = cb.const_type in
+      hov 0 (
+        match cb.const_body with
+        | Undef _ ->
+          str "Parameter " ++
+          print_basename con ++ str " :" ++ spc () ++ pr_ltype_env env sigma typ
+        | OpaqueDef lc ->
+          str "Theorem " ++ print_basename con ++ cut () ++
+          str " : " ++ pr_ltype_env env sigma typ ++ str "." ++ fnl () ++
+          str "Proof " ++ pr_lconstr_env env sigma
+            (fst (Global.force_proof Library.indirect_accessor lc))
+        | Def c ->
+          str "Definition " ++ print_basename con ++ cut () ++
+          str "  : " ++ pr_ltype_env env sigma typ ++ cut () ++ str " := " ++
+          pr_lconstr_env env sigma c
+        | Primitive _ ->
+          str "Primitive " ++
+          print_basename con ++ str " :" ++ spc () ++ pr_ltype_env env sigma typ)
+      ++ str "." ++ fnl () ++ fnl ()
+    end @@
+    DynHandleF.add DeclareInd.Internal.objInductive begin fun (id,_) ->
+      let kn = KerName.make mp (Label.of_id id) in
+      let mind = Global.mind_of_delta_kn kn in
+      let mib = Global.lookup_mind mind in
+      Printmod.pr_mutual_inductive_body (Global.env()) mind mib None ++
+      str "." ++ fnl () ++ fnl ()
+    end @@
+    DynHandleF.empty
+  in
+  handleF handler lobj
+
+let print_full_pure_leaf env sigma mp = function
+  | AtomicObject lobj -> print_full_pure_atomic env sigma mp lobj
+  | ModuleObject (id, _) ->
+    (* TODO: make it reparsable *)
+    print_module (MPdot (mp,Label.of_id id)) ++ str "." ++ fnl () ++ fnl ()
+  | ModuleTypeObject (id, _) ->
+    (* TODO: make it reparsable *)
+    print_modtype (MPdot (mp,Label.of_id id)) ++ str "." ++ fnl () ++ fnl ()
+  | _ -> mt()
+
 let print_full_pure_context env sigma =
   let rec prec = function
-  | ((_,kn),Lib.Leaf AtomicObject lobj)::rest ->
-    let handler =
-      DynHandleF.add Declare.Internal.Constant.tag begin fun _ ->
-          let con = Global.constant_of_delta_kn kn in
-          let cb = Global.lookup_constant con in
-          let typ = cb.const_type in
-          hov 0 (
-            match cb.const_body with
-              | Undef _ ->
-                str "Parameter " ++
-                print_basename con ++ str " : " ++ cut () ++ pr_ltype_env env sigma typ
-              | OpaqueDef lc ->
-                str "Theorem " ++ print_basename con ++ cut () ++
-                str " : " ++ pr_ltype_env env sigma typ ++ str "." ++ fnl () ++
-                str "Proof " ++ pr_lconstr_env env sigma
-                  (fst (Opaqueproof.force_proof Library.indirect_accessor
-                          (Global.opaque_tables ()) lc))
-              | Def c ->
-                str "Definition " ++ print_basename con ++ cut () ++
-                str "  : " ++ pr_ltype_env env sigma typ ++ cut () ++ str " := " ++
-                pr_lconstr_env env sigma c
-              | Primitive _ ->
-                 str "Primitive " ++
-                   print_basename con ++ str " : " ++ cut () ++ pr_ltype_env env sigma typ)
-          ++ str "." ++ fnl () ++ fnl ()
-      end @@
-      DynHandleF.add DeclareInd.Internal.objInductive begin fun _ ->
-          let mind = Global.mind_of_delta_kn kn in
-          let mib = Global.lookup_mind mind in
-          pr_mutual_inductive_body (Global.env()) mind mib None ++
-            str "." ++ fnl () ++ fnl ()
-      end @@
-      DynHandleF.empty
-    in
-    let pp = handleF handler lobj in
+    | (node,leaves)::rest ->
+      let mp = (Lib.node_prefix node).Nametab.obj_mp in
+      let pp = Pp.prlist (print_full_pure_leaf env sigma mp) leaves in
       prec rest ++ pp
-  | ((_,kn),Lib.Leaf ModuleObject _)::rest ->
-          (* TODO: make it reparsable *)
-    let (mp,l) = KerName.repr kn in
-          prec rest ++ print_module true (MPdot (mp,l)) ++ str "." ++ fnl () ++ fnl ()
-  | ((_,kn),Lib.Leaf ModuleTypeObject _)::rest ->
-          (* TODO: make it reparsable *)
-    let (mp,l) = KerName.repr kn in
-          prec rest ++ print_modtype (MPdot (mp,l)) ++ str "." ++ fnl () ++ fnl ()
-  | _::rest -> prec rest
-  | _ -> mt () in
+  | [] -> mt ()
+  in
   prec (Lib.contents ())
 
 (* For printing an inductive definition with
@@ -853,9 +854,9 @@ let read_sec_context qid =
   let dir =
     try Nametab.locate_section qid
     with Not_found ->
-      user_err ?loc:qid.loc ~hdr:"read_sec_context" (str "Unknown section.") in
+      user_err ?loc:qid.loc (str "Unknown section.") in
   let rec get_cxt in_cxt = function
-    | (_,Lib.OpenedSection ({Nametab.obj_dir;_},_) as hd)::rest ->
+    | (Lib.OpenedSection ({Nametab.obj_dir;_},_), _ as hd)::rest ->
         if DirPath.equal dir obj_dir then (hd::in_cxt) else get_cxt (hd::in_cxt) rest
     | [] -> []
     | hd::rest -> get_cxt (hd::in_cxt) rest
@@ -873,9 +874,9 @@ let maybe_error_reject_univ_decl na udecl =
   let open GlobRef in
   match na, udecl with
   | _, None | Term (ConstRef _ | IndRef _ | ConstructRef _), Some _ -> ()
-  | (Term (VarRef _) | Syntactic _ | Dir _ | ModuleType _ | Other _ | Undefined _), Some udecl ->
+  | (Term (VarRef _) | Abbreviation _ | Dir _ | Module _ | ModuleType _ | Other _ | Undefined _), Some udecl ->
     (* TODO Print na somehow *)
-    user_err ~hdr:"reject_univ_decl" (str "This object does not support universe names.")
+    user_err (str "This object does not support universe names.")
 
 let print_any_name env sigma na udecl =
   maybe_error_reject_univ_decl na udecl;
@@ -885,9 +886,8 @@ let print_any_name env sigma na udecl =
   | Term (IndRef (sp,_)) -> print_inductive sp udecl
   | Term (ConstructRef ((sp,_),_)) -> print_inductive sp udecl
   | Term (VarRef sp) -> print_section_variable env sigma sp
-  | Syntactic kn -> print_syntactic_def env kn
-  | Dir (Nametab.GlobDirRef.DirModule Nametab.{ obj_dir; obj_mp; _ } ) ->
-    print_module (printable_body obj_dir) obj_mp
+  | Abbreviation kn -> print_abbreviation env kn
+  | Module mp -> print_module mp
   | Dir _ -> mt ()
   | ModuleType mp -> print_modtype mp
   | Other (obj, info) -> info.print obj
@@ -895,11 +895,9 @@ let print_any_name env sigma na udecl =
   try  (* Var locale de but, pas var de section... donc pas d'implicits *)
     let dir,str = repr_qualid qid in
     if not (DirPath.is_empty dir) then raise Not_found;
-    str |> Global.lookup_named |> print_named_decl env sigma
+    str |> Global.lookup_named |> print_named_decl_gen env sigma
 
-  with Not_found ->
-    user_err
-      ~hdr:"print_name" (pr_qualid qid ++ spc () ++ str "not a defined object.")
+  with Not_found -> user_err ?loc:qid.loc (pr_qualid qid ++ spc () ++ str "not a defined object.")
 
 let print_name env sigma na udecl =
   match na with
@@ -911,24 +909,46 @@ let print_name env sigma na udecl =
   | {loc; v=Constrexpr.AN ref} ->
     print_any_name env sigma (locate_any_name ref) udecl
 
-let print_opaque_name env sigma qid =
-  let open GlobRef in
-  match Nametab.global qid with
-    | ConstRef cst ->
-      let cb = Global.lookup_constant cst in
-      if Declareops.constant_has_body cb then
-        print_constant_with_infos cst None
-      else
-        user_err Pp.(str "Not a defined constant.")
-    | IndRef (sp,_) ->
-      print_inductive sp None
-    | ConstructRef cstr as gr ->
-      let ty, ctx = Typeops.type_of_global_in_context env gr in
-      let ty = EConstr.of_constr ty in
-      let open EConstr in
-      print_typed_value_in_env env sigma (mkConstruct cstr, ty)
-    | VarRef id ->
-      env |> lookup_named id |> print_named_decl env sigma
+let print_notation_grammar env sigma ntn =
+  let ng = List.hd (Notgram_ops.grammar_of_notation ntn) in
+  let assoc = ng.Notation_gram.notgram_assoc in
+  let prdf () = Pp.str "no associativity" in
+  Pp.(pr_opt_no_spc_default prdf Gramlib.Gramext.pr_assoc assoc)
+
+exception PrintNotationNotFound of Constrexpr.notation_entry * string
+
+let () = CErrors.register_handler @@ function
+  | PrintNotationNotFound (entry, ntn_str) ->
+      let entry_string = match entry with
+      | Constrexpr.InConstrEntry -> "."
+      | Constrexpr.InCustomEntry e -> " in " ^ e ^ " entry."
+      in
+      Some Pp.(str "\"" ++ str ntn_str ++ str "\"" ++ spc ()
+        ++ str "cannot be interpreted as a known notation" ++ str entry_string ++ spc ()
+        ++ strbrk "Make sure that symbols are surrounded by spaces and that holes are explicitly denoted by \"_\".")
+  | _ -> None
+
+let error_print_notation_not_found e s =
+  raise @@ PrintNotationNotFound (e, s)
+
+let print_notation env sigma entry raw_ntn =
+  (* make sure entry exists *)
+  let () =
+    match entry with
+    | Constrexpr.InConstrEntry -> ()
+    | Constrexpr.InCustomEntry e -> Metasyntax.check_custom_entry e
+  in
+  (* convert notation string to key. eg. "x + y" to "_ + _" *)
+  let interp_ntn = Notation.interpret_notation_string raw_ntn in
+  let ntn = (entry, interp_ntn) in
+  try
+    let lvl = Notation.level_of_notation ntn in
+    let args = Notgram_ops.non_terminals_of_notation ntn in
+    let pplvl = Metasyntax.pr_level ntn lvl args in
+    Pp.(str "Notation \"" ++ str interp_ntn ++ str "\"" ++ spc ()
+      ++ pplvl ++ pr_comma () ++ print_notation_grammar env sigma ntn
+      ++ str ".")
+  with Not_found -> error_print_notation_not_found entry raw_ntn
 
 let print_about_any ?loc env sigma k udecl =
   maybe_error_reject_univ_decl k udecl;
@@ -944,14 +964,11 @@ let print_about_any ?loc env sigma k udecl =
         print_opacity ref @
   print_bidi_hints ref @
   [hov 0 (str "Expands to: " ++ pr_located_qualid k)])
-  | Syntactic kn ->
-    let () = match Syntax_def.search_syntactic_definition kn with
-    | [],Notation_term.NRef (ref,_) -> Dumpglob.add_glob ?loc ref
-    | _ -> () in
+  | Abbreviation kn ->
       v 0 (
-      print_syntactic_def env kn ++ fnl () ++
+      print_abbreviation env kn ++ fnl () ++
       hov 0 (str "Expands to: " ++ pr_located_qualid k))
-  | Dir _ | ModuleType _ | Undefined _ ->
+  | Dir _ | Module _ | ModuleType _ | Undefined _ ->
       hov 0 (pr_located_qualid k)
   | Other (obj, info) -> hov 0 (info.about obj)
 
@@ -979,7 +996,8 @@ let print_path ((i,j),p) =
   hov 2 (
     str"[" ++ hov 0 (prlist_with_sep pr_semicolon print_coercion_value p) ++
     str"] : ") ++
-  pr_class i ++ str" >-> " ++ pr_class j
+  pr_class i ++ str" >-> " ++ pr_class j ++
+  str (if path_is_reversible p then " (reversible)" else "")
 
 let _ = Coercionops.install_path_printer print_path
 
@@ -997,7 +1015,7 @@ let print_path_between cls clt =
     try
       lookup_path_between_class (cls, clt)
     with Not_found ->
-      user_err ~hdr:"index_cl_of_id"
+      user_err
         (str"No path between " ++ pr_class cls ++ str" and " ++ pr_class clt
          ++ str ".")
   in
@@ -1006,12 +1024,12 @@ let print_path_between cls clt =
 let print_canonical_projections env sigma grefs =
   let open Structures in
   let match_proj_gref { CSTable.projection; value; solution } gr =
-    GlobRef.equal projection gr ||
+    QGlobRef.equal env projection gr ||
     begin match value with
-      | ValuePattern.Const_cs y -> GlobRef.equal y gr
+      | ValuePattern.Const_cs y -> GlobRef.CanOrd.equal y gr
       | _ -> false
     end ||
-    GlobRef.equal solution gr
+    QGlobRef.equal env solution gr
   in
   let projs =
     List.filter (fun p -> List.for_all (match_proj_gref p) grefs)
@@ -1039,7 +1057,7 @@ let print_typeclasses () =
     prlist_with_sep fnl (pr_typeclass env) (typeclasses ())
 
 let pr_instance env i =
-  (*   gallina_print_constant_with_infos i.is_impl *)
+  (*   print_constant_with_infos i.is_impl *)
   (* lighter *)
   print_ref false (instance_impl i) None ++
   begin match hint_priority i with
@@ -1054,6 +1072,5 @@ let print_all_instances () =
 
 let print_instances r =
   let env = Global.env () in
-  let sigma = Evd.from_env env in
-  let inst = instances env sigma r in
-    prlist_with_sep fnl (pr_instance env) inst
+  let inst = instances_exn env (Evd.from_env env) r in
+  prlist_with_sep fnl (pr_instance env) inst
